#define VERSION 1

/* Summarize for BAMBE, Version 1.
 * (c) Copyright 1998, Donald Simon & Bret Larget
 * Department of Mathematics/Computer Science, Duquesne University
 * 7/8/98
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define DEFAULTTHRESHOLD .80
#define MAXCLADETOPS 100

#define MAXLINE 513
#define SSIZE 20707
#define CSIZE 10709
#define MAXLEAVES 64
#define SETSIZE ((MAXLEAVES+30)/31)
#define MAXCLADES 1000
#define MAXTREES 15000
const unsigned int bit = (1<<30);

#define SHASH(k1,k2)	((((k1)<<6)+(k2)) % SSIZE)

void usage_error(char *name)
{
  fprintf(stderr,"Usage: %s [-n skipped_lines] [-p threshold] "
	  "[-c maximum_clade_topologies] <file 1> <file 2> ...\n",
	  name);
  exit(1);
}

typedef unsigned int Set[SETSIZE];

unsigned int chash(const Set s)	
{
  int i;
  unsigned int n;

  n = s[0];
  for(i=1;i<SETSIZE;i++)
    n = ((n << 6) + s[i]) % CSIZE;
  return n;
}

void make_empty_set(Set s)
{
  int i;

  for(i=0;i<SETSIZE;i++)
    s[i] = 0;
}

void copy_set(Set s1, const Set s2)
{
  int i;
  
  for(i=0;i<SETSIZE;i++)
    s1[i] = s2[i];
}

void set_union(const Set s1, const Set s2, Set s)
{
  int i;

  for(i=0;i<SETSIZE;i++)
    s[i] = s1[i] | s2[i];
}

void set_difference(const Set s1, const Set s2, Set s)
{
  int i;

  for(i=0;i<SETSIZE;i++)
    s[i] = s1[i] & ~s2[i];
}

int set_equal(const Set s1, const Set s2)
{
  int i;

  for(i=0;i<SETSIZE;i++)
    if(s1[i] != s2[i])
      return 0;
  return 1;
}

int subset (const Set s1, const Set s2)
{
  int i;

  for(i=0;i<SETSIZE;i++)
    if(s1[i] & ~s2[i])
      return 0;
  return 1;
}

void singleton(const int n, Set s)
{
  int i;

  for(i=0;i<SETSIZE;i++)
    s[i] = 0;
  s[(n-1)/31] = (1 << (30 - ((n-1) % 31)));
}

void print_set(const Set s)
{
  int first,i,j;
  unsigned int n;

  first = 1;
  printf("{");
  for(i=0;i<SETSIZE;i++) {
    n = s[i];
    for(j=1;j<32;j++,n<<=1)
      if(n & bit) {
	if(first)
	  first = 0;
	else 
	  printf(",");
	printf("%d",j+31*i);
      }
  }
  printf("}");
}

void print_elements(const Set s)
{
  int i,j;
  unsigned int n;

  for(i=0;i<SETSIZE;i++) {
    n = s[i];
    for(j=1;j<32;j++,n<<=1)
    if(n & bit)
      printf("%12s%3d\n","",j+31*i);
  }
}

int cmp_set(const Set s1, const Set s2)
{
  int c,i;

  for(i=0;i<SETSIZE;i++)
    if(c=(s1[i]-s2[i]))
      return c;
  return 0;
}

int cindex[CSIZE], sindex[SSIZE];
int num_clades = 0, num_trees = 0;

struct {
  Set set;
  int count;
  int trees;
  int first;
  int name;
  int time;
  int lasttopnum;
  int trans[MAXCLADETOPS+1][MAXCLADETOPS+1];
} ctab[MAXCLADES];

struct {
  int left;
  int right;
  int count;
  int clade;
  int next;
  int name;
  int num;
} stab[MAXTREES];

void print_tree(int k,int print_id)
{
  if(k<0)
    printf("%d",-k);
  else if(stab[k].name) {
    printf("%c",ctab[stab[k].clade].name+'A'-1);
    if(print_id)
      printf("%d",stab[k].name);
  }
  else {
    printf("(");
    print_tree(stab[k].left,print_id);
    printf(",");
    print_tree(stab[k].right,print_id);
    printf(")");
  }
}

void add_trans(int c, int t, int time, int cladetops)
{
  if(ctab[c].trees<=cladetops) {
    if(ctab[c].time == time-1)
      ctab[c].trans[ctab[c].lasttopnum][stab[t].num]++;
    else {
      ctab[c].trans[ctab[c].lasttopnum][0]++;
      ctab[c].trans[0][stab[t].num]++;
    }
    ctab[c].time = time;
    ctab[c].lasttopnum = stab[t].num;
  }
}

int increment(const Set s, unsigned int k, int i1, int i2, int time,
	      int cladetops)
{
  /* s is the set of taxa. k is the hash value of the subtree.
     i1 and i2 are the indices into the subtree table of the left
     and right subtrees, respectively. */

  int c,t;
  unsigned int j;

  while(t=sindex[k])
    if(stab[t].left == i1 && stab[t].right == i2) {
      stab[t].count++;
      c = stab[t].clade;
      ctab[c].count++;
      add_trans(c,t,time,cladetops);
      return t;
    }
    else
      k = (k+1) % SSIZE;

  if(num_trees == MAXTREES) {
    fprintf(stderr,"Error: Too many distinct subtrees.\n");
    exit(0);
  }

  t = sindex[k] = ++num_trees;
  stab[t].left = i1;
  stab[t].right = i2;
  stab[t].count = 1;

  j = chash(s);
  while(c=cindex[j])
    if(set_equal(ctab[c].set,s)) {
      stab[t].clade = c;
      ctab[c].count++;
      stab[t].num = ++ctab[c].trees;
      stab[t].next = ctab[c].first;
      ctab[c].first = t;
      stab[t].name = 0;
      add_trans(c,t,time,cladetops);
      return t;
    }
    else
      j = (j+1) % CSIZE;

  if(num_clades == MAXCLADES) {
    fprintf(stderr,"Error: Too many distinct clades.\n");
    exit(0);
  }

  c = cindex[j] = ++num_clades;
  stab[t].clade = c;
  copy_set(ctab[c].set,s);
  ctab[c].count = 1;
  ctab[c].trees = 1;
  stab[t].next = -1;
  ctab[c].first = t;
  stab[t].name = 0;
  ctab[c].name = 0;
  ctab[c].time = time;
  ctab[c].lasttopnum = stab[t].num = 1;
  if(time>1) 
    ctab[c].trans[0][1]++;
  return t;
}
  
char *store_tree (char *top, int time, Set s, unsigned int *k, int *i,
		  int cladetops)
{
  char *top2, *top3;
  Set s1,s2;
  unsigned int k1,k2;
  int i1,i2;

  if(*top == '\0')
    return top;
  if(*top == '(') {
    top2 = store_tree(top+1,time,s1,&k1,&i1,cladetops);
    if(*top2 == ',')
      top2++;
    else {
      fprintf(stderr,"Error: Missing comma - /%s/\n",top);
      exit(1);
    }
    top3 = store_tree(top2,time,s2,&k2,&i2,cladetops);
    if(*top3 != ')') {
      fprintf(stderr,"Error: Mismatched parentheses - /%s/\n",top);
      exit(1);
    }
    set_union(s1,s2,s);
    *k = SHASH(k1,k2);
    *i = increment(s, *k, i1, i2, time, cladetops);
    return top3+1;
  }
  *k = 0;
  for(;isdigit(*top);top++)
    *k = 10 * *k + (*top-'0');
  if(*k>MAXLEAVES) {
    fprintf(stderr,"Error: taxa number too large - %d\n",*k);
    exit(1);
  }
  *i = -*k;
  singleton(*k,s);
  return top;
}

char *make_set (char *top, Set s)
{
  char *top2, *top3;
  Set s1,s2;
  int k;

  if(*top == '\0')
    return top;
  if(*top == '(') {
    top2 = make_set(top+1,s1);
    if(*top2 == ',')
      top2++;
    else {
      fprintf(stderr,"Error: Missing comma - /%s/\n",top);
      exit(1);
    }
    top3 = make_set(top2,s2);
    if(*top3 != ')') {
      fprintf(stderr,"Error: Mismatched parentheses - /%s/\n",top);
      exit(1);
    }
    set_union(s1,s2,s);
    return top3+1;
  }
  k = 0;
  for(;isdigit(*top);top++)
    k = 10 * k + (*top-'0');
  if(k>MAXLEAVES) {
    fprintf(stderr,"Error: taxa number too large - %d\n",k);
    exit(1);
  }
  singleton(k,s);
  return top;
}

int cmp_tops2_int(int a, int b)
{ int c;

  if(a==b) 
    return 0;
  if(a<0)
    if(b<0)
      return b-a;
    else
      return -1;
  else if(b<0)
    return 1;
  else {
    if(stab[a].name)
      if(stab[b].name)
	return ctab[stab[a].clade].name - ctab[stab[b].clade].name;
      else 
	return -1;
    else if(stab[b].name)
      return 1;
    else 
      return ((c = cmp_tops2_int(stab[a].left,stab[b].left)) ? c :
	      cmp_tops2_int(stab[a].right,stab[b].right));
  }
}

int cmp_tops2(const void *a, const void *b)
{
  return (cmp_tops2_int(*((int *)a),*((int *)b)));
}

int cmp_tops(const void *a, const void *b)
{
  return stab[*((int *)b)].count - stab[*((int *)a)].count;
}
  

int cmp_clades(const void *a, const void *b)
{
  return cmp_set(ctab[*((int *)b)].set,ctab[*((int *)a)].set);
}

void add_clade(int i, int clades[], int *num)
{
  int j=0;

  while(j<*num)
    if(subset(ctab[i].set,ctab[clades[j]].set))
      return;
    else if(subset(ctab[clades[j]].set,ctab[i].set))
      clades[j] = clades[--(*num)];
    else
      j++;
  clades[(*num)++] = i;
}

int print_clades(int min, int maxsub, int clades[], Set universal_set)
{
  unsigned int m;
  int i,j,k,l,c;
  int num=0;
  int tops[MAXCLADETOPS];
  Set allset;

  make_empty_set(allset);
  for(m=0;m<CSIZE;m++)
    if((c=cindex[m]) && ctab[c].count >= min && ctab[c].trees <= maxsub) 
      add_clade(c,clades,&num);
  qsort(clades,num,sizeof(int),cmp_clades);
  for(i=0;i<num;i++) {
    j = clades[i];
    ctab[j].name = i+1;
    set_union(ctab[j].set,allset,allset);
    printf("%6d    %c  ",ctab[j].count,i+'A');
    print_set(ctab[j].set);
    printf("\n");
    for(k=ctab[j].first,l=0;k!=-1;k=stab[k].next,l++)
      tops[l] = k;
    qsort(tops,l,sizeof(tops[0]),cmp_tops);
    for(k=0;k<l;k++) {
      printf("  %6d  %c%d ",stab[tops[k]].count,i+'A',k+1);
      print_tree(tops[k],1);
      printf("\n");
      stab[tops[k]].name = k+1;
    }
    printf("\n");
  }
  set_difference(universal_set,allset,allset);
  print_elements(allset);
  return num;
}

void print_trees(int trees[], int n, int total, int print_id)
{
  int sum = 0, k, m;

  qsort(trees,n,sizeof(int),cmp_tops);
  for(k=0;k<n;k++) {
    m = stab[trees[k]].count;
    sum += m;
    printf("%5d %6.3lf %6.3lf ",m,m/(double)total,sum/(double)total);
    print_tree(trees[k],print_id);
    printf("\n");
  }
}

void print_trans(const int clades[], int n, int total)
{
  int i, j, k, s, sum;
  int zerorow, zerocol;
  char cname;
  int (*clade_matrix)[MAXCLADETOPS+1];
  int num[MAXCLADETOPS+1];

  for(i=0;i<n;i++) {
    s = ctab[clades[i]].trees;
    cname = ctab[clades[i]].name+'A'-1;
    clade_matrix = ctab[clades[i]].trans;
    for(k=ctab[clades[i]].first;k!=-1;k=stab[k].next)
      num[stab[k].name] = stab[k].num;
    printf("     | ");
    for(j=1;j<s+1;j++)
      printf("  %c%-2d ",cname,j);
    zerocol = zerorow = sum = 0;
    for(k=0;k<s+1;k++)
      for(j=0;j<s+1;j++)
	sum += clade_matrix[k][j];
    clade_matrix[0][0] = total - 1 - sum;
    for(k=0;k<s+1;k++) {
      zerocol = zerocol || clade_matrix[k][0];
      zerorow = zerorow || clade_matrix[0][k];
    }
    if(zerocol)
      printf("   -  ");
    printf("\n-----+");
    for(j=1;j<s;j++)
      printf("------");
    if(zerocol)
      printf("------");
    printf("-----");
    printf("\n");
    for(k=1;k<s+1;k++) {
      printf(" %c%-2d |",cname,k);
      for(j=1;j<s+1;j++)
	printf("%5d ",clade_matrix[num[k]][num[j]]);
      if(zerocol)
	printf("%5d ",clade_matrix[num[k]][0]);
      printf("\n");
    }
    if(zerorow) {
      printf("  -  |");
      for(j=1;j<s+1;j++)
	printf("%5d ",clade_matrix[0][num[j]]);
      if(zerocol)
	printf("%5d ",clade_matrix[0][0]);
      printf("\n");
    }
    printf("\n\n");
  }
}

void add_zero_trans(int i, int time)
{
  int c;

  if(i>0) {
    c = stab[i].clade;
    if(ctab[c].time != time)
      ctab[c].trans[ctab[c].lasttopnum][0]++;
    add_zero_trans(stab[i].left,time);
    add_zero_trans(stab[i].right,time);
  }
}

void print_clade_count(int k, int total)
{
  int c;

  if(k>=0) {
    c = stab[k].clade;
    /*printf("%10d %5d ",ctab[c].count,ctab[c].trees);*/
    printf("%10d %6.3lf ",ctab[c].count,(double)ctab[c].count/(double)total);
    print_set(ctab[c].set);
    printf("\n");
    print_clade_count(stab[k].left,total);
    print_clade_count(stab[k].right,total);
  }
}

void summarize(int n, double threshold, int cladetops,
	       char **names, int num_names)
{
  FILE *f;
  char s[MAXLINE];
  int total=0;
  Set set,universal_set;
  int i,j,l,r;
  unsigned int k;
  unsigned int m;
  int trees[SSIZE];
  int clades[MAXLEAVES],num;
  char **name;

  for(name=names;name<names+num_names;name++) {
    if(strcmp(*name,"-")==0)
      f = stdin;
    else if(!(f=fopen(*name,"r"))) {
      fprintf(stderr,"Error: Could not open %s\n",*name);
      exit(1);
    }
    for(i=0;i<n;i++)
      if(fgets(s,MAXLINE,f)==NULL) {
	fprintf(stderr,"Error: Input contains fewer than %d lines.\n",n);
	exit(1);
      }
    /*n = 0;	set number of skipped lines to 0 for successive files. */
    while(fgets(s,MAXLINE,f))
      store_tree(s, ++total, set, &k, &i, cladetops);
    fclose(f);
  }

  if(total==0) {
    fprintf(stderr,"Error: Input contains exactly %d lines.\n",n);
    exit(1);
  }

  add_zero_trans(i,total);
  make_set(s,universal_set);
  printf("******************** Named clades "
	 "********************\n\n");
  num = print_clades((int)total*threshold,cladetops,clades,universal_set);
  printf("\n");

  m = chash(universal_set);
  while(!(set_equal(ctab[cindex[m]].set,universal_set)))
      m = (m+1) % CSIZE;
  for(r=ctab[cindex[m]].first,l=0;r!=-1;r=stab[r].next,l++)
    trees[l] = r;
  printf("******************** Tree topologies "
	 "********************\n\n");
  printf("Count  Prob.  Cum.  Tree topology\n");
  print_trees(trees,l,total,1);
  printf("\n");

  printf("***** Posterior probabilities of clades in most probable tree "
	 "topology *****\n\n");
  printf("     Count  Prob. Tree topology\n");
  print_clade_count(trees[0],total);
  printf("\n");

  printf("******************** Clade transition matrices "
	 "********************\n\n");
  print_trans(clades,num,total);

  qsort(trees,l,sizeof(int),cmp_tops2);
  for(r=1,i=0;r<l;r++)
    if(cmp_tops2_int(trees[i],trees[r])==0)
      stab[trees[i]].count += stab[trees[r]].count;
    else
      trees[++i] = trees[r];

  printf("******************** Clade tree topologies "
	 "********************\n\n");
  printf("Count  Prob.  Cum.  Tree topology\n");
  print_trees(trees,i+1,total,0);
}

int main(int argc, char *argv[])
{
  int n=0,cladetops=MAXCLADETOPS,num_names=0;
  double threshold=DEFAULTTHRESHOLD;
  char *names[100],**p;
  int i;

  printf("******************** BAMBE Summarize Version %d "
	 "********************\n\n",
	 VERSION);
  printf("Invocation: ");
  for(i=0;i<argc;i++)
    printf("%s ",argv[i]);
  printf("\n\n");
  if(argc > 1) {
    for(p=argv+1;p<argv+argc;p++) {
      if(strcmp(*p,"-n")==0) {
	if(++p>=argv+argc || sscanf(*p,"%d",&n)!=1)
	  usage_error(argv[0]);
      }
      else if(strcmp(*p,"-p")==0) {
	if(++p>=argv+argc || sscanf(*p,"%lf",&threshold)!=1)
	  usage_error(argv[0]);
	if(threshold < 0.5 || threshold > 1.0) {
	  fprintf(stderr,"Error: threshold value (%7.3lf) must be between "
		  "0.5 and 1.0.\n",threshold);
	  exit(1);
	}
      }
      else if(strcmp(*p,"-c")==0) {
	if(++p>=argv+argc || sscanf(*p,"%d",&cladetops)!=1)
	  usage_error(argv[0]);
	if(cladetops < 2 || cladetops > MAXCLADETOPS) {
	  fprintf(stderr,"Error: Number of topologies (%d) of a named clade "
		  "must be between 2 and %d.\n",
		  cladetops,MAXCLADETOPS);
	  exit(1);
	}
      }
      else if(num_names<100)
	names[num_names++] = *p;
    }
    summarize(n,threshold,cladetops,names,num_names);
  }
  else 
    usage_error(argv[0]);
  return 0;
}
