#define VERSION		"1.02 Beta"

/* Bayesian Analysis in Molecular Biology and Evolution (BAMBE) version 1.02
 * Beta
 * (c) Copyright 1999, Donald Simon & Bret Larget
 * Department of Mathematics/Computer Science, Duquesne University
 */

typedef int int32;
typedef unsigned int uint32;

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>

#define REROOT_PROB	.01
#define MAXIT           20 
#define TOLERANCE       1.0e-14

#define MAX_LINE 	257
#define MAX_CATEGORIES	10
#define NUM_BASES	4
#define NUM_SYMBOLS     16
#define GAP		0xf
#define LAMBDA1		1.3862943611198906  /* 2 * log(2) */ 
#define LAMBDA2		0.19062035960864987 /* 2 * log(1.1) */ 
#define SKIPLINE(f,c)   {while(((c)=getc(f))!=EOF && (c)!='\n');}
#define SWAP(t,a,b)	{(t)=(a);(a)=(b);(b)=(t);}
#define MINIMUM(a,b)	((a)<(b)?(a):(b))
#define MAXIMUM(a,b)	((a)>(b)?(a):(b))
#define LEAF(n)		((n)<num_leaves)

/**************************************************************************

  Type Declarations

**************************************************************************/

/*
 *  typedef struct parameters
 *
 *  pi[b] = p  <==>  the equilibrium proportion of base b is p
 *  kappa      <==>  the rate of transition over the rate of transversion
 */

/* Parameters of the probability model. */

/* typedef struct {...} Parameters[num_tags]; */
typedef struct parameters
{
  double kappa, theta, ttp, gamma;
  double pi[NUM_BASES];
  double pi_r, pi_y, pry, pyr;
} *Parameters;


/*
 *  num_leaves	     is the number of taxa
 *  num_nodes	     is 2*num_leaves - 1
 *  num_all_sites    is the total number of sites
 *  num_unique_sites is the number of unique sites
 */

int num_leaves;
int num_nodes;
int num_all_sites,num_unique_sites;


/* The phylogeny tree. */
/* struct tree { int lchild[num_nodes], rchild[num_nodes], parent[num_nodes];
                 double branclen[num_nodes];
		 int root;
		 double loglikelihood; } */
typedef struct tree
{
  int *lchild, *rchild, *parent;
  double *branchlen;
  int root;
  double loglikelihood;
} Tree;

void out_of_space(char *name)
{
  fprintf(stderr,"Error: Ran out of space allocating space for %s.\n",
	  name);
  exit(1);
}

Tree *make_tree(void)
{
  Tree *t;

  if(!(t=(Tree *)malloc(sizeof(struct tree))))
    out_of_space("tree");
  if(!(t->lchild=(int *)malloc(num_nodes*sizeof(int))))
    out_of_space("lchild");
  if(!(t->rchild=(int *)malloc(num_nodes*sizeof(int))))
    out_of_space("rchild");
  if(!(t->parent=(int *)malloc(num_nodes*sizeof(int))))
    out_of_space("parent");
  if(!(t->branchlen=(double *)malloc(num_nodes*sizeof(double))))
    out_of_space("branchlen");
  return t;
}

void free_tree(Tree *t)
{
  free(t->lchild);
  free(t->rchild);
  free(t->parent);
  free(t->branchlen);
  free(t);
}

/* Files to which the output is written. */

typedef struct {
  FILE *flogp,*fkappa,*ftop,*fout,*ftree;
} Files;

/* Acceptance counts. */

typedef struct {
  int accept, accepts, kaccept, kaccepts;
} Accept_counts;

/* Possible relationships between a node and its neighbor. */

typedef enum { LEFTCHILD, RIGHTCHILD, PARENTL, PARENTR, ROOTLEFT, 
	       ROOTRIGHT } 
  Relation;

/* A pair consisting of a node and its relationship one of its neighbors. */

typedef struct {
  int nd;
  Relation rel;
} Neighbor;

/* Information necessary to restore the tree and associated global arrays
   back to their original state after a proposed tree is rejected. */

/* typedef struct {... int nvs[num_nodes] ...} Saved_state; */
typedef struct saved_state {
  Tree *t;
  int *nvs;
  int n, s, rotated;
  Neighbor a, b, c, d, e;
  double loglikelihood;
} *Saved_state;

Saved_state make_saved_state(void)
{
  Saved_state ss;

  if(!(ss=(Saved_state)malloc(sizeof(struct saved_state))))
    out_of_space("saved_state");
  ss->t = make_tree();
  if(!(ss->nvs=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("nvs");
  return ss;
}

void free_saved_state(Saved_state ss)
{
  free(ss->nvs);
  free_tree(ss->t);
  free(ss);
}

/*  A base is represented by a bit vector of size 4. If the base is
 *  known only one bit is set. If the base may be one of several 
 *  (uncertain) then each corresponding bit is set. A gap is
 *  indicated by having all four bits cleared.
 *
 *  A is coded as 0001, G is coded as 0010, C is coded as 0100, 
 *  and T is coded as 1000.
 *
 *  A site is a sequence of bases. A bit vector comprised of an array
 *  of integers represents sites. The bit vectors representing the bases
 *  are packed into the integers. The sign bit of each integer is unused
 *  so that the comparison function used by quicksort can subtract
 *  two integers. On our systems, an integer is 32 bits, so 31 bits are used.
 *  The non-sign bits of the integers in the array are used consecutively,
 *  so that a base may be split between two integers.
 * 
 *  Sites are stored from left to right in the order the taxa are read in.
 */

/*typedef int Site[(num_leaves*NUM_BASES+30)/31];*/
typedef int *Site;

/* When first read in, all sites in the input data. The array is
   then compacted so that each unique site is listed once and
   a count of the the number of times that site occurs in the data
   is kept. In both cases, the array is ordered so that all sites
   with tag 0 occur first, then all sites with tag 1, etc. */

/* int sites[num_all_sites][site_len]; where site_len = 
   (num_leaves*NUM_BASES+30)/31]; */
Site sites;
#define SITES(s)	(sites+(s)*site_len)

/* first_site[n] is the index of the first site in array sites with
   tag n. first_site[num_tags] is the number of sites in the array. */

/* int first_site[num_tags+1]; */
int *first_site;

/*
 *  A mask is a special type of site.
 *  Each node has a mask.
 *  The bits corresponding to the taxon in the subtree rooted at the node
 *  are 1, with all others being 0.
 *  mask[0..num_leaves-1] store the unchanging masks for a leaf. 
 *  They are indexed by the node number.
 */

/* The only bits in mask[n] that are set are the ones that correspond
   to taxon of the leaves of the tree rooted at node n. */

/* int mask[num_nodes][site_len]; */
Site mask;
#define MASK(n)	(mask+(n)*site_len)

/* typedef int SiteArray[NUM_BASES][site_len]; */
typedef int *SiteArray;

/* constant_mask[b] has the bits corresponding to base b set for every
   taxa. */

/* int constant_mask[NUM_BASES][site_len]; */
SiteArray constant_mask; 
#define CONSTANT_MASK(n) (constant_mask+(n)*site_len)
#define CONST_MASK(n)    (const_mask+(n)*site_len)


/*  The number of integers needed to store one site or mask. */

int site_len;

int siteeq(Site s1, Site s2)
{
  /* Returns 1 if sites s1 and s2 are equal. Returns 0 otherwise. */
  int i;

  for(i=0;i<site_len;i++)
    if(s1[i]!=s2[i])
      return 0;
  return 1;
}

void siteand(Site s, Site s1, Site s2)
{
  /* s is the bitwise conjunction of s1 and s2. */
  int i;

  for(i=0;i<site_len;i++)
    s[i] = (s1[i] & s2[i]);
}

void siteor(Site s, Site s1, Site s2)
{
  /* s is the bitwise disjunction of s1 and s2. */
  int i;

  for(i=0;i<site_len;i++)
    s[i] = (s1[i] | s2[i]);
}

void copysite(Site s1, Site s2)
{
  /* s2 is copied to s1. */
  int i;

  for(i=0;i<site_len;i++)
    s1[i] = s2[i];
}

void printsite(Site s)
{
  /* Prints out s. */
  int i;

  printf("0x%08x",s[0]);
  for(i=1;i<site_len;i++)
    printf(" 0x%08x",s[i]);
}

void fprintsite(FILE *f,Site s)
{
  /* Prints out s. */
  int i;

  fprintf(f,"0x%08x",s[0]);
  for(i=1;i<site_len;i++)
    fprintf(f," 0x%08x",s[i]);
}

int sitemember(Site s, SiteArray l, int n)
{
  /* Returns an index into array of sites l (of size n) whose element is
     equal to s if one exists. Returns n otherwise. */
  int i = 0;

  while(i<n && !siteeq(s,l+i*site_len))
    i++;
  return i;
}

/* Sites are sorted by the lexicographical ordering of (bit vector of bases,
   index into the global site array). Numbered_site is used for the sort. */

/* typedef struct numbered_site {int index, n, site[site_len];}
   Numbered_site; */

typedef int *Numbered_site;

#define NS_INDEX(ns)	((ns)[0])
#define NS_N(ns)	((ns)[1])
#define NS_SITE(ns)	((ns)+2)
#define SL(n)		(sl+((n)*(site_len+2)))
/* 
 *  cmp_nsites compares two Numbered_sites
 *
 *  Receive: *n1, a Numbered_site
 *	   : *n2, a Numbered_site
 *  Returns: -# if the first site precedes the second in lex order,
 *	      0 if two sites are the same, and 
 *	     +# if the first site follows the second in lex order.
 */

int cmp_nsites(const void *n1, const void *n2)
{
  int c,i;

  for(i=0;i<site_len;i++)
    if((c = NS_SITE((int *)n1)[i] - NS_SITE((int *)n2)[i])!=0)
      return c;
  return NS_INDEX((int *)n1) - NS_INDEX((int *)n2);
}

/* ordered_names is a sorted list of the taxon names (with the taxon
   numbers attached) for faster lookup of the names. */

typedef struct {
  int n;
  char name[MAX_LINE];
} Numbered_name;

/* Numbered_name ordered_names[num_leaves]; */
Numbered_name *ordered_names;

int cmp_name(const void *a, const void *b)
{
  return strcmp(((Numbered_name *)a)->name,((Numbered_name *)b)->name);
}

/* Trees can be constructed from an inorder listing of the nodes, plus
   the depth of each node (distance from node to root). Dlist is a 
   doubly-linked list containing the inorder traversal of the nodes.
   done is a flag indicating the a node has been processed and has
   be added to the tree. */

typedef struct dlist {
  int node,prev,next,done;
} Dlist;

/* Probability transition matrix. */

/* typedef double Table[num_tags][NUM_BASES][NUM_BASES]; */
typedef double *Table;
#define PTAB(tg,b)	(ptab+((tg)*NUM_BASES+b)*NUM_BASES)

/*****************

 Functions for picky machines.

*****************/

#ifdef ALPHA
#define SAFE_EXP(x) ((x)<-200.0 ? 0.0 : exp(x))
#else
#define SAFE_EXP(x) (exp(x))
#endif

/***********************************************************************

 Global Variables

************************************************************************/

/*  
 *  seed is an odd unsigned int between 0 and (4294967295 = 2^32 - 1)
 *  num_cycles is the number of cycles
 *  sample_interval is the interval for storing the tree
 *  param_update_interval is the interval for updating parameters
 *  initial_kappa is the inital value for kappa
 *  valley_win is the size of half the window for updating valley depths
 *  pi_win is the size of half the window for updating equilibrium 
 *  proportions
 *  kappa_win is the size of half the window for updating kappa
 *  max_depth is the maximum size of a valley depth for the initial tree
 *  data_file contains the name of the file where the data is stored.
 */

unsigned int seed;
int num_cycles, sample_interval, param_update_interval, window_interval;
int burn, num_subcycles, update_kappa, update_theta, update_pi;
int tune_interval;
double *initial_kappa, *initial_theta;
double *initial_ttp, *initial_gamma; 
double valley_win, pi_win;
double input_kappa[MAX_CATEGORIES], input_theta[MAX_CATEGORIES];
double input_ttp[MAX_CATEGORIES], input_gamma[MAX_CATEGORIES];
double kappa_win, max_depth;
char data_file[81], tree_file[81], file_root[81];
int num_tags, num_categories;
int burn_algorithm,main_algorithm,single_kappa,mclock,model,outgroup;
/* char taxaname[num_leaves][MAX_LINE]; */
char *taxaname;
#define TAXANAME(n)	(taxaname+(n)*MAX_LINE)

int newick_format;
double theta_const, pi_const;

/*
 * num_csites[i] = j <==>  node i has j canonical sites
 * csites[i][j] = k  <==>  the jth canonical site of node i is site[k]
 * lsites[i][k] = j  <==>  the canonical site for site[k] at node i is 
 *				site[csites[i][j]]; in other words, lsites
 *			   for node i tells you where in the array csites
 *			   for node i you need to look to find the actual
 *			   site number.
 *			   *** lsites is only set at node i for sites
 *			       which are canonical sites for the parent of i.
 *			       lsites is set for all sites at the root ***
 * tag_symbols[b] = c  <==>  base b appeared c times in the datafile
 * sitecount[k] = c  <==>  site[k] appeared c times
 * base[i][k] = b    <==>  node i has base b at site k
 * frag[i][k][b] = p  <==>  given the parent of i has value b
 *			       at site csite[i][k],
 *			   the probability of observing all data at site k
 *			   in the subtree rooted at i is p
 */

/* csites[n] is the list of canonical variable sites for node n. */

/* int csites[num_nodes][num_unique_sites]; */
int *csites;
#define CSITES(n)	(csites+(n)*num_unique_sites)

/* first_csite[n][tg] is the first site in the list of canonical variable
   sites for node n whose tag is tg. first_csite[n][num_tags] is the number
   of canonical variable sites for node n. */

/* int first_csite[num_nodes][num_tags+1]; */
int *first_csite;
#define FIRST_CSITE(n)	(first_csite+(n)*(num_tags+1))

/* root_ksites[tg][b] is the number of constant sites of the root
   of base b with tag tg. */

/* int root_ksites[num_tags][NUM_BASES]; */
int *root_ksites;
#define ROOT_KSITES(tg) (root_ksites+(tg)*NUM_BASES)

/* lsites[c][i] is the partial calculation (frag or kfrag) stored with 
   the child c for its parent n and the parent's canonical site i. */

/* double *lsites[num_nodes][num_unique_sites]; */
double **lsites;
#define LSITES(n)	(lsites+(n)*num_unique_sites)

/* tag_symbols[tg][s] is the number of times symbol s occurs with tag tg
   in the data. */

/* int tag_symbols[num_tags][NUM_SYMBOLS]; */
int *tag_symbols;
#define TAG_SYMBOLS(tg) (tag_symbols+(tg)*NUM_SYMBOLS)

/* tag_sites[tg]  This may not be properly set! */

/* int tag_sites[num_tags]; */
int *tag_sites;

/* double tag_weights[num_tags]; */
double *tag_weights;

/* sitecount[i] is the number of times that site i occurs in the data. */

/* int sitecount[num_all_sites]; */
int *sitecount;

/* taxa_symbols[n][s] is the number of times symbol s occurs in taxon n
   in the data. */

/* int taxa_symbols[num_leaves][NUM_SYMBOLS]; */
int *taxa_symbols;
#define TAXA_SYMBOLS(n) (taxa_symbols+(n)*NUM_SYMBOLS)

/* total_symbols[s] is the number of times symbol s occurs in the data. */

int total_symbols[NUM_SYMBOLS];

/* frag[c][i][pb] is a stored partial calculation for child c and site i 
   of c.
   This stored result is used for calculating the likelihood for the parent,
   n, of c, for base pb, and for a site j of n whose restriction to
   c is i. Site i (and hence, site j) are variable sites. */

/* double frag[num_nodes][num_unique_sites][NUM_BASES]; */
double *frag;
#define FRAG(n)	        (frag+(n)*num_unique_sites*NUM_BASES)
#define FRAG2(n,s)	(frag+((n)*num_unique_sites+(s))*NUM_BASES)

/* A copy of part of the frag array used to restore the frag array
   after a proposed tree has been rejected. */

/* double saved_frag[num_nodes][num_unique_sites][NUM_BASES]; */
double *saved_frag;
#define SAVED_FRAG(n)	 (saved_frag+(n)*num_unique_sites*NUM_BASES)
#define SAVED_FRAG2(n,s) (saved_frag+((n)*num_unique_sites+(s))*NUM_BASES)

/* kfrag[c][tg][b][pb] is a stored partial calculation for child c and
   a site which is the constant base b with tag tg. The stored result is
   used for calculating the likelihood for the parent, n,
   of c, whose base is pb, for a site whose restriction to c is the
   constant site b. */

/* double kfrag[num_nodes][num_tags][NUM_BASES][NUM_BASES]; */
double *kfrag;
#define KFRAG(n)         (kfrag+(n)*num_tags*NUM_BASES*NUM_BASES)
#define KFRAG2(n,tg,b)   (kfrag+((((n)*num_tags+(tg))*NUM_BASES)+(b))* \
			  NUM_BASES)

/* double saved_kfrag[num_nodes][num_tags][NUM_BASES][NUM_BASES]; */
double *saved_kfrag;
#define SAVED_KFRAG(n)       (saved_kfrag+(n)*num_tags*NUM_BASES*NUM_BASES)
#define SAVED_KFRAG2(n,tg,b) (saved_kfrag+((((n)*num_tags+(tg))*NUM_BASES)+ \
					  (b))*NUM_BASES)

/* ufrag[c][tg][s][pb] is a stored partial calculation for child c (a leaf),
   for site s, and a single uncertain base s, with tag tg. The stored
   result is used for calculating the likelihood for the parent, n,
   of c, whose base is pb, for a site whose restriction to c is the
   the constant uncertain base s. */

/* double ufrag[num_leaves][num_tags][NUM_SYMBOLS][NUM_BASES]; */
double *ufrag;
#define UFRAG(n)       (ufrag+(n)*num_tags*NUM_SYMBOLS*NUM_BASES)
#define UFRAG3(n,tg,b) (ufrag+((((n)*num_tags+(tg))*NUM_SYMBOLS)+(b))* \
			NUM_BASES)

/* A copy of part of the ufrag array used to restore the ufrag array
   after a proposed tree has been rejected. */

/* double saved_ufrag[num_leaves][num_tags][NUM_SYMBOLS][NUM_BASES]; */
double *saved_ufrag;
#define SAVED_UFRAG(n)       (saved_ufrag+(n)*num_tags*NUM_SYMBOLS*NUM_BASES)
#define SAVED_UFRAG3(n,tg,b) (saved_ufrag+((((n)*num_tags+(tg))* \
					    NUM_SYMBOLS)+(b))*NUM_BASES)

/* onefrag[pb] is the likelihood of the parent's base being pb given
   that the child is a gap (1.0 in all cases). */

double onefrag[NUM_BASES] = { 1.0, 1.0, 1.0, 1.0 };

/* num_uncertain[n] is the number of uncertain bases of leaf n. */

/* int num_uncertain[num_leaves]; */
int *num_uncertain;

/* Uncertain bases of a leaf are uniquely stored consecutively in a list.
   uncertain_base[n][i] is the ith uncertain base of leaf n. */

/* int uncertain_base[num_leaves][NUM_SYMBOLS]; */
int *uncertain_base;
#define UNCERTAIN_BASE(n) (uncertain_base+(n)*NUM_SYMBOLS)

/* uncertain_tag[n][i] is the tag of the ith uncertain base of leaf n. */

/* int uncertain_tag[num_leaves][NUM_SYMBOLS]; */
int *uncertain_tag;
#define UNCERTAIN_TAG(n) (uncertain_tag+(n)*NUM_SYMBOLS)

/* base[n][i] is the symbol of taxon n at site i. i is the index of the
   site in the original (uncompacted) array of sites. */

/* int base[num_leaves][num_all_sites]; */
int *base;
#define BASE(n)	(base+(n)*num_all_sites)

/* category_list is the category-list from the rc file. */

char category_list[MAX_LINE];

/* categories[] is the list of categories used, listed in order. */

int categories[MAX_CATEGORIES];

/************************************************************
 *  
 *  Borrowed function declarations
 *
 *  Several functions follow which are have been modified from sources
 *  on the Internet.  Each contains a reference to its original source.
 *
 ************************************************************/

double pnorm(double z)
/*
    pnorm(z)  returns the area to the left of under the standard normal curve

    Contributed by Alan Miller of CSIRO Division of Mathematics & Statistics,
    Clayton, Victoria.

    Distributed with Algorithm AS66 Applied Statistics (1973) vol22 no.3
    in StatLib.

    Translated to C by Bret Larget on June 2, 1998

*/
{
  const double P[7] = {220.2068679123761,
                 221.2135961699311,
                 112.0792914978709,
                 33.91286607838300,
                 6.373962203531650,
                 0.7003830644436881,
                 0.03526249659989109};
  const double Q[8] = {440.4137358247522,
                 793.8265125199484,
                 637.3336333788311,
                 296.5642487796737,
                 86.78073220294608,
                 16.06417757920695,
                 1.755667163182642,
                 0.08838834764831844};
  const double CUTOFF = 7.071;
  const double ROOT2PI = 2.506628274631001;
  double zabs = fabs(z);
  double e = SAFE_EXP(-0.5*zabs*zabs);
  double sign, p;
  
  sign = (z < 0 ? -1.0 : 1.0);
  if(zabs > 37.0)
    return (sign+1.0)/2.0;
  else if(zabs < CUTOFF)
    p = e*((((((P[6]*zabs + P[5])*zabs + P[4])*zabs + P[3])*zabs +
                 P[2])*zabs + P[1])*zabs + P[0])/
            (((((((Q[7]*zabs + Q[6])*zabs + Q[5])*zabs + Q[4])*zabs +
                Q[3])*zabs + Q[2])*zabs + Q[1])*zabs + Q[0]);
  else
    p = e/ROOT2PI/(zabs + 1.0/(zabs + 2.0/(zabs + 3.0/(zabs + 4.0/
						       (zabs + 0.650)))));
  p = (sign < 0 ? p : 1 - p);
  return p;
}

double qnorm(double p)
{
/*
  ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3,
  from StatLib.  Translated from Fortran to C by Bret Larget
  on June 2, 1998.
*/

  const double SPLIT1 = 0.425;
  const double SPLIT2 = 5.0;
  const double CONST1 = 0.180625;
  const double CONST2 = 1.6;
  
  const double A[8] = {3.3871328727963666080e0,
                       1.3314166789178437745e+2,
                       1.9715909503065514427e+3,
                       1.3731693765509461125e+4,
                       4.5921953931549871457e+4,
                       6.7265770927008700853e+4,
                       3.3430575583588128105e+4,
                       2.5090809287301226727e+3};
  const double B[8] = {1.0,
                       4.2313330701600911252e+1,
                       6.8718700749205790830e+2,
                       5.3941960214247511077e+3,
                       2.1213794301586595867e+4,
                       3.9307895800092710610e+4,
                       2.8729085735721942674e+4,
                       5.2264952788528545610e+3};
  const double C[8] = {1.42343711074968357734e0,
                       4.63033784615654529590e0,
                       5.76949722146069140550e0,
                       3.64784832476320460504e0,
                       1.27045825245236838258e0,
                       2.41780725177450611770e-1,
                       2.27238449892691845833e-2,
                       7.74545014278341407640e-4};
  const double D[8] = {1.0,
                       2.05319162663775882187e0,
                       1.67638483018380384940e0,
                       6.89767334985100004550e-1,
                       1.48103976427480074590e-1,
                       1.51986665636164571966e-2,
                       5.47593808499534494600e-4,
                       1.05075007164441684324e-9};
  const double E[8] = {6.65790464350110377720e0,
                       5.46378491116411436990e0,
                       1.78482653991729133580e0,
                       2.96560571828504891230e-1,
                       2.65321895265761230930e-2,
                       1.24266094738807843860e-3,
                       2.71155556874348757815e-5,
                       2.01033439929228813265e-7};
  const double F[8] = {1.0,
                       5.99832206555887937690e-1,
                       1.36929880922735805310e-1,
                       1.48753612908506148525e-2,
                       7.86869131145613259100e-4,
                       1.84631831751005468180e-5,
                       1.42151175831644588870e-7,
                       2.04426310338993978564e-15};

  double q, r, ans;

  q = p - 0.5;
  if(fabs(q) < SPLIT1)
  {
    r = CONST1 - q * q;
    return( q * (((((((A[7] * r + A[6]) * r + A[5]) * r + A[4]) * r + A[3])
                   * r + A[2]) * r + A[1]) * r + A[0]) /
            (((((((B[7] * r + B[6]) * r + B[5]) * r + B[4]) * r + B[3])
               * r + B[2]) * r + B[1]) * r + B[0]));
  }
  else
  {
    if(q < 0.0)
      r = p;
    else
      r = 1.0 - p;
    if(r < 0)
      fprintf(stderr,"Error: p = %lf must be strictly between 0 and 1.\n",p);
    r = sqrt(-log(r));
    if(r < SPLIT2)
    {
      r = r - CONST2;
      ans = (((((((C[7] * r + C[6]) * r + C[5]) * r + C[4]) * r + C[3])
               * r + C[2]) * r + C[1]) * r + C[0]) /
        (((((((D[7] * r + D[6]) * r + D[5]) * r + D[4]) * r + D[3])
           * r + D[2]) * r + D[1]) * r + D[0]);
    }
    else
    {
      r = r - SPLIT2;
      ans = (((((((E[7] * r + E[6]) * r + E[5]) * r + E[4]) * r + E[3])
               * r + E[2]) * r + E[1]) * r + E[0]) /
        (((((((F[7] * r + F[6]) * r + F[5]) * r + F[4]) * r + F[3])
           * r + F[2]) * r + F[1]) * r + F[0]);
    }
    ans = (q < 0.0 ? -ans : ans);
    return ans;
  }
}

double lgamma(double alpha)
{
/*
    function to return natural log of gamma function

    From StatLib, Applied Statistics Algorithm 66

    Translated from Fortran to C by Bret Larget, June 2, 1998
*/
  double a[9] = {0.9999999999995183,
                 676.5203681218835,
                 -1259.139216722289,
                 771.3234287757674,
                 -176.6150291498386,
                 12.50734324009056,
                 -0.1385710331296526,
                 0.9934937113930748e-05,
                 0.1659470187408462e-06};
  double ans,tmp;
  int i, j;
  const double LNSQRT2PI=0.9189385332046727;

  if(alpha <= 0.0)
    alpha = DBL_EPSILON;
  /*{
    fprintf(stderr,"Error: alpha = %lf <= 0 in lgamma.\n",alpha);
    exit(1);
    }*/
  ans = 0.0;
  tmp = alpha + 7.0;
  for(j=8;j>0;j--)
  {
    ans += a[j]/tmp;
    tmp -= 1.0;
  }
  ans += a[0];
  ans = log(ans) + LNSQRT2PI - (alpha+6.5) + (alpha-0.5)*log(alpha+6.5);
  return ans;
}

double incomplete_gamma(double x, double alpha)
{
  /*
    ALGORITHM AS239  APPL. STATIST. (1988) VOL. 37, NO. 3

    From StatLib.

    Translated from Fortran to C by Bret Larget on June 2, 1998.
  */
  double ans=0.0;
  const double OFLO=1.0e+37;
  const double TOL=1.0e-14;
  const double XBIG=1.0e+8;
  const double PLIMIT=1000.0;
  const double ELIMIT=-88.0;
  double pn1, pn2, pn3, pn4, pn5, pn6;
  double arg, c, rn, a, b, an;

  if(alpha <= 0.0)
  {
    fprintf(stderr,"Error: Alpha = %lf too small in incomplete_gamma\n",
	    alpha);
    exit(1);
  }
  if(x < 0.0)
  {
    fprintf(stderr,"Error: x = %lf too small in incomplete_gamma\n",x);
    exit(1);
  }
  if(x == 0)
    return 0.0;
  if(alpha > PLIMIT)
    return pnorm(3.0*sqrt(alpha)*(pow(x/alpha,1.0/3.0) + 1.0/(9.0*alpha) 
				  - 1.0));
  if(x > XBIG)
    return 1.0;
  if(x <= 1.0 || x < alpha)
  {
    arg = alpha*log(x) - x - lgamma(alpha + 1.0);
    c = 1.0;
    ans = 1.0;
    a = alpha;
    do
    {
      a += 1.0;
      c *= x/a;
      ans += c;
    }
    while(c > TOL);
    arg += log(ans);
    ans = 0.0;
    if(arg >= ELIMIT)
      ans = SAFE_EXP(arg);
  }
  else
  {
    arg = alpha*log(x) - x - lgamma(alpha);
    a = 1.0 - alpha;
    b = a + x + 1.0;
    c = 0.0;
    pn1 = 1.0;
    pn2 = x;
    pn3 = x + 1.0;
    pn4 = x*b;
    ans = pn3/pn4;
    for(;;)
    {
      a +=1.0;
      b += 2.0;
      c += 1.0;
      an = a * c;
      pn5 = b*pn3 - an*pn1;
      pn6 = b*pn4 - an*pn2;
      if(fabs(pn6) > 0.0)
      {
        rn = pn5/pn6;
        if(fabs(ans - rn) <= MINIMUM(TOL,TOL*rn))
          break;
        ans = rn;
      }
      pn1 = pn3;
      pn2 = pn4;
      pn3 = pn5;
      pn4 = pn6;
      if(fabs(pn5) >= OFLO)
      {
        pn1 /= OFLO;
        pn2 /= OFLO;
        pn3 /= OFLO;
        pn4 /= OFLO;
      }
    }
    arg += log(ans);
    ans = 1.0;
    if(arg >= ELIMIT)
      ans = 1.0 - SAFE_EXP(arg);
  }
  return ans;
}

double qchisq(double p, double df)
{
/* qchisq -- function to find the quantile from a chi-square distribution */
/* 
   Algorithm AS 91   Appl. Statist. (1975) Vol.24, P.35

   Incorporates the suggested changes in AS R85 (vol.40(1), 
   pp.233-5, 1991) which should eliminate the need for the limited
   range for p above, though these limits have not been removed
   from the routine.

   Auxiliary routines required: PPND = AS 111 (or AS 241) and
   GAMMAD = AS 239.

   Fortran code obtained from StatLib.

   Translated to C by Bret Larget on June 2, 1998
*/

  double g = lgamma(df/2.0);
  const double aa = 0.6931471806;
  const double e = 0.5e-6;
  const double pmin = 0.000002;
  const double pmax = 0.999998;
  const double C[39] = {0.0, 0.01, 0.222222, 0.32, 0.4, 1.24, 2.2,
                        4.67, 6.66, 6.73, 13.32, 60.0, 70.0,
                        84.0, 105.0, 120.0, 127.0, 140.0, 1175.0,
                        210.0, 252.0, 2264.0, 294.0, 346.0, 420.0,
                        462.0, 606.0, 672.0, 707.0, 735.0, 889.0,
                        932.0, 966.0, 1141.0, 1182.0, 1278.0, 1740.0,
                        2520.0, 5040.0};
  double a, b, c, ch, p1, p2, q, s1, s2, s3, s4, s5, s6, t, x, xx, temp;
  int i;

  if(p < pmin || p > pmax)
  {
    fprintf(stderr,"Error: p = %lf out of range in qchisq\n",p);
    exit(1);
  }
  if(df <= 0.0)
  {
    fprintf(stderr,"Error: df = %lf out of range in qchisq\n",df);
    exit(1);
  }
  xx = 0.5*df;
  c = xx - 1.0;
  if(df < C[5]*log(p))
  {
    ch = pow(p*xx*SAFE_EXP(g + xx*aa),1.0/xx);
    if(ch < e)
      return ch;
  }
  else
  {
    if(df > C[3])
    {
      x = qnorm(p);
      p1 = C[2]/df;
      temp = (x*sqrt(p1) + 1.0 - p1);
      ch = df*temp*temp*temp;
      if(ch > C[6]*df + 6.0)
        ch = -2.0*(log(1.0 - p) - c*log(0.5*ch) + g);
    }
    else
    {
      ch = C[4];
      a = log(1.0 - p);
      do
      {
        q = ch;
        p1 = 1.0 + ch * (C[7] + ch);
        p2 = ch * (C[9] + ch*(C[8] + ch));
        t = -0.5 + (C[7] + 2.0*ch)/p1 - (C[9] + ch*(C[10] + 3.0*ch))/p2;
        ch -= (1.0 - SAFE_EXP(a + g + 0.5*ch + c*aa)*p2/p1)/t;
      }
      while(fabs(q/ch - 1.0) > C[1]);
    }
  }
  for(i=0;i<MAXIT;i++)
  {
    q = ch;
    p1 = 0.5*ch;
    p2 = p - incomplete_gamma(p1,xx);
    t = p2*SAFE_EXP(xx*aa + g + p1 - c*log(ch));
    b = t/ch;
    a = 0.5*t - b*c;
    s1 = (C[19] + a*(C[17] + a*(C[14] + a*(C[13] + a*(C[12] + C[11]*a)))))
      /C[24];
    s2 = (C[24] + a*(C[29] + a*(C[32] + a*(C[33] + C[35]*a))))/C[37];
    s3 = (C[19] + a*(C[25] + a*(C[28] + C[31]*a)))/C[37];
    s4 = (C[20] + a*(C[27] + C[34]*a) + c*(C[22] + a*(C[30] + C[36]*a)))
      /C[38];
    s5 = (C[13] + C[21]*a + c*(C[18] + C[26]*a))/C[37];
    s6 = (C[15] + c*(C[23] + C[16]*c))/C[38];
    ch += t*(1.0 + 0.5*t*s1 - b*c*(s1 - b*(s2 - b*
					   (s3 - b*(s4 - b*(s5 - b*s6))))));
    if(fabs(q/ch - 1.0) > e)
      break;
  }
  return ch;
}

double qgamma(double p, double alpha, double lambda)
{
  return qchisq(p,2.0*alpha)/2.0/lambda;
}

/* Prototype */

double runif(void);

/*********************************************************************
 *  Some functions from NetLib (http://www.netlib.no/).
 *  
 *  Someday, I should clean up these functions (get rid of goto statements)
 *  and make sure they're sufficiently correct.
 *
 */

double snorm(void)
/*
**********************************************************************
                                                                      
                                                                      
     (STANDARD-)  N O R M A L  DISTRIBUTION                           
                                                                      
                                                                      
**********************************************************************
**********************************************************************
                                                                      
     FOR DETAILS SEE:                                                 
                                                                      
               AHRENS, J.H. AND DIETER, U.                            
               EXTENSIONS OF FORSYTHE'S METHOD FOR RANDOM             
               SAMPLING FROM THE NORMAL DISTRIBUTION.                 
               MATH. COMPUT., 27,124 (OCT. 1973), 927 - 937.          
                                                                      
     ALL STATEMENT NUMBERS CORRESPOND TO THE STEPS OF ALGORITHM 'FL'  
     (M=5) IN THE ABOVE PAPER     (SLIGHTLY MODIFIED IMPLEMENTATION)  
                                                                      
     Modified by Barry W. Brown, Feb 3, 1988 to use RANF instead of   
     SUNIF.  The argument IR thus goes away.                          
                                                                      
**********************************************************************
     THE DEFINITIONS OF THE CONSTANTS A(K), D(K), T(K) AND
     H(K) ARE ACCORDING TO THE ABOVEMENTIONED ARTICLE
*/
{
static double a[32] = {
    0.0,3.917609E-2,7.841241E-2,0.11777,0.1573107,0.1970991,0.2372021,
    0.2776904,
    0.3186394,0.36013,0.4022501,0.4450965,0.4887764,0.5334097,0.5791322,
    0.626099,0.6744898,0.7245144,0.7764218,0.8305109,0.8871466,0.9467818,
    1.00999,1.077516,1.150349,1.229859,1.318011,1.417797,1.534121,1.67594,
    1.862732,2.153875
};
static double d[31] = {
    0.0,0.0,0.0,0.0,0.0,0.2636843,0.2425085,0.2255674,0.2116342,0.1999243,
    0.1899108,0.1812252,0.1736014,0.1668419,0.1607967,0.1553497,0.1504094,
    0.1459026,0.14177,0.1379632,0.1344418,0.1311722,0.128126,0.1252791,
    0.1226109,0.1201036,0.1177417,0.1155119,0.1134023,0.1114027,0.1095039
};
static double t[31] = {
    7.673828E-4,2.30687E-3,3.860618E-3,5.438454E-3,7.0507E-3,8.708396E-3,
    1.042357E-2,1.220953E-2,1.408125E-2,1.605579E-2,1.81529E-2,2.039573E-2,
    2.281177E-2,2.543407E-2,2.830296E-2,3.146822E-2,3.499233E-2,3.895483E-2,
    4.345878E-2,4.864035E-2,5.468334E-2,6.184222E-2,7.047983E-2,8.113195E-2,
    9.462444E-2,0.1123001,0.136498,0.1716886,0.2276241,0.330498,0.5847031
};
static double h[31] = {
    3.920617E-2,3.932705E-2,3.951E-2,3.975703E-2,4.007093E-2,4.045533E-2,
    4.091481E-2,4.145507E-2,4.208311E-2,4.280748E-2,4.363863E-2,4.458932E-2,
    4.567523E-2,4.691571E-2,4.833487E-2,4.996298E-2,5.183859E-2,5.401138E-2,
    5.654656E-2,5.95313E-2,6.308489E-2,6.737503E-2,7.264544E-2,7.926471E-2,
    8.781922E-2,9.930398E-2,0.11556,0.1404344,0.1836142,0.2790016,0.7010474
};
static int32 i;
static double snorm,u,s,ustar,aa,w,y,tt;
    u = runif();
    s = 0.0;
    if(u > 0.5) s = 1.0;
    u += (u-s);
    u = 32.0*u;
    i = (int32) (u);
    if(i == 32) i = 31;
    if(i == 0) goto S100;
/*
                                START CENTER
*/
    ustar = u-(double)i;
    aa = *(a+i-1);
S40:
    if(ustar <= *(t+i-1)) goto S60;
    w = (ustar-*(t+i-1))**(h+i-1);
S50:
/*
                                EXIT   (BOTH CASES)
*/
    y = aa+w;
    snorm = y;
    if(s == 1.0) snorm = -y;
    return snorm;
S60:
/*
                                CENTER CONTINUED
*/
    u = runif();
    w = u*(*(a+i)-aa);
    tt = (0.5*w+aa)*w;
    goto S80;
S70:
    tt = u;
    ustar = runif();
S80:
    if(ustar > tt) goto S50;
    u = runif();
    if(ustar >= u) goto S70;
    ustar = runif();
    goto S40;
S100:
/*
                                START TAIL
*/
    i = 6;
    aa = *(a+31);
    goto S120;
S110:
    aa += *(d+i-1);
    i += 1;
S120:
    u += u;
    if(u < 1.0) goto S110;
    u -= 1.0;
S140:
    w = u**(d+i-1);
    tt = (0.5*w+aa)*w;
    goto S160;
S150:
    tt = u;
S160:
    ustar = runif();
    if(ustar > tt) goto S50;
    u = runif();
    if(ustar >= u) goto S150;
    u = runif();
    goto S140;
}

double sexpo(void)
/*
**********************************************************************
                                                                      
                                                                      
     (STANDARD-)  E X P O N E N T I A L   DISTRIBUTION                
                                                                      
                                                                      
**********************************************************************
**********************************************************************
                                                                      
     FOR DETAILS SEE:                                                 
                                                                      
               AHRENS, J.H. AND DIETER, U.                            
               COMPUTER METHODS FOR SAMPLING FROM THE                 
               EXPONENTIAL AND NORMAL DISTRIBUTIONS.                  
               COMM. ACM, 15,10 (OCT. 1972), 873 - 882.               
                                                                      
     ALL STATEMENT NUMBERS CORRESPOND TO THE STEPS OF ALGORITHM       
     'SA' IN THE ABOVE PAPER (SLIGHTLY MODIFIED IMPLEMENTATION)       
                                                                      
     Modified by Barry W. Brown, Feb 3, 1988 to use RANF instead of   
     SUNIF.  The argument IR thus goes away.                          
                                                                      
**********************************************************************
     Q(N) = SUM(ALOG(2.0)**K/K!)    K=1,..,N ,      THE HIGHEST N
     (HERE 8) IS DETERMINED BY Q(N)=1.0 WITHIN STANDARD PRECISION
*/
{
static double q[8] = {
    0.6931472,0.9333737,0.9888778,0.9984959,0.9998293,0.9999833,0.9999986,
    1.0
};
static int32 i;
static double sexpo,a,u,ustar,umin;
static double *q1 = q;
    a = 0.0;
    u = runif();
    goto S30;
S20:
    a += *q1;
S30:
    u += u;
    if(u <= 1.0) goto S20;
    u -= 1.0;
    if(u > *q1) goto S60;
    sexpo = a+u;
    return sexpo;
S60:
    i = 1;
    ustar = runif();
    umin = ustar;
S70:
    ustar = runif();
    if(ustar < umin) umin = ustar;
    i += 1;
    if(u > *(q+i-1)) goto S70;
    sexpo = a+umin**q1;
    return sexpo;
}

double fsign( double num, double sign )
/* Transfers sign of argument sign to argument num */
{
if ( ( sign>0.0f && num<0.0f ) || ( sign<0.0f && num>0.0f ) )
    return -num;
else return num;
}

double sgamma(double a)
/*
**********************************************************************
                                                                      
                                                                      
     (STANDARD-)  G A M M A  DISTRIBUTION                             
                                                                      
                                                                      
**********************************************************************
**********************************************************************
                                                                      
               PARAMETER  A >= 1.0  !                                 
                                                                      
**********************************************************************
                                                                      
     FOR DETAILS SEE:                                                 
                                                                      
               AHRENS, J.H. AND DIETER, U.                            
               GENERATING GAMMA VARIATES BY A                         
               MODIFIED REJECTION TECHNIQUE.                          
               COMM. ACM, 25,1 (JAN. 1982), 47 - 54.                  
                                                                      
     STEP NUMBERS CORRESPOND TO ALGORITHM 'GD' IN THE ABOVE PAPER     
                                 (STRAIGHTFORWARD IMPLEMENTATION)     
                                                                      
     Modified by Barry W. Brown, Feb 3, 1988 to use RANF instead of   
     SUNIF.  The argument IR thus goes away.                          
                                                                      
**********************************************************************
                                                                      
               PARAMETER  0.0 < A < 1.0  !                            
                                                                      
**********************************************************************
                                                                      
     FOR DETAILS SEE:                                                 
                                                                      
               AHRENS, J.H. AND DIETER, U.                            
               COMPUTER METHODS FOR SAMPLING FROM GAMMA,              
               BETA, POISSON AND BINOMIAL DISTRIBUTIONS.              
               COMPUTING, 12 (1974), 223 - 246.                       
                                                                      
     (ADAPTED IMPLEMENTATION OF ALGORITHM 'GS' IN THE ABOVE PAPER)    
                                                                      
**********************************************************************
     INPUT: A =PARAMETER (MEAN) OF THE STANDARD GAMMA DISTRIBUTION
     OUTPUT: SGAMMA = SAMPLE FROM THE GAMMA-(A)-DISTRIBUTION
     COEFFICIENTS Q(K) - FOR Q0 = SUM(Q(K)*A**(-K))
     COEFFICIENTS A(K) - FOR Q = Q0+(T*T/2)*SUM(A(K)*V**K)
     COEFFICIENTS E(K) - FOR EXP(Q)-1 = SUM(E(K)*Q**K)
     PREVIOUS A PRE-SET TO ZERO - AA IS A', AAA IS A"
     SQRT32 IS THE SQUAREROOT OF 32 = 5.656854249492380
*/
{
extern double fsign( double num, double sign );
static double q1 = 4.166669E-2;
static double q2 = 2.083148E-2;
static double q3 = 8.01191E-3;
static double q4 = 1.44121E-3;
static double q5 = -7.388E-5;
static double q6 = 2.4511E-4;
static double q7 = 2.424E-4;
static double a1 = 0.3333333;
static double a2 = -0.250003;
static double a3 = 0.2000062;
static double a4 = -0.1662921;
static double a5 = 0.1423657;
static double a6 = -0.1367177;
static double a7 = 0.1233795;
static double e1 = 1.0;
static double e2 = 0.4999897;
static double e3 = 0.166829;
static double e4 = 4.07753E-2;
static double e5 = 1.0293E-2;
static double aa = 0.0;
static double aaa = 0.0;
static double sqrt32 = 5.656854;
static double sgamma,s2,s,d,t,x,u,r,q0,b,si,c,v,q,e,w,p;
    if(a == aa) goto S10;
    if(a < 1.0) goto S120;
/*
     STEP  1:  RECALCULATIONS OF S2,S,D IF A HAS CHANGED
*/
    aa = a;
    s2 = a-0.5;
    s = sqrt(s2);
    d = sqrt32-12.0*s;
S10:
/*
     STEP  2:  T=STANDARD NORMAL DEVIATE,
               X=(S,1/2)-NORMAL DEVIATE.
               IMMEDIATE ACCEPTANCE (I)
*/
    t = snorm();
    x = s+0.5*t;
    sgamma = x*x;
    if(t >= 0.0) return sgamma;
/*
     STEP  3:  U= 0,1 -UNIFORM SAMPLE. SQUEEZE ACCEPTANCE (S)
*/
    u = runif();
    if(d*u <= t*t*t) return sgamma;
/*
     STEP  4:  RECALCULATIONS OF Q0,B,SI,C IF NECESSARY
*/
    if(a == aaa) goto S40;
    aaa = a;
    r = 1.0/ a;
    q0 = ((((((q7*r+q6)*r+q5)*r+q4)*r+q3)*r+q2)*r+q1)*r;
/*
               APPROXIMATION DEPENDING ON SIZE OF PARAMETER A
               THE CONSTANTS IN THE EXPRESSIONS FOR B, SI AND
               C WERE ESTABLISHED BY NUMERICAL EXPERIMENTS
*/
    if(a <= 3.686) goto S30;
    if(a <= 13.022) goto S20;
/*
               CASE 3:  A .GT. 13.022
*/
    b = 1.77;
    si = 0.75;
    c = 0.1515/s;
    goto S40;
S20:
/*
               CASE 2:  3.686 .LT. A .LE. 13.022
*/
    b = 1.654+7.6E-3*s2;
    si = 1.68/s+0.275;
    c = 6.2E-2/s+2.4E-2;
    goto S40;
S30:
/*
               CASE 1:  A .LE. 3.686
*/
    b = 0.463+s+0.178*s2;
    si = 1.235;
    c = 0.195/s-7.9E-2+1.6E-1*s;
S40:
/*
     STEP  5:  NO QUOTIENT TEST IF X NOT POSITIVE
*/
    if(x <= 0.0) goto S70;
/*
     STEP  6:  CALCULATION OF V AND QUOTIENT Q
*/
    v = t/(s+s);
    if(fabs(v) <= 0.25) goto S50;
    q = q0-s*t+0.25*t*t+(s2+s2)*log(1.0+v);
    goto S60;
S50:
    q = q0+0.5*t*t*((((((a7*v+a6)*v+a5)*v+a4)*v+a3)*v+a2)*v+a1)*v;
S60:
/*
     STEP  7:  QUOTIENT ACCEPTANCE (Q)
*/
    if(log(1.0-u) <= q) return sgamma;
S70:
/*
     STEP  8:  E=STANDARD EXPONENTIAL DEVIATE
               U= 0,1 -UNIFORM DEVIATE
               T=(B,SI)-DOUBLE EXPONENTIAL (LAPLACE) SAMPLE
*/
    e = sexpo();
    u = runif();
    u += (u-1.0);
    t = b+fsign(si*e,u);
/*
     STEP  9:  REJECTION IF T .LT. TAU(1) = -.71874483771719
*/
    if(t < -0.7187449) goto S70;
/*
     STEP 10:  CALCULATION OF V AND QUOTIENT Q
*/
    v = t/(s+s);
    if(fabs(v) <= 0.25) goto S80;
    q = q0-s*t+0.25*t*t+(s2+s2)*log(1.0+v);
    goto S90;
S80:
    q = q0+0.5*t*t*((((((a7*v+a6)*v+a5)*v+a4)*v+a3)*v+a2)*v+a1)*v;
S90:
/*
     STEP 11:  HAT ACCEPTANCE (H) (IF Q NOT POSITIVE GO TO STEP 8)
*/
    if(q <= 0.0) goto S70;
    if(q <= 0.5) goto S100;
    w = SAFE_EXP(q)-1.0;
    goto S110;
S100:
    w = ((((e5*q+e4)*q+e3)*q+e2)*q+e1)*q;
S110:
/*
               IF T IS REJECTED, SAMPLE AGAIN AT STEP 8
*/
    if(c*fabs(u) > w*SAFE_EXP(e-0.5*t*t)) goto S70;
    x = s+0.5*t;
    sgamma = x*x;
    return sgamma;
S120:
/*
     ALTERNATE METHOD FOR PARAMETERS A BELOW 1  (.3678794=EXP(-1.))
*/
    aa = 0.0;
    b = 1.0+0.3678794*a;
S130:
    p = b*runif();
    if(p >= 1.0) goto S140;
    sgamma = SAFE_EXP(log(p)/ a);
    if(sexpo() < sgamma) goto S130;
    return sgamma;
S140:
    sgamma = -log((b-p)/ a);
    if(sexpo() < (1.0-a)*log(sgamma)) goto S130;
    return sgamma;
}

double gengam(double a,double r)
/*
**********************************************************************
     double gengam(double a,double r)
           GENerates random deviates from GAMma distribution
                              Function
     Generates random deviates from the gamma distribution whose
     density is
          (A**R)/Gamma(R) * X**(R-1) * Exp(-A*X)
                              Arguments
     a --> Location parameter of Gamma distribution
     r --> Shape parameter of Gamma distribution
                              Method
     Renames SGAMMA from TOMS as slightly modified by BWB to use RANF
     instead of SUNIF.
     For details see:
               (Case R >= 1.0)
               Ahrens, J.H. and Dieter, U.
               Generating Gamma Variates by a
               Modified Rejection Technique.
               Comm. ACM, 25,1 (Jan. 1982), 47 - 54.
     Algorithm GD
               (Case 0.0 <= R <= 1.0)
               Ahrens, J.H. and Dieter, U.
               Computer Methods for Sampling from Gamma,
               Beta, Poisson and Binomial Distributions.
               Computing, 12 (1974), 223-246/
     Adapted algorithm GS.
**********************************************************************
*/
{
static double gengam;

    gengam = sgamma(r);
    gengam /= a;
    return gengam;
}

double rgamma(double alpha,double lambda)
{
  static double x;
  
  x = gengam(lambda,alpha);
  return x;
}

void rdirich(const double alpha[],double y[],int k)
/*
 * alpha is an array of k positive numbers.
 * return y distributed according to the Dirichlet(alpha) distribution
 */
{
  double sum=0.0;
  int i;
  
  for(i=0;i<k;i++)
  {
    y[i] = rgamma(alpha[i],1.0);
    sum += y[i];
  }
  for(i=0;i<k;i++)
    y[i] /= sum;
}

double hastings(const double c,const double x[],const double y[],int k)
/* c is the constant where y ~ Dirichlet(cx)
   x is the current state
   y is the proposed state
   k is the length of each array
   return q(y,x) / q(x,y)
*/
{
  double lhr = 0.0,sumx=0.0,sumy=0.0,cx,cy;
  int i;
  
  for(i=0;i<k;i++)
  {
    sumx += x[i];
    sumy += y[i];
  }
  lhr += lgamma(c*sumy) - lgamma(c*sumx);
  for(i=0;i<k;i++)
  {
    cx = c*x[i];
    cy = c*y[i];
    lhr += (cy-1)*log(x[i]) - (cx-1)*log(y[i]) - lgamma(cy) + lgamma(cx);
  }
  return SAFE_EXP(lhr);
}

/**************************************************************************

  Function Declarations

**************************************************************************/

/*****

  I/O Functions

*****/

/***********************************************************************

  Parameter Information

************************************************************************/

/* Algorithms used to deform the tree. */

char *algorithms[] = { "global", "local" };
typedef enum { GLOBAL, LOCAL } Algorithms;
#define NUM_ALGORITHMS 	(sizeof(algorithms)/sizeof(algorithms[0]))

/* Probability models. */

char *models[] = { "HKY85", "F84", "TN93" };
typedef enum { HKY85, F84, TN93 } Models;
#define NUM_MODELS 	(sizeof(models)/sizeof(models[0]))

/* Site models. */

/* Types of allowable parameters in the rc file. */

char *param_types[] = { "int", "algorithm", "double", "double[]", "string",
		        "boolean", "model" };
typedef enum { INT, ALGORITHM, DOUBLE, DOUBLEARRAY, STRING, BOOLEAN, MODEL }
  Param_types;
#define NUM_PARAM_TYPES	(sizeof(param_types)/sizeof(param_types[0]))

/* Names, types, and default values of parameters in the rc file.
   The parameters will be processed in the order that they are
   listed in this array. */

char *input_params[][3] = {
  "seed",			"int",	       "194024933", 
  "burn",			"int",	       "1000", 
  "burn-algorithm",		"algorithm",   "global", 
  "main-algorithm",		"algorithm",   "local", 
  "subcycles",			"int",	       "1", 
  "cycles",			"int",	       "6000", 
  "sample-interval",		"int",	       "200", 
  "parameter-update-interval",	"int",	       "0",
  "update-kappa",               "boolean",     "true",
  "update-theta",               "boolean",     "true",
  "update-pi",                  "boolean",     "true",
  "tune-interval",              "int",         "200",
  "window-interval",		"int",         "200", 
  "molecular-clock",		"boolean",     "true",
  "likelihood-model",		"model",       "HKY85",
  "category-list",		"string",      "1*",
  "single-kappa",		"boolean",     "false",
  "initial-kappa",		"double[]",    "2.0,2.0,2.0",
  "initial-theta",		"double[]",    "1.0,1.0,1.0",
  "initial-ttp",		"double[]",    "1.0,1.0,1.0",
  "initial-gamma",		"double[]",    "1.0,1.0,1.0",
  "data-file",			"string",      "infile",
  "outgroup",                   "int",         "1",
  "initial-delta",		"double",      "0.01", 
  "initial-kappa-delta",	"double",      "0.5", 
  "theta-const",                "double",      "2000.0",
  "pi-const",                   "double",      "2000.0",
  "max-initial-tree-height",	"double",      ".1", 
  "tree-file",			"string",      "random", 
  "file-root",			"string",      "run1", 
  "newick-format",		"boolean",     "true"
};

typedef enum { SEED, BURN, BURN_ALGORITHM, MAIN_ALGORITHM, SUBCYCLES, 
	       CYCLES, SAMPLE_INTERVAL, PARAMETER_UPDATE_INTERVAL,
               UPDATE_KAPPA, UPDATE_THETA, UPDATE_PI, TUNE_INTERVAL,
	       WINDOW_INTERVAL, MOLECULAR_CLOCK, LIKELIHOOD_MODEL,
	       CATEGORY_LIST, SINGLE_KAPPA, INITIAL_KAPPA, 
	       INITIAL_THETA, INITIAL_TTP, INITIAL_GAMMA,
               DATA_FILE, OUTGROUP, INITIAL_DELTA, INITIAL_KAPPA_DELTA,
               THETA_CONST, PI_CONST, MAX_INITIAL_TREE_HEIGHT, TREE_FILE, 
	       FILE_ROOT, NEWICK_FORMAT } Input_params;
#define NUM_INPUT_PARAMS (sizeof(input_params)/sizeof(input_params[0]))

int find_input_param(char *s)
{
  /* Returns the number of a parameter given the name s of a valid parameter.
     Otherwise returns the total number of parameters. */
  int i;

  for(i=0;i<NUM_INPUT_PARAMS;i++)
    if(strcmp(s,input_params[i][0])==0)
      return i;
  return i;
}

int string_member(char *s, int n, char *sl[])
{
  /* Returns the index of string s in the list of strings sl (of size n),
     if s is a member of that list. Returns n otherwise. */
  int i;

  for(i=0;i<n;i++)
    if(strcmp(s,sl[i])==0)
      return i;
  return i;
}

int string_prefix(char *s, int n, char *sl[])
{
  /* Returns the index of a string in the list of strings sl (of size n),
     of which s is a prefix, if such a string exists. Returns n otherwise. */
  int i;

  for(i=0;i<n;i++)
    if(strncmp(s,sl[i],strlen(sl[i]))==0)
      return i;
  return i;
}

char *strip_right_spaces(char *s, char *end)
{
  /* Strips the right spaces from string s where end points to the 
     character after the string. */
  char *pos;

  for(pos=end-1;pos>=s&&isspace(*pos);pos--)
    ;
  *(pos+1) = '\0';
  return s;
}

void print_bad_param(char *param, int lineno)
{
  fprintf(stderr,"'%s' (line %d).\n",
	  strip_right_spaces(param,param+strlen(param)),lineno);
}

void add_category(int n, int *categories)
{
  int *i,*j;

  for(i=categories;i<categories+num_tags && *i<=n;i++)
    if(*i == n)
      return;
  for(j=categories+num_tags;j>i;j--)
    *j = *(j-1);
  *j = n;
  num_tags++;
}

int get_categories(char **str, int *categories, char *param, int lineno) 
{
  /* Parses a string representing a list of categories up to, but
     not including the final asterisk (if it exists).
     Checks for syntax errors and sets up the category array.
     str is the unscanned part of the category list string.
     categories is an array of categories (in order) seen so far.
     num_tags is the number of distinct categories seen so far.
     Returns whether or not a category was read in from the string. */

  int n;

  while(isspace(**str))
    (*str)++;
  if(**str == '(') {
    (*str)++;
    if(!get_categories(str,categories,param,lineno)) {
      fprintf(stderr,"Error: Empty group in category list ");
      print_bad_param(param,lineno);
      exit(1);
    }
    while(isspace(**str))
      (*str)++;
    if(**str == '*') {
      fprintf(stderr,"Error: Asterisk must be the last symbol in the "
	      "category list ");
      print_bad_param(param,lineno);
      exit(1);
    }
    if(**str != ')') {
      fprintf(stderr,"Error: Mismatched parentheses in category list ");
      print_bad_param(param,lineno);
      exit(1);
    }
    (*str)++;
  }
  else if(isdigit(**str)) {
    n = **str - '0';
    for((*str)++;isdigit(**str);(*str)++)
      n = 10 * n + **str - '0';
    if(n == 0) {
      fprintf(stderr,"Error: Category must be a positive number in "
	      "category list ");
      print_bad_param(param,lineno);
      exit(1);
    }
    else if(n > MAX_CATEGORIES) {
      fprintf(stderr,"Error: Category must be less than %d in "
	      "category list ",
	      MAX_CATEGORIES);
      print_bad_param(param,lineno);
      exit(1);
    }
    add_category(n,categories);
  }
  else 
    return 0;
  while(isspace(**str))
    (*str)++;
  if(**str == '^') {
    (*str)++;
    while(isspace(**str))
      (*str)++;
    if(!isdigit(**str)) {
      fprintf(stderr,"Error: Repeat count must be an integer in the "
	      "category list ");
      print_bad_param(param,lineno);
      exit(1);
    }
    n = **str - '0';
    for((*str)++;isdigit(**str);(*str)++)
      n = 10 * n + **str - '0';
    if(n == 0) {
      fprintf(stderr,"Error: Repeat count may not be zero in the "
	      "category list ");
      print_bad_param(param,lineno);
      exit(1);
    }
    while(isspace(**str))
      (*str)++;
  }
  if(**str == ',') {
    (*str)++;
    get_categories(str,categories,param,lineno);
  }
  return 1;
}

int int_in_range(int x, int positive, int i, int *seen)
{
  if(positive ? x > 0 : x >= 0)
    return 1;
  fprintf(stderr,"Error: %s (%d) must be %s (line %d).\n",
	  input_params[i][0],x,
	  positive?"positive":"non-negative",seen[i]);
  return 0;
}

int double_in_range(double x, int positive, int i, int *seen)
{
  if(positive ? x > 0 : x >= 0)
    return 1;
  fprintf(stderr,"Error: %s (%lf) must be %s (line %d).\n",
	  input_params[i][0],x,
	  positive?"positive":"non-negative",seen[i]);
  return 0;
}

void parse_param_list(int numvals, double *input, double **valarray, 
		      int *given, int i, int *seen, double *ardoubleval,
		      int single_kap, int positive)
{ 
  int j,k,warned = 0;

  if(!(*valarray=(double *)malloc(sizeof(double)*num_tags)))
    out_of_space(input_params[i][0]);
  if(numvals >= num_categories) {
    if(numvals > num_categories)
      fprintf(stderr,"Warning: Too many values in %s (line %d). "
	      "Extra values ignored.\n",
	      input_params[i][0],seen[i]);
    for(j=0;j<num_categories;j++)
      input[j] = ardoubleval[j];
    for(k=0;k<num_tags;k++) {
      j = categories[k];
      if(!given[j-1]) {
	fprintf(stderr,"Error: Value must be specified for %s of "
		"category %d (line %d).\n",
		input_params[i][0],j,seen[i]);
	exit(1);
      }
      else if(!double_in_range(ardoubleval[j-1],positive,i,seen)) 
	exit(1);
      else if(single_kap && k > 0 && 
	      fabs(ardoubleval[j-1]-(*valarray)[0])>1e-10) {
	if(!warned) {
	  fprintf(stderr,"Warning: %s values differ in single kappa"
		  " mode. Only first value will be used (line %d).\n",
		  input_params[i][0], seen[i]);
	  warned = 1;
	}
	(*valarray)[k] = (*valarray)[0];
      }
      else
	(*valarray)[k] = ardoubleval[j-1];
    }
  }
  else if(numvals == 1 && single_kap) {
    if(!double_in_range(ardoubleval[0],positive,i,seen)) 
      exit(1);
    for(k=0;k<num_tags;k++)
      (*valarray)[k] = ardoubleval[0];
    for(j=0;j<num_categories;j++)
      input[j] = ardoubleval[0];
  }
  else {
    fprintf(stderr,"Error: Too few values in %s (line %d).\n",
	    input_params[i][0],seen[i]);
    exit(1);
  }
}

void read_input_params(FILE *f)
{
  /* Reads in and processes the parameters from file f. Sets
     the corresponding global variables. A parameter not listed in the
     file is set to its default value. Then the parameters are processed
     in the order given by the parameter array. */

  char *pos,*eqpos,*pos1,*pos2,(*val)[MAX_LINE],buf[MAX_LINE];
  char valx[MAX_LINE];
  int i,j,k,n,*seen,okay;
  int intval;
  double doubleval, ardoubleval[MAX_CATEGORIES];
  char *stringval;
  int element_given[MAX_CATEGORIES];

  /* Set the value of each parameter to its default value and mark
     each as unseen. */
  if(!(val=(char (*)[MAX_LINE])malloc(sizeof(char[MAX_LINE])
				      * NUM_INPUT_PARAMS)))
    out_of_space("val");
  if(!(seen=(int *)malloc(sizeof(int)*NUM_INPUT_PARAMS)))
    out_of_space("seen");
  for(i=0;i<NUM_INPUT_PARAMS;i++) {
    strcpy(val[i],input_params[i][2]);
    seen[i] = 0;
  }
  /* Read in each line of the file. Parse the parameter name and its
     value. */
  i = 0;
  while(fgets(buf,MAX_LINE,f)) {
    i++;
    if(pos=strchr(buf,'#'))
      *pos = '\0';
    pos = buf+strspn(buf," \t\n");
    if(*pos!='\0')
      if(eqpos=strchr(pos,'=')) {
	strip_right_spaces(pos,eqpos);
	if((n=find_input_param(pos)) < NUM_INPUT_PARAMS) {
	  if(seen[n])
	    fprintf(stderr,"Warning: parameter (%s) multiply defined "
		    "(line %d and line %d).\n",
		    seen[n],i);
	  eqpos = eqpos+1+strspn(eqpos+1," \t\n");
	  seen[n] = i;
	  strcpy(val[n],eqpos);
	}
	else {
	  fprintf(stderr,"Error: Unknown parameter name (%s) (line %d).\n",
		  pos,i);
	  exit(1);
	}
      }
      else
	fprintf(stderr,"Error: Illegal syntax. Missing '=' (line %d).\n",i);
  }
  /* Process the parameters in order, checking the types and numbers 
     of the values. */
  for(i=0;i<NUM_INPUT_PARAMS;i++) {
    switch (string_member(input_params[i][1],NUM_PARAM_TYPES,param_types)) {
    case INT: okay = (sscanf(val[i],"%d",&intval)==1); break;
    case ALGORITHM: 
      strip_right_spaces(val[i],val[i]+strlen(val[i]));
      okay = ((intval=string_member(val[i],NUM_ALGORITHMS,algorithms))
	      < NUM_ALGORITHMS);
      break;
    case MODEL: 
      strip_right_spaces(val[i],val[i]+strlen(val[i]));
      okay = ((intval=string_member(val[i],NUM_MODELS,models))
	      < NUM_MODELS);
      break;
    case DOUBLE: okay = (sscanf(val[i],"%lf",&doubleval)==1); break;
    case STRING: 
      strip_right_spaces(val[i],val[i]+strlen(val[i]));
      stringval = val[i];
      okay = 1;
      break;
    case DOUBLEARRAY:
      okay = 1;
      strcpy(valx,val[i]);
      for(pos1=valx,intval=0;*pos1;pos1=pos2,intval++) {
	for(pos2=pos1;*pos2;pos2++)
	  if(*pos2 == ',') {
	    *pos2++ = '\0';
	    break;
	  }
	if(*pos1 == '\0') {
	  element_given[intval] = 0;
	  ardoubleval[intval] = 0;
	}
	else if(sscanf(pos1,"%lf",&(ardoubleval[intval]))==1) {
	  element_given[intval] = 1;
	}
	else {
	  ardoubleval[intval] = 0;
	  okay = 0;
	}
      }
      break;
    case BOOLEAN:
      strcpy(val[i],val[i]+strspn(buf," \t"));
      strip_right_spaces(val[i],val[i]+strlen(val[i]));
      if(strcmp(val[i],"true")==0) {
	intval = 1;
	okay = 1;
      }
      else if(strcmp(val[i],"false")==0) {
	intval = 0;
	okay = 1;
      }
      else 
	okay = 0;
      break;
    default: fprintf(stderr,"Internal error: Unknown parameter type (%s).\n",
		     input_params[i][1]);
      exit(1);
    }
    if(!okay) {
      fprintf(stderr,"Error: Could not understand parameter value '%s' "
	      "(line %d).\n",
	      strip_right_spaces(val[i],val[i]+strlen(val[i])),seen[i]);
      exit(1);
    }
    /* Process the parameter values in order, setting the appropriate
       global variables. */
    switch (i) {
    case SEED: seed = intval % (1UL << 31); break;
    case BURN: 
      if(!int_in_range(intval,0,i,seen))
	exit(1);
      burn = intval; 
      break;
    case BURN_ALGORITHM: burn_algorithm = intval; break;
    case MAIN_ALGORITHM: main_algorithm = intval; break;
    case SUBCYCLES: 
      if(!int_in_range(intval,1,i,seen))
	exit(1);
      num_subcycles = intval; 
      break;
    case CYCLES: 
      if(!int_in_range(intval,0,i,seen))
	exit(1);
      num_cycles = intval;
      break;
    case SAMPLE_INTERVAL: 
      if(!int_in_range(intval,1,i,seen))
	exit(1);
      sample_interval = intval;
      break;
    case PARAMETER_UPDATE_INTERVAL: 
      if(!int_in_range(intval,0,i,seen))
	exit(1);
      param_update_interval = intval; 
      break;
    case UPDATE_KAPPA: update_kappa = intval; break;
    case UPDATE_THETA: update_theta = intval; break;
    case UPDATE_PI: update_pi = intval; break;
    case TUNE_INTERVAL: 
      if(!int_in_range(intval,1,i,seen))
	exit(1);
      tune_interval = intval; 
      break;
    case WINDOW_INTERVAL: 
      if(!int_in_range(intval,1,i,seen))
	exit(1);
      window_interval = intval; 
      break;
    case MOLECULAR_CLOCK: mclock = intval; break;
    case LIKELIHOOD_MODEL: model = intval; break; 
    case CATEGORY_LIST:
      strcpy(category_list,stringval);
      num_tags = 0;
      if(!get_categories(&stringval,categories,val[i],seen[i])) {
	fprintf(stderr,"Error: Empty category list ");
	print_bad_param(val[i],seen[i]);
	exit(1);
      }
      if(*stringval == '*') {
	(stringval)++;
	while(isspace(*stringval))
	  (stringval)++;
      }
      if(*stringval != '\0') {
	fprintf(stderr,"Error: Category list ends badly ");
	print_bad_param(val[i],seen[i]);
	exit(1);
      }
      num_categories = categories[num_tags-1];
      break;
    case SINGLE_KAPPA: single_kappa = intval; break;
    case INITIAL_KAPPA: 
      parse_param_list(intval,input_kappa,&initial_kappa,element_given,
		      i,seen,ardoubleval,single_kappa,1);
      break;
    case INITIAL_THETA:
      parse_param_list(intval,input_theta,&initial_theta,element_given,
		       i,seen,ardoubleval,0,0);
      break;
    case INITIAL_TTP:
      if(model==TN93) 
	parse_param_list(intval,input_ttp,&initial_ttp,element_given,
			i,seen,ardoubleval,single_kappa,1);
      else if(seen[INITIAL_TTP])
	fprintf(stderr,"Warning: parameter (%s) not used in model %s "
		"(line %d).\n",
		input_params[INITIAL_TTP][0],models[model],
		seen[INITIAL_TTP]);
      break;
    case INITIAL_GAMMA: 
      if(model==TN93) 
	parse_param_list(intval,input_gamma,&initial_gamma,element_given,
			i,seen,ardoubleval,single_kappa,1);
      else if(seen[INITIAL_GAMMA])
	fprintf(stderr,"Warning: parameter (%s) not used in model %s "
		"(line %d).\n",
		input_params[INITIAL_GAMMA][0],models[model],
		seen[INITIAL_GAMMA]);
      break;
    case DATA_FILE: strcpy(data_file,stringval); break;
    case OUTGROUP: 
      if(!int_in_range(intval,1,i,seen))
	exit(1);
      outgroup = intval; 
      break;
    case INITIAL_DELTA: 
      if(!double_in_range(doubleval,0,i,seen))
	exit(1);
      valley_win  = doubleval; 
      break;
    case INITIAL_KAPPA_DELTA: 
      if(!double_in_range(doubleval,0,i,seen))
	exit(1);
      kappa_win = doubleval; 
      break;
    case THETA_CONST:
      if(!double_in_range(doubleval,0,i,seen))
	exit(1);
      theta_const  = doubleval; 
      break;
    case PI_CONST:
      if(!double_in_range(doubleval,0,i,seen))
	exit(1);
      pi_const  = doubleval; 
      break;
    case MAX_INITIAL_TREE_HEIGHT: 
      if(!double_in_range(doubleval,1,i,seen))
	exit(1);
      max_depth = doubleval; 
      break;
    case TREE_FILE: strcpy(tree_file,stringval); break;
    case FILE_ROOT: strcpy(file_root,stringval); break;
    case NEWICK_FORMAT: newick_format = intval; break;
    default: fprintf(stderr,"Internal error: Unknown parameter number %d.\n",
		     i);
    }
  }
  free(val);
  free(seen);
}

void print_input_params(FILE *f)
{
  /* Prints out the parameters and their values. */
  int i;
  fprintf(f,"%s=%d\n",input_params[SEED][0],seed);
  fprintf(f,"%s=%d\n",input_params[BURN][0],burn);
  fprintf(f,"%s=%s\n",input_params[BURN_ALGORITHM][0],
	  algorithms[burn_algorithm]);
  fprintf(f,"%s=%s\n",input_params[MAIN_ALGORITHM][0],
	  algorithms[main_algorithm]);
  fprintf(f,"%s=%d\n",input_params[SUBCYCLES][0],num_subcycles);
  fprintf(f,"%s=%d\n",input_params[CYCLES][0],num_cycles);
  fprintf(f,"%s=%d\n",input_params[SAMPLE_INTERVAL][0],sample_interval);
  fprintf(f,"%s=%d\n",input_params[PARAMETER_UPDATE_INTERVAL][0],
	  param_update_interval);
  fprintf(f,"%s=%s\n",input_params[UPDATE_KAPPA][0],
	  update_kappa?"true":"false");
  fprintf(f,"%s=%s\n",input_params[UPDATE_THETA][0],
	  update_theta?"true":"false");
  fprintf(f,"%s=%s\n",input_params[UPDATE_PI][0],
	  update_pi?"true":"false");
  fprintf(f,"%s=%d\n",input_params[TUNE_INTERVAL][0],tune_interval);
  fprintf(f,"%s=%d\n",input_params[WINDOW_INTERVAL][0],window_interval);
  fprintf(f,"%s=%s\n",input_params[MOLECULAR_CLOCK][0],
	  mclock?"true":"false");
  fprintf(f,"%s=%s\n",input_params[LIKELIHOOD_MODEL][0],models[model]);
  fprintf(f,"%s=%s\n",input_params[CATEGORY_LIST][0],category_list);
  fprintf(f,"%s=%s\n",input_params[SINGLE_KAPPA][0],
	  single_kappa?"true":"false");
  fprintf(f,"%s=%lf",input_params[INITIAL_KAPPA][0],input_kappa[0]);
  if(!single_kappa)
    for(i=1;i<num_categories;i++)
      fprintf(f,",%lf",input_kappa[i]);
  fprintf(f,"\n");
  fprintf(f,"%s=%lf",input_params[INITIAL_THETA][0],input_theta[0]);
  for(i=1;i<num_categories;i++)
    fprintf(f,",%lf",input_theta[i]);
  fprintf(f,"\n");
  if(model==TN93) {
    fprintf(f,"%s=%lf",input_params[INITIAL_TTP][0],input_ttp[0]);
    if(!single_kappa)
      for(i=1;i<num_categories;i++)
	fprintf(f,",%lf",input_ttp[i]);
    fprintf(f,"\n");
    fprintf(f,"%s=%lf",input_params[INITIAL_GAMMA][0],input_gamma[0]);
    if(!single_kappa)
      for(i=1;i<num_categories;i++)
	fprintf(f,",%lf",input_gamma[i]);
    fprintf(f,"\n");
  }
  fprintf(f,"%s=%s\n",input_params[DATA_FILE][0],data_file);
  fprintf(f,"%s=%d\n",input_params[OUTGROUP][0],outgroup);
  fprintf(f,"%s=%lf\n",input_params[INITIAL_DELTA][0],valley_win);
  fprintf(f,"%s=%lf\n",input_params[INITIAL_KAPPA_DELTA][0],kappa_win);
  fprintf(f,"%s=%lf\n",input_params[THETA_CONST][0],theta_const);
  fprintf(f,"%s=%lf\n",input_params[PI_CONST][0],pi_const);  
  fprintf(f,"%s=%lf\n",input_params[MAX_INITIAL_TREE_HEIGHT][0],max_depth);
  fprintf(f,"%s=%s\n",input_params[TREE_FILE][0],tree_file);
  fprintf(f,"%s=%s\n",input_params[FILE_ROOT][0],file_root);
  fprintf(f,"%s=%s\n",input_params[NEWICK_FORMAT][0],
	  newick_format?"true":"false");
}

/*
 *  int conv_char(char c) 
 *
 *  A is coded as 0001, G is coded as 0010, C is coded as 0100, 
 *  and T(U) is coded as 1000
 */

int conv_char(char c, int n, int s)
{
  /* Returns the bit vector associated with the symbol. */

  switch (c) {
  case 'A' : case 'a' : 			return 0x1;
  case 'G' : case 'g' : 			return 0x2;
  case 'C' : case 'c' : 			return 0x4;
  case 'U' : case 'u' : case 'T' : case 't' :   return 0x8;
  case 'M' : case 'm' : 			return 0x1 | 0x4;
  case 'R' : case 'r' : 			return 0x1 | 0x2;
  case 'W' : case 'w' : 			return 0x1 | 0x8;
  case 'S' : case 's' : 			return 0x2 | 0x4;
  case 'Y' : case 'y' : 			return 0x4 | 0x8;
  case 'K' : case 'k' : 			return 0x2 | 0x8;
  case 'V' : case 'v' : 			return 0x1 | 0x2 | 0x4;
  case 'H' : case 'h' : 			return 0x1 | 0x4 | 0x8;
  case 'D' : case 'd' : 			return 0x1 | 0x2 | 0x8;
  case 'B' : case 'b' : 			return 0x2 | 0x4 | 0x8;
  case 'X' : case 'x' : case 'N' : case 'n' :   return 0x1 | 0x2 | 0x8 | 0x4;
  case '~' : case '-' : case '?' :		return 0x1 | 0x2 | 0x8 | 0x4;
  case '.' : 
    if(n==0) {
      fprintf(stderr,"Error: Base '.' appears in the first taxon.\n");
      exit(1);
    }
    return base[s];
  default:
    fprintf(stderr,"Warning: %c unknown base. - used.\n",c );
    return GAP;
  }
}

double runif()
{
  /* Returns a pseudo-random number between 0 and 1.
   *  runif is from the book Applied Statistics with S-PLUS by Ripley
   *  (check the title/author) */

  static int init = 0;
  static uint32 congrval;
  static uint32 tausval;
  uint32 n, lambda = 69069;

  if(init == 0)
  {
    init = 1;
    tausval = 644672833;
    if( seed % 2 == 1 )
      congrval = seed;
    else
    {  
      congrval = seed + 1;
      fprintf(stdout,"Changing initial seed to be odd number\n");
    }
  }    

  congrval = congrval * lambda;
  tausval ^= tausval >> 15;
  tausval ^= tausval << 17;
  n = tausval ^ congrval;
  return( (((n>>1) & 017777777777)) / 2147483648.);
}

void find_permutation(int perm[], int count)
{
  /* perm is set to a permutation of [0..count-1]. */

  int t,i;
  int n;

  for(i=0;i<count;i++)
    perm[i] = i;
  for(i=0;i<count-1;i++) {
    n = (int)(runif()*(count-i))+i;
    SWAP(t,perm[i],perm[n]);
  }
}

void add_base(Site s, int n, int b)
{
  /* Adds base b of taxon n to a site vector s. */
  int p, r, d, e;

  p = 4*(n / 31);
  r = n % 31;
  d = (r + 1) / 8;
  e = (r + 1) % 8;
  if(e==0) { 
    s[p+d-1] |= (b >> d);
    s[p+d]   |= ((b & (GAP >> (4 - d))) << (31 - d));
  }
  else
    s[p+d]   |= (b << (8*(d+1)-2-r)*4+(3-d)); 
}

int single_base(int sym, int *b)
{
  /* If sym corresponds to a single base, returns 1 and sets b to the base.
     Returns 0 otherwise. */

  switch (sym) {
  case 0x1: *b = 0; return 1;
  case 0x2: *b = 1; return 1;
  case 0x4: *b = 2; return 1;
  case 0x8: *b = 3; return 1;
  default: return 0;
  }
}

int get_next_line(FILE *f, char *buf, int r, int *lineno, char *filename)
{ 
  /* r is the expected number of additional lines of input.
     r = -1 means that no data has been read yet. */

  if(fgets(buf,MAX_LINE,f)) {
    (*lineno)++;
    return 1;
  }
  if(r == 0) 
    return 0;
  if(r == -1)
    fprintf(stderr,"Error: Data file %s is empty.\n",filename);
  else 
    fprintf(stderr,"Error: Data file %s ends prematurely.\n",filename);
  exit(1);
}

char *read_taxa_name(char *buf, int lineno, int n, int first, char *filename,
		     char *name)
{
  if(strlen(buf) < 17 || isspace(*(buf+16))) {
    fprintf(stderr,"Error: Line (%d) in data file %s contains no site "
	    "information.\n",
	    lineno,filename);
    exit(1);
  }
  *(buf+15) = '\0';
  strip_right_spaces(buf,buf+15);
  strcpy(name,buf);
  return buf+16;
}

int process_first_taxa(char *buf, int lineno, int first, int num_sites, 
		       char *filename, char **lbase)
{
  char *p,*q,name[MAX_LINE];
  int len, new_num_sites;

  p=read_taxa_name(buf,lineno,0,first,filename,name);
  if(first)
    strcpy(TAXANAME(0),name);
  len = strlen(p);
  if(p[len-1]=='\n') 
    p[--len] = '\0';
  new_num_sites = num_sites + len;
  if(!lbase[0]) {
    if(!(lbase[0]=(char *)malloc(sizeof(char)*(new_num_sites+1))))
      out_of_space("lbase (process_first_taxa)");
    strcpy(lbase[0]+num_sites,p);
  }
  else {
    if(!(lbase[0]=(char *)realloc(lbase[0],sizeof(char)*(new_num_sites+1))))
      out_of_space("lbase (process_first_taxa)");
    strcat(lbase[0]+num_sites,p);
  }
  return new_num_sites;
}

void process_other_taxa(char *buf, int lineno, int n, int first, 
			int num_sites, int new_num_sites, char *filename,
			char ***lbase, int *lmax_taxa)
{
  char *p,*q,name[MAX_LINE];
  int len,new_lmax_taxa,i;

  p=read_taxa_name(buf,lineno,0,first,filename,name);
  len = strlen(p);
  if(p[len-1]=='\n') 
    p[--len] = '\0';
  if(num_sites + len < new_num_sites) {
    fprintf(stderr,"Error: Line %d in data file %s ends prematurely.\n",
	    lineno,filename);
    exit(1);
  }
  if(num_sites + len > new_num_sites) {
    p[new_num_sites - num_sites] = '\0';
    fprintf(stderr,"Warning: Line %d of data file %s contains extra "
	    "information. Ignored.\n",
	    lineno,filename);
  }
  if(n >= *lmax_taxa) {
    new_lmax_taxa = 2 * *lmax_taxa;
    if(!(*lbase=(char **)realloc(*lbase,sizeof(char *)*new_lmax_taxa)))
      out_of_space("lbase (process_other_taxa 1)");
    for(i=*lmax_taxa;i<new_lmax_taxa;i++)
      (*lbase)[i] = NULL;
    if(!(taxaname=(char *)realloc(taxaname,sizeof(char)*
				  new_lmax_taxa*MAX_LINE)))
      out_of_space("taxaname");
    for(i=*lmax_taxa;i<new_lmax_taxa;i++) {
      (*lbase)[i] = NULL;
      TAXANAME(i)[0] = '\0';
    }
    *lmax_taxa = new_lmax_taxa;
  }
  if(first)
    strcpy(TAXANAME(n),name);
  if(!(*lbase)[n]) {
    if(!((*lbase)[n]=(char *)malloc(sizeof(char)*(new_num_sites+1))))
      out_of_space("lbase (process_other_taxa 2)");
    strcpy((*lbase)[n]+num_sites,p);
  }
  else {
    if(!((*lbase)[n]=(char *)realloc((*lbase)[n],
				     sizeof(char)*(new_num_sites+1))))
      out_of_space("lbase (process_other_taxa 2)");
    strcat((*lbase)[n]+num_sites,p);
  }
}

void read_clustal_file_int(FILE *f, char *filename,
			   char ***lbase, int *lmax_taxa)
{
  char buf[MAX_LINE];
  int lineno,n,k,i;

  /* Allocate initial space for the array of bases. */
  *lmax_taxa=10;
  if(!((*lbase)=(char **)malloc(sizeof(char *)*(*lmax_taxa))))
    out_of_space("lbase (read_clustal_file_int)");
  if(!(taxaname=(char *)malloc(sizeof(char)*(*lmax_taxa)*MAX_LINE)))
    out_of_space("taxaname");
  for(i=0;i<*lmax_taxa;i++) {
    (*lbase)[i] = NULL;
    TAXANAME(i)[0] = '\0';
  }
  lineno = 1; /* Header line read by read_datafile(). */
  /* skip blank lines */
  while(get_next_line(f,buf,-1,&lineno,filename) && isspace(*buf))
    ;
  /* read first section of data */
  /* read first line of first section */
  num_leaves = 1;
  num_all_sites = process_first_taxa(buf,lineno,1,0,filename,*lbase);
  if(!fgets(buf,MAX_LINE,f))
    return;
  lineno++;
  /* read rest of first section */
  while(*buf && !isspace(*buf)) {
    num_leaves++;
    process_other_taxa(buf,lineno,num_leaves-1,1,0,num_all_sites,filename,
		       lbase,lmax_taxa);
    if(!fgets(buf,MAX_LINE,f)) 
      return;
    lineno++;
  }
  /* read other sections */
  while(*buf) {
    /* skip blank lines */
    do {
      if(!fgets(buf,MAX_LINE,f))
	return;
      lineno++;
    }
    while(isspace(*buf));
    /* read first line of other section */
    k = process_first_taxa(buf,lineno,0,num_all_sites,filename,*lbase);
    if(!get_next_line(f,buf,num_leaves-1,&lineno,filename))
      return;
    for(n=1;n<num_leaves;n++) {
      process_other_taxa(buf,lineno,n,0,num_all_sites,k,filename,
			 lbase,lmax_taxa);
      if(!get_next_line(f,buf,num_leaves-n-1,&lineno,filename))
	return;
    }
    if(*buf && !isspace(*buf)) {
      fprintf(stderr,"Error: Section of data file %s is not terminated by "
	      "blank line or end of file. At line %d.\n",
	      filename,lineno);
      exit(1);
    }
    num_all_sites = k;
  }
}

void read_clustal_file(FILE *f, char *filename)
{
  char **lbase;
  int lmax_taxa,n,s;

  read_clustal_file_int(f,filename,&lbase,&lmax_taxa);
  if(!(base=(int *)malloc(sizeof(int)*num_leaves*num_all_sites)))
    out_of_space("base");
  for(n=0;n<num_leaves;n++)
    for(s=0;s<num_all_sites;s++)
      BASE(n)[s] = conv_char(lbase[n][s],n,s);
  for(n=0;n<lmax_taxa;n++)
    free(lbase[n]);
  free(lbase);
}

int isempty(char *buf)
{
  char *p;

  for(p=buf;*p;p++)
    if(!isspace(*p))
      return 0;
  return 1;
}

void read_bambe_file(FILE *f, char *filename)
{
  /* Reads in the bambe file f and stores the site information
     in base[n][k]. */

  int n,site,c,len;
  char buf[MAX_LINE],*comment;

  if(!(base=(int *)malloc(sizeof(int)*num_leaves*num_all_sites)))
    out_of_space("base");
  if(!(taxaname=(char *)malloc(sizeof(char)*num_leaves*MAX_LINE)))
    out_of_space("taxaname");
  for(n=0;n<num_leaves;n++)
  {
    /* Read the taxon name */
    do
      if(!fgets(buf,MAX_LINE,f))
	return;
    while(isempty(buf)) ;
    if(*buf && buf[(len=strlen(buf))-1]!='\n')
      SKIPLINE(f,c);
    if(comment=strchr(buf,'#'))
      strip_right_spaces(buf,comment);
    else
      strip_right_spaces(buf,buf+len);
    strcpy(TAXANAME(n),buf);
    for(site=0;site<num_all_sites;site++) {
      while(isspace(c=getc(f)))
	;
      if(c == EOF) {
	fprintf(stderr,"Error: Premature end of data file %s.\n",
		filename);
	exit(1);
      }
      BASE(n)[site] = conv_char(c,n,site);
    }
    SKIPLINE(f,c);
  }
}

void parse_catstring(char **str, int *tag_list, int *len_tl, int *lastgroup,
		     int *internal_cat) 
{
  /* Parses a string representing a list of categories up to, but
     not including the final asterisk (if it exists).
     str is the unscanned part of the category list string.
     tag_list is the corresponding array of tags.
     *len_tl is the size of tag_list.
     *lastgroup is the starting position of the last group in tag_list.
     internal_cat maps the category names to the tag names. */

  int i,j,n,size,new_len,x;

  while(isspace(**str))
    (*str)++;
  if(**str == '(') {
    (*str)++;
    *lastgroup = *len_tl;
    parse_catstring(str,tag_list,len_tl,&x,internal_cat);
    while(isspace(**str))
      (*str)++;
    (*str)++;
  }
  else if(isdigit(**str)) {
    n = **str - '0';
    for((*str)++;isdigit(**str);(*str)++)
      n = 10 * n + **str - '0';
    *lastgroup = *len_tl;
    if((*len_tl) >= num_all_sites) {
      fprintf(stderr,"Warning: There are more categories than sites. The "
	      "remainder will be ignored.\n");
      return;
    }
    tag_list[(*len_tl)++] = internal_cat[n];
  }
  else 
    return;
  while(isspace(**str))
    (*str)++;
  if(**str == '^') {
    (*str)++;
    while(isspace(**str))
      (*str)++;
    n = **str - '0';
    for((*str)++;isdigit(**str);(*str)++)
      n = 10 * n + **str - '0';
    size = *len_tl - *lastgroup;
    new_len = *len_tl + size * (n-1);
    if(new_len > num_all_sites) {
      fprintf(stderr,"Warning: There are more categories than sites. The "
	      "remainder will be ignored.\n");
      for(i=*len_tl,j=*lastgroup;i<num_all_sites;i++,j++)
	tag_list[i] = tag_list[j];
      *len_tl = num_all_sites;
      return;
    }
    for(i=*len_tl,j=*lastgroup;i<new_len;i++,j++)
      tag_list[i] = tag_list[j];
    *len_tl = new_len;
    *lastgroup = new_len - size;
    while(isspace(**str))
      (*str)++;
  }
  if(**str == ',') {
    (*str)++;
    parse_catstring(str,tag_list,len_tl,lastgroup,internal_cat);
  }
}

void process_bases()
{
  /* Processes the bases in the global array base[n][k] and stores the
     sites (ordered by tags) in the site array. Also counts the number 
     of symbols and sets up the list of uncertain bases for the taxa. 
     Assumes that base[n][k] is in the order read in. */

  int tag,site,sym,n,word,b,j,x,i;
  int (*base_seen)[NUM_SYMBOLS],*base2;
  int *count,*sitepos;
  int internal_cat[MAX_CATEGORIES],*tag_list,len_tl,lastgroup;
  char *p;

  if(!(base_seen=(int (*)[NUM_SYMBOLS])malloc(sizeof(int[NUM_SYMBOLS])
					      * num_tags)))
    out_of_space("base_seen");
  if(!(base2=(int *)malloc(sizeof(int)*num_all_sites)))
    out_of_space("base2");
  if(!(count=(int *)malloc(sizeof(int)*num_tags)))
    out_of_space("count");
  if(!(sitepos=(int *)malloc(sizeof(int)*num_all_sites)))
    out_of_space("sitepos");

  num_nodes = 2 * num_leaves - 1;
  site_len = (num_leaves * NUM_BASES + 30) / 31;

  if(outgroup>num_leaves)
    if(mclock)
      fprintf(stderr,"Warning: outgroup (%d) (unused in this model) exceeds"
	      " number of taxa (%d).\n",
	      outgroup,num_leaves);
    else {
      fprintf(stderr,"Error: outgroup (%d) exceeds number of taxa (%d).\n",
	      outgroup,num_leaves);
      exit(1);
    }

  if(!(tag_list=(int *)malloc(sizeof(int)*num_all_sites)))
    out_of_space("tag_list");
  for(tag=0;tag<num_tags;tag++)
    internal_cat[categories[tag]] = tag;
  len_tl = 0;
  p = category_list;
  parse_catstring(&p,tag_list,&len_tl,&lastgroup,internal_cat);
  if(len_tl < num_all_sites)
    if(*p != '*') {
    fprintf(stderr,"Error: There are fewer categories than sites.\n");
    exit(1);
    }
  else
    for(i=len_tl,j=lastgroup;i<num_all_sites;i++,j++)
      tag_list[i] = tag_list[j];


  if(!(tag_sites=(int *)malloc(sizeof(int)*num_tags)))
    out_of_space("tag_sites");
  if(!(tag_weights=(double *)malloc(sizeof(double)*num_tags)))
    out_of_space("tag_weights");
  if(!(tag_symbols=(int *)malloc(sizeof(int)*num_tags*NUM_SYMBOLS)))
    out_of_space("tag_symbols");
  if(!(taxa_symbols=(int *)malloc(sizeof(int)*num_leaves*NUM_SYMBOLS)))
    out_of_space("taxa_symbols");
  if(!(num_uncertain=(int *)malloc(sizeof(int)*num_leaves)))
    out_of_space("num_uncertain");
  if(!(uncertain_base=(int *)malloc(sizeof(int)*num_leaves*NUM_SYMBOLS)))
    out_of_space("uncertain_base");
  if(!(uncertain_tag=(int *)malloc(sizeof(int)*num_leaves*NUM_SYMBOLS)))
    out_of_space("uncertain_tag");
  if(!(first_site=(int *)malloc(sizeof(int)*(num_tags+1))))
    out_of_space("first_site");
  if(!(sites=(int *)malloc(sizeof(int)*num_all_sites*site_len)))
    out_of_space("sites");

  for(tag=0;tag<num_tags;tag++)
    tag_sites[tag] = 0;
  for(site=0;site<num_all_sites;site++)
    tag_sites[tag_list[site]]++;

  first_site[0] = 0;
  for(tag=0;tag<num_tags;tag++) {
    count[tag] = 0;
    first_site[tag+1] = first_site[tag] + tag_sites[tag];
    tag_weights[tag] = (double)tag_sites[tag] / num_all_sites;
  }

  for(sym=1;sym<NUM_SYMBOLS;sym++) {
    total_symbols[sym] = 0;
    for(tag=0;tag<num_tags;tag++)
      TAG_SYMBOLS(tag)[sym] = 0;
    for(n=0;n<num_leaves;n++)
      TAXA_SYMBOLS(n)[sym] = 0;
  }

  for(site=0;site<num_all_sites;site++) {
      tag = tag_list[site];
      sitepos[site] = first_site[tag] + count[tag]++;
      for(word=0;word<site_len;word++)
	SITES(site)[word] = 0;
  }

  for(n=0;n<num_leaves;n++)
  {
    num_uncertain[n] = 0;
    for(tag=0;tag<num_tags;tag++)
      for(sym=1;sym<NUM_SYMBOLS;sym++)
	base_seen[tag][sym] = 0;

    for(site=0;site<num_all_sites;site++)
      base2[site] = BASE(n)[site];

    for(site=0;site<num_all_sites;site++) {
      tag = tag_list[site];
      j = sitepos[site];

      b = BASE(n)[j] = base2[site];
      add_base(SITES(j),n,b);
      TAG_SYMBOLS(tag)[b]++;
      TAXA_SYMBOLS(n)[b]++;
      total_symbols[b]++;
      if(b!=GAP && !single_base(b,&x) && !base_seen[tag][b]) {
	base_seen[tag][b] = 1;
	UNCERTAIN_BASE(n)[num_uncertain[n]] = b;
	UNCERTAIN_TAG(n)[num_uncertain[n]++] = tag;
      }
    }
  }
  free(tag_list);
  free(base_seen);
  free(base2);
  free(count);
  free(sitepos);
}

void read_datafile(char *filename)
{
  /* Reads in either a BAMBE or clustal datafile named filename and
     stores the sites (ordered by tags) in the site array. Also counts 
     the number of symbols and sets up the list of uncertain bases 
     for the taxa. */

  FILE *f;
  char buf[MAX_LINE];
  int i;

  if((f=fopen(filename,"r"))==NULL) {
    fprintf(stderr,"Error: Cannot open data file %s.\n", filename);
    exit(1);
  }
  if(!fgets(buf,MAX_LINE,f)) {
    fprintf(stderr,"Error: Data file %s is empty.\n",
	    filename);
    exit(1);
  }
  if(sscanf(buf,"%d%d",&num_leaves,&num_all_sites)==2)
    read_bambe_file(f,filename);
  else if(strstr(buf,"CLUSTAL"))
    read_clustal_file(f,filename);
  else {
    fprintf(stderr,"Error: Data file %s is not in a readable format.\n",
	    filename);
    exit(1);
  }
  fclose(f);
  process_bases();
  if(!(ordered_names=(Numbered_name *)malloc(sizeof(Numbered_name)*
					     num_leaves)))
    out_of_space("ordered_names");
  for(i=0;i<num_leaves;i++) {
    ordered_names[i].n = i;
    strcpy(ordered_names[i].name,TAXANAME(i));
  }
  qsort(ordered_names,num_leaves,sizeof(Numbered_name),cmp_name);
}

void open_file(FILE **f,char *extension)
{
  /* Opens a file for writing. */
  char filename[81];

  strcpy(filename,file_root); 
  strcat(filename,extension);
  if( (*f = fopen(filename,"r")) == NULL)
    *f = fopen(filename,"w");
  else
  {
    fprintf(stderr,"Error: You probably do not wish to overwrite %s\n",
	    filename);
    exit(1);
  }
}

void open_files(Files *files)
{
  /* Opens the output files. */
  open_file(&(files->flogp),".lpd");
  open_file(&(files->fkappa),".par");
  open_file(&(files->ftop),".top");
  open_file(&(files->fout),".out");
  open_file(&(files->ftree),".tre");
}

void close_files(Files *files)
{
  /* Closes the output files. */
  fclose(files->flogp);
  fclose(files->fkappa);
  fclose(files->ftop);
  /*fclose(files->fout);*/
  fclose(files->ftree);
}

void print_initial_information(FILE *f, const Parameters params)
{
  /* Prints out summary information. */

  int b,d,i,count,row_total,x,sym,total,column_total[NUM_BASES];
  double sum,gapprop,prop;
  char base_symbol[NUM_SYMBOLS] = {'\0', 'A', 'G', 'R', 'C', 'M', 'S', 'V', 
				   'T', 'W', 'K', 'D', 'Y', 'H', 'B', '-'};
  char format[] = "   %c      %5d      %6.4lf\n";

  fprintf(f,"Output generated by BAMBE version %s\n", VERSION);
  fprintf(f,"Data Summary\n\n");
  fprintf(f,"%5d\t\tnumber of taxa\n",num_leaves);
  fprintf(f,"%5d\t\ttotal number of sites\n",num_all_sites);
  fprintf(f,"%5d\t\ttotal number of unique sites\n\n",num_unique_sites);
  d = num_leaves * num_all_sites;
  fprintf(f,"Number of constant sites\n\nCategory");
  total = 0;
  for(b=0;b<NUM_BASES;b++) {
    fprintf(f,"   %c   ",base_symbol[0x1<<b]);
    column_total[b] = 0;
  }
  fprintf(f,"Total\n");
  for(i=0;i<num_tags;i++) {
    row_total = 0;
    fprintf(f,"%3d     ",categories[i]);
    for(b=0;b<NUM_BASES;b++) {
      fprintf(f,"%4d   ",ROOT_KSITES(i)[b]);
      row_total += ROOT_KSITES(i)[b];
      column_total[b] += ROOT_KSITES(i)[b];
    }
    total += row_total;
    fprintf(f,"%4d\n",row_total);
  }
  fprintf(f,"Total   ");
  for(b=0;b<NUM_BASES;b++) 
    fprintf(f,"%4d   ",column_total[b]);
  fprintf(f,"%4d\n\n",total);
  fprintf(f,"Overall   Count   Proportion\n\n");
  for(b=0;b<NUM_BASES;b++)
    fprintf(f,format,
	   base_symbol[0x1<<b], total_symbols[0x1<<b], 
	   (double)total_symbols[0x1<<b]/(double)(d));
  fprintf(f,format,
         base_symbol[GAP], total_symbols[GAP], 
	 (double)total_symbols[GAP]/(double)(d));
  for(sym=1;sym<NUM_SYMBOLS;sym++)
    if(sym!=GAP && !single_base(sym,&x))
      fprintf(f,format,
	     base_symbol[sym], total_symbols[sym], 
	     (double)total_symbols[sym]/(double)(d));
  fprintf(f,"\nCategory     ");
  for(b=0;b<NUM_BASES;b++)
    fprintf(f,"%c       ",base_symbol[0x1<<b]);
  fprintf(f,"%c      other\n",base_symbol[GAP]);
  for(i=0;i<num_tags;i++) {
    fprintf(f,"%3d    Pi",categories[i]);
    for(b=0;b<NUM_BASES;b++)
      fprintf(f,"  %6.4lf",params[i].pi[b]);
    fprintf(f,"\n     Prop");
    count = tag_sites[i];
    sum = gapprop = (double)TAG_SYMBOLS(i)[GAP]/(double)(count*num_leaves);
    for(b=0;b<NUM_BASES;b++) {
      prop = (double)TAG_SYMBOLS(i)[0x1<<b]/(double)(count*num_leaves);
      sum += prop;
      fprintf(f,"  %6.4lf",prop);
    }
    fprintf(f,"  %6.4lf  %6.4lf\n",gapprop,1-sum);
  }
  fprintf(f,"\nTaxon        ");
  for(b=0;b<NUM_BASES;b++)
    fprintf(f,"%c       ",base_symbol[0x1<<b]);
  fprintf(f,"%c      other\n",base_symbol[GAP]);
  for(i=0;i<num_leaves;i++)
  {
    fprintf(f,"%3d      ",i+1);
    sum = gapprop = (double)TAXA_SYMBOLS(i)[GAP]/(double)num_all_sites;
    for(b=0;b<NUM_BASES;b++) {
      prop = (double)TAXA_SYMBOLS(i)[0x1<<b]/(double)num_all_sites;
      sum += prop;
      fprintf(f,"  %6.4lf", prop);
    }
    fprintf(f,"  %6.4lf  %6.4lf\n", gapprop, 1-sum);
  }
  fprintf(f,"\n\nTaxa names:\n");
  for(i=0;i<num_leaves;i++)
    fprintf(f,"%3d  %s\n",i+1,TAXANAME(i));
  fprintf(f,"\n\nRun Settings:\n\n");
  print_input_params(f);
}

void read_bambe_tree(FILE *f, Tree *t)
{
  /* Reads a bambe format tree from the file tree_file. */

  int i,ttaxa,tlchild,trchild,tnode,lc,rc,m1,m2,m3,parent,c;
  double tbranchlen;

  if(fscanf(f,"%d",&t->root) != 1) {
    fprintf(stderr,"Error: Tree file not in BAMBE format.\n");
    exit(1);
  }
  SKIPLINE(f,c);
  SKIPLINE(f,c);
  for(i=0;i<num_leaves;i++)
  {
    /* tnode should always be the same as i. */
    if(fscanf(f,"%d%d%d%d%d%lf",&ttaxa,&tnode,&tlchild,&trchild,&parent,
	      &tbranchlen) != 6) {
      fprintf(stderr,"Error: Premature end of data in tree file.\n");
      exit(1);
    }
    t->lchild[i] = tlchild;
    t->rchild[i] = trchild;
    t->branchlen[i] = tbranchlen;
    SKIPLINE(f,c);
  }
  for(i=num_leaves;i<num_nodes;i++)
  {
    if(fscanf(f,"%d%d%d%d%d%lf",&ttaxa,&tnode,&tlchild,&trchild,&parent,
	      &tbranchlen) != 6) {
      fprintf(stderr,"Error: Premature end of data in tree file.\n");
      exit(1);
    }
    t->parent[tlchild] = i;
    t->parent[trchild] = i;
    t->lchild[i] = tlchild;
    t->rchild[i] = trchild;
    t->branchlen[i] = tbranchlen;
    SKIPLINE(f,c);
  }
  t->parent[t->root] = -1;
}

int find_taxa(Numbered_name *key)
{
  Numbered_name *p;

  if(p=(Numbered_name*)bsearch(key,ordered_names,num_leaves,sizeof(Numbered_name),cmp_name))
    return p->n;
  fprintf(stderr,"Error: Taxon name /%s/ not found.\n",key->name);
  exit(1);
}

int read_one_char(FILE *f)
{ 
  int c;

  if((c=getc(f)) == EOF) {
    fprintf(stderr,"Error: Input file ends prematurely.\n");
    exit(1);
  }
  return c;
}

int read_nonspace_char(FILE *f)
{
  char d;

  while(isspace(d=read_one_char(f)))
    ;
  return d;
}

int read_special_char(FILE *f, char c, char *name)
{
  char d;

  d = read_nonspace_char(f);
  if(d != c) {
    fprintf(stderr,"Error: Expected %s in input file. Found %c.\n",
	    name,d);
    exit(1);
  }
  return d;
}

double read_branchlen(FILE *f)
{
  double b;

  if(fscanf(f,"%lf",&b) != 1) {
    fprintf(stderr,"Error: Missing branchlen in input tree.\n");
    exit(1);
  }
  return b;
}

int read_newick_int(FILE *f, Tree *t, int *n)
{
  /* n is the number of the next internal node to be created. 
     Returns the root of the subtree read in. */

  int c,i,lc,rc,n1;
  Numbered_name nbuf;
  double bl;

  c = read_nonspace_char(f);
  if(c == '(') {
    lc = read_newick_int(f,t,n);
    read_special_char(f,',',"comma");
    rc = read_newick_int(f,t,n);
    read_special_char(f,')',"right parenthesis");
    read_special_char(f,':', "colon");
    if(*n>=num_nodes) {
      fprintf(stderr,"Error: Too many nodes in input tree.\n");
      exit(1);
    }
    t->branchlen[*n] = read_branchlen(f);
    t->lchild[*n] = lc;
    t->rchild[*n] = rc;
    t->parent[lc] = t->parent[rc] = *n;
    return (*n)++;
  }
  else {
    nbuf.name[0] = c;
    for(c=read_one_char(f),i=1;c!=':';c=read_one_char(f))
      if(i<MAX_LINE-1)
	nbuf.name[i++] = c;
    nbuf.name[i] = '\0';
    n1 = find_taxa(&nbuf);
    t->lchild[n1] = t->rchild[n1] = 0;
    bl = read_branchlen(f);
    t->branchlen[n1] = bl; 
    return n1;
  }
}

void read_newick_tree(FILE *f, Tree *t)
{
  int c,n,lc,rc,len;
  Numbered_name nbuf;
  int i;

  c = read_nonspace_char(f);
  if(c == '(') {
    n = num_leaves;	/* number of first internal node */
    lc = read_newick_int(f,t,&n);
    read_special_char(f,',',"comma");
    rc = read_newick_int(f,t,&n);
    read_special_char(f,')',"right parenthesis");
  }
  else {
    ungetc(c,f);
    fgets(nbuf.name,MAX_LINE,f);
    if(*nbuf.name && nbuf.name[(len=strlen(nbuf.name))-1]=='\n')
      nbuf.name[len-1] = '\0';
    n = find_taxa(&nbuf);
    if(n != 0) {
      fprintf(stderr,"Internal error: The only taxon is not taxon 0.\n");
      exit(1);
    }
    lc = rc = -1;
  }
  if(n!=num_nodes-1) {
    fprintf(stderr,"Error: Wrong number of nodes in input tree.\n");
    exit(1);
  }
  t->root = n;
  t->lchild[n] = lc;
  t->rchild[n] = rc;
  if(lc != -1)
    t->parent[lc] = t->parent[rc] = n;
  t->branchlen[n] = 0;
  t->parent[n] = -1;
}

void find_heights(const Tree *t, double *ht, int n, double h,
		 double *height)
{
  /* h is the distance between n and the subroot.. Sets ht[i] to 
     the distance between i and the subroot for each i in the subtree.
     Returns the height of the subtree. */

  ht[n] = h;
  if(LEAF(n))
    *height = h;
  else {
    find_heights(t,ht,t->lchild[n],h+t->branchlen[t->lchild[n]],height);
    find_heights(t,ht,t->rchild[n],h+t->branchlen[t->rchild[n]],height);
  }
}

void read_initial_tree(Tree *t)
{
  /* Reads a tree from the file tree_file. */

  FILE *f;
  double *ht,height;
  int n,i,bad=0;

  if(!(f = fopen(tree_file,"r"))) {
    fprintf(stderr,"Error: Cannot open treefile %s.\n",tree_file);
    exit(1);
  }
  if(newick_format) {
    for(i=0;i<num_leaves-1;i++)
      if(strcmp(ordered_names[i].name,ordered_names[i+1].name)==0) {
	fprintf(stderr,"Error: Taxa have identical names (%s). Cannot read "
		"tree in newick format.\n",
		ordered_names[i].name);
	bad = 1;
      }
    if(bad)
      exit(1);
    read_newick_tree(f,t);
  }
  else
    read_bambe_tree(f,t);
  fclose(f);
  if(mclock) {
    if(!(ht=(double *)malloc(sizeof(double)*num_nodes)))
      out_of_space("ht");
    find_heights(t,ht,t->root,0,&height);
    for(n=0;n<num_leaves;n++)
      if(fabs(ht[n]-height)>1e-5) {
	fprintf(stderr,"Error: Molecular clock does not hold for initial "
		"tree.\n");
	exit(1);
      }
    for(n=0;n<num_leaves;n++)
      t->branchlen[n] += height - ht[n];
    free(ht);
  }
}

double tree_height(const Tree *t,int i)
{
  /* returns the height of tree t rooted at i (assumes clock model). */

  return (LEAF(i) ?
	  0 :
	  t->branchlen[t->lchild[i]] + tree_height(t,t->lchild[i]));
}

void tree_diameter(const Tree *t, int i, double *h, double *d)
{
  /* For tree t, rooted at i, sets h to the maximum distance from 
     a leaf to i, and sets d to the diameter (maximum distance between
     any two nodes). Does not assume the clock model. */

  double h1,h2,d1,d2,l1,l2;

  if(LEAF(i))
    *h = *d = 0;
  else {
    tree_diameter(t,t->lchild[i],&h1,&d1);
    tree_diameter(t,t->rchild[i],&h2,&d2);
    l1 = h1+t->branchlen[t->lchild[i]];
    l2 = h2+t->branchlen[t->rchild[i]];
    *h = MAXIMUM(l1,l2);
    *d = MAXIMUM(MAXIMUM(d1,d2),l1+l2);
  }
}

double tree_radius(const Tree *t)
{
  /* Returns the radius of tree t. Does not assume the clock model. */

  double h,d;
  tree_diameter(t,t->root,&h,&d);
  return d/2;
}

void fprint_tree_topology_int(FILE *f, const Tree *t, int n, 
			      const int *min_taxa)
{
  /* Prints out the tree topology of t to file f. */
  if(LEAF(n))
    fprintf(f,"%d",n+1);
  else {
    fprintf(f,"(");
    if (min_taxa[t->lchild[n]] < min_taxa[t->rchild[n]]) {
      fprint_tree_topology_int(f,t,t->lchild[n],min_taxa);
      fprintf(f,",");
      fprint_tree_topology_int(f,t,t->rchild[n],min_taxa);
    }
    else {
      fprint_tree_topology_int(f,t,t->rchild[n],min_taxa);
      fprintf(f,",");
      fprint_tree_topology_int(f,t,t->lchild[n],min_taxa);
    }
    fprintf(f,")");
  }
}

void find_min_taxa(const Tree *t, int n, int min_taxa[])
{
  if(LEAF(n)) {
    min_taxa[n] = n;
  }
  else {
    find_min_taxa(t,t->lchild[n],min_taxa);
    find_min_taxa(t,t->rchild[n],min_taxa);
    min_taxa[n] = MINIMUM(min_taxa[t->lchild[n]],min_taxa[t->rchild[n]]);
  }
}

void copy_subtree_topology_int(Tree *t1, int n, const Tree *t2)
{
  t1->lchild[n] = t2->lchild[n];
  t1->rchild[n] = t2->rchild[n];
  t1->parent[n] = t2->parent[n];
  t1->branchlen[n] = t2->branchlen[n];
  if(!LEAF(n)) {
    copy_subtree_topology_int(t1,t2->lchild[n],t2);
    copy_subtree_topology_int(t1,t2->rchild[n],t2);
  }
}

void copy_subtree_topology(Tree *t1, int subroot, const Tree *t2)
{
  int subroot_parent=t2->parent[subroot];

  if(subroot_parent!=-1) {
    t1->lchild[subroot_parent] = t2->lchild[subroot_parent];
    t1->rchild[subroot_parent] = t2->rchild[subroot_parent];
  }
  t1->root = t2->root;
  t1->loglikelihood = t2->loglikelihood;
  copy_subtree_topology_int(t1,subroot,t2);
}

int split_branch_int(Tree *t, int x, int r, int s, int inside[],
		     double bf)
{
  int c;
  double b;

  /* The subtree rooted at r contains x. s is the sibling of r.
     p is the parent of both. */

  if(r == x) {
    t->branchlen[r] = t->branchlen[s] = t->branchlen[s]/2;
    /*b = t->branchlen[s];
    t->branchlen[r] = b * bf;
    t->branchlen[s] = b * (1 - bf);*/
    return s;
  }
  if(inside[(c=t->lchild[r])])
    t->lchild[r] = s;
  else {
    c = t->rchild[r];
    t->rchild[r] = s;
  }
  t->parent[s] = r;
  t->branchlen[r] = t->branchlen[c];
  return split_branch_int(t, x, c, r, inside, bf);
}

int split_branch(Tree *t, int x)
{
  /* x should not be the root of t. */

  int i,*inside,n;
  int r;	/* The child of the root whose subtree contains x. */
  int s;	/* The other child of the root. */

  if(!(inside=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("inside");
  for(i=0;i<num_nodes;i++)
    inside[i] = 0;
  for(i=x;i!=-1;i=t->parent[i])
    inside[i] = 1;
  if(inside[t->lchild[t->root]]) {
    r = t->lchild[t->root]; s = t->rchild[t->root];
  }
  else {
    r = t->rchild[t->root]; s = t->lchild[t->root];
  }
  t->branchlen[s] = t->branchlen[r] + t->branchlen[s];
  n = split_branch_int(t,x,r,s,inside,t->branchlen[r]/t->branchlen[s]);
  t->lchild[t->root] = x;
  t->rchild[t->root] = n;
  t->parent[x] = t->parent[n] = t->root;
  free(inside);
  return s;
}

void fprint_tree_topology(FILE *f, const Tree *t)
{
  int *min_taxa;
  Tree *t2;
  
  if(!(min_taxa=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("min_taxa");
  if(mclock) {
    find_min_taxa(t,t->root,min_taxa);
    fprint_tree_topology_int(f,t,t->root,min_taxa);
  }
  else {
    t2 = make_tree();
    copy_subtree_topology(t2,t->root,t);
    split_branch(t2,outgroup-1);
    find_min_taxa(t2,t2->root,min_taxa);
    fprint_tree_topology_int(f,t2,t2->root,min_taxa);
    free_tree(t2);
  }
  fprintf(f,"\n");
  free(min_taxa);
}

void fprint_tree_bambe(FILE *f, const Tree *t)
{
  /* Prints out tree t to file f. */
  int i;
  double *ht,height;

  if(!(ht=(double *)malloc(sizeof(double)*num_nodes)))
    out_of_space("ht");
  fprintf(f," %d %d\n",t->root,num_leaves);
  fprintf(f,"taxon node lchild rchild parent branchlen    height\n");
  find_heights(t,ht,t->root,0,&height);
  for(i=0;i<num_nodes;i++)
  {
    fprintf(f,"%4d %5d %6d %6d %6d %9.6lf %9.6lf\n",
	    LEAF(i)?i+1:-1,i,t->lchild[i],t->rchild[i],t->parent[i], 
	    t->branchlen[i], ht[i]); 
  }
  fprintf(f,"log likelihood = %lf\n\n",t->loglikelihood);
  free(ht);
}

void fprint_tree_newick_int(FILE *f, const Tree *t, int n,
			    const int *min_taxa, int isroot)
{
  /* Prints out the tree t in newick format to file f. */
  if(LEAF(n))
      fprintf(f,"%s",TAXANAME(n));
  else {
    fprintf(f,"(");
    if (min_taxa[t->lchild[n]] < min_taxa[t->rchild[n]]) {
      fprint_tree_newick_int(f,t,t->lchild[n],min_taxa,0);
      fprintf(f,",");
      fprint_tree_newick_int(f,t,t->rchild[n],min_taxa,0);
    }
    else {
      fprint_tree_newick_int(f,t,t->rchild[n],min_taxa,0);
      fprintf(f,",");
      fprint_tree_newick_int(f,t,t->lchild[n],min_taxa,0);
    }
    fprintf(f,")");
  }
  if(!isroot)
    fprintf(f,":%.6lf",t->branchlen[n]);
}

void fprint_tree_newick(FILE *f, const Tree *t)
{
  int *min_taxa;
  Tree *t2;
  
  if(!(min_taxa=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("min_taxa");
  if(mclock) {
    find_min_taxa(t,t->root,min_taxa);
    fprint_tree_newick_int(f,t,t->root,min_taxa,1);
  }
  else {
    t2 = make_tree();
    copy_subtree_topology(t2,t->root,t);
    split_branch(t2,outgroup-1);
    find_min_taxa(t2,t2->root,min_taxa);
    fprint_tree_newick_int(f,t2,t2->root,min_taxa,1);
    free_tree(t2);
  }
  fprintf(f,"\n");
  free(min_taxa);
}

void fprint_tree(FILE *f, const Tree *t)
{
  if(newick_format)
    fprint_tree_newick(f,t);
  else
    fprint_tree_bambe(f,t);
}

/*
 *  find_ptable computes the probability transition matrix for a given
 *  branchlength and set of parameters according to the Tamura-Nei model.
 *  
 *  This assumes the infintesimal rate matrix: where K = 4*ttr/(gamma+1)
 *
 *             A        G        C              T
 *         A  -()      K pi_g   pi_c           pi_t
 *         G  K pi_a   -()      pi_c           pi_t
 *         C  pi_a     pi_g     -()            K gamma pi_t
 *         T  pi_a     pi_g     K gamma pi_c   -()
 *
 *  ptable is the transpose of the conventional representation
 *  (with column sums = 1)
 *
 */

void find_ptable(Table ptab, const double blen, 
		 const Parameters params)

{
  double ebetar, ebetay, p, q, btheta, pia, pig, pic, pit;
  double pir, piy, pry, pyr, ttp, gamma;
  double tempr1, tempr2, tempy1, tempy2;
  int i;

  for(i=0;i<num_tags;i++) {
    btheta = blen * params[i].theta;
    pia = params[i].pi[0];
    pig = params[i].pi[1];
    pic = params[i].pi[2];
    pit = params[i].pi[3];
    pir = params[i].pi_r;
    piy = params[i].pi_y;
    pry = params[i].pry;
    pyr = params[i].pyr;
    ttp = params[i].ttp;
    gamma = params[i].gamma;

    p = SAFE_EXP(-1 * btheta);
    q = 1 - p;
    ebetar = SAFE_EXP(-1 * btheta * ( (4 * ttp * pir) / (gamma + 1) + piy) );
    ebetay = SAFE_EXP(-1 * btheta * ( (4 * ttp * gamma * piy)
				 / (gamma + 1) + pir));
    tempr1 = 1 + pyr * p;
    tempy1 = 1 + pry * p;
    tempr2 = (1 + pyr * p - ebetar / pir);
    tempy2 = (1 + pry * p - ebetay / piy);

    PTAB(i,0)[0] = pia*tempr1 + (pig/pir)*ebetar;
    PTAB(i,0)[1] = pia*tempr2;
    PTAB(i,0)[2] = PTAB(i,0)[3] = pia*q;
    PTAB(i,1)[0] = pig*tempr2;
    PTAB(i,1)[1] = pig*tempr1 + (pia/pir)*ebetar;
    PTAB(i,1)[2] = PTAB(i,1)[3] = pig*q;
    PTAB(i,2)[0] = pic*q;
    PTAB(i,2)[1] = PTAB(i,2)[0];
    PTAB(i,2)[2] = pic*tempy1 + (pit/piy)*ebetay;
    PTAB(i,2)[3] = pic*tempy2;
    PTAB(i,3)[0] = PTAB(i,3)[1] = pit*q;
    PTAB(i,3)[2] = pit*tempy2;
    PTAB(i,3)[3] = pit*tempy1 + (pic/piy) * ebetay;
  }
}

void create_initial_parameters(Parameters *params)
{
  int b, i, missing_base;
  double denom, t;

  /* Rescale initial_theta if weighted mean is not 1 */

  denom = 0;

  if(!(*params=(Parameters)malloc(sizeof(struct parameters)*num_tags)))
    out_of_space("parameters");
  for(i=0;i<num_tags;i++)
    denom += initial_theta[i] * tag_sites[i];
  denom /= num_all_sites;
  for(i=0;i<num_tags;i++)
    initial_theta[i] /= denom;
  if(fabs(denom-1.0) > TOLERANCE)
    fprintf(stderr,"Warning: Initial theta values rescaled"
            " so weighted average equals 1.\n");
  
  for(i=0;i<num_tags;i++) {
    (*params)[i].kappa = initial_kappa[i];
    (*params)[i].theta = initial_theta[i];
    denom = 0;
    /* If a base is missing for the entire data (a rare occurrence, we hope)
       then bump up all base counts by 1 to avoid 0's in the transition
       table. */
    missing_base = 0;
    for(b=0;b<NUM_BASES;b++) {
      denom += TAG_SYMBOLS(i)[0x1 << b];
      if(TAG_SYMBOLS(i)[0x1 << b]==0)
	missing_base = 1;
    }
    if(missing_base) 
      for(b=0;b<NUM_BASES;b++)
	(*params)[i].pi[b] = (double)(TAG_SYMBOLS(i)[0x1 << b]+1) 
	  / (denom+NUM_BASES);
    else
      for(b=0;b<NUM_BASES;b++)
	(*params)[i].pi[b] = (double)TAG_SYMBOLS(i)[0x1 << b] / denom;
    (*params)[i].pi_r = (*params)[i].pi[0] + (*params)[i].pi[1];
    (*params)[i].pi_y = (*params)[i].pi[2] + (*params)[i].pi[3];
    (*params)[i].pry = (*params)[i].pi_r / (*params)[i].pi_y;
    (*params)[i].pyr = (*params)[i].pi_y / (*params)[i].pi_r;
    if(model==HKY85) {
      (*params)[i].ttp = (*params)[i].kappa/2;
      (*params)[i].gamma = 1;
    }
    else if(model==F84) {
      (*params)[i].ttp = ((*params)[i].kappa + 2 * (*params)[i].pi_r 
		      * (*params)[i].pi_y) 
	/ (4 * (*params)[i].pi_r * (*params)[i].pi_y);
      (*params)[i].gamma = (((*params)[i].kappa + (*params)[i].pi_y) 
			* (*params)[i].pi_r) /
	(((*params)[i].kappa + (*params)[i].pi_r) * (*params)[i].pi_y);
    }
    else {/* TN93 */
      (*params)[i].ttp = initial_ttp[i];
      (*params)[i].gamma = initial_gamma[i];
    }
  }
}

/*
 *  find_unique_sites determines the unique sites
 *
 *  Sets Globals: num_unique_sites 
 *		  num_csites[] at the root
 *		  csites[][] at the root
 *		  lsites[][] at the root
 *
 *  Algorithm:	(1) Create sl, a site and an index by copying sites.
 *		(2) Use qsort (built in quick sort routine)
 *		    to sort sl so that all repeated sites are contiguous.
 *		(3) The first canonical site is the first site in the sorted
 *		    list.
 *		(4) If a site agrees with its predecesor
 *		    in the sorted list, it is not a canonical site.
 *		    The count for the corresponding canonical site
 *		    increases by one and lsites gives the same index 
 *		    in csites as the corresponding canonical site.
 *
 *		    Otherwise, the site is a canonical site.  The 
 *		    corresponding counter is set to 1, the site is stored 
 *		    in csites for the root, and lsites points to this 
 *		    location in csites.
 *		(5) The number of canonical sites at the root (the number of
 *		    unique sites) is stored.
 */

void find_unique_sites()
{
  int i,j,tag,first,last,b,num_nonconst;
  int *sl,*org_site,*base2;

  if(!(sl=(int *)malloc(sizeof(int)*num_all_sites*(site_len+2))))
    out_of_space("sl");
  if(!(org_site=(int *)malloc(sizeof(int)*num_all_sites)))
    out_of_space("org_site");
  if(!(base2=(int *)malloc(sizeof(int)*num_all_sites)))
    out_of_space("base2");

  num_unique_sites = 0;

  if(!(root_ksites=(int *)malloc(sizeof(int)*num_tags*NUM_BASES)))
    out_of_space("root_ksites");
  if(!(sitecount=(int *)malloc(sizeof(int)*num_all_sites)))
    out_of_space("sitecount");

  for(tag=0,first=first_site[0],last=first_site[1];
      tag<num_tags;
      tag++,first=last,last=first_site[tag+1]) {
    num_nonconst = 0;
    for(b=0;b<NUM_BASES;b++)
      ROOT_KSITES(tag)[b] = 0;
    for(i=first;i<last;i++)
      if((b=sitemember(SITES(i),constant_mask,NUM_BASES)) < NUM_BASES)
	ROOT_KSITES(tag)[b]++;
      else {
	copysite(NS_SITE(SL(num_nonconst)), SITES(i));
	NS_INDEX(SL(num_nonconst++)) = i;
      }

    if(num_nonconst>0) {
      qsort((void *)sl,num_nonconst,sizeof(int)*(site_len+2),cmp_nsites);
      copysite(SITES(num_unique_sites),NS_SITE(SL(0)));
      org_site[num_unique_sites] = NS_INDEX(SL(0));
      sitecount[num_unique_sites++] = 1;
      for(i=1;i<num_nonconst;i++)
	if (siteeq(NS_SITE(SL(i)), NS_SITE(SL(i-1))))
	  sitecount[num_unique_sites-1]++;
	else {
	  copysite(SITES(num_unique_sites),NS_SITE(SL(i)));
	  org_site[num_unique_sites] = NS_INDEX(SL(i));
	  sitecount[num_unique_sites++] = 1;
	}
    }
    first_site[tag+1] = num_unique_sites;
  }
  for(i=0;i<num_leaves;i++) {
    for(j=0;j<num_all_sites;j++)
      base2[j] = BASE(i)[j];
    for(j=0;j<num_unique_sites;j++)
      BASE(i)[j] = base2[org_site[j]];
  }
  free(sl);
  free(org_site);
  free(base2);
}

void generate_masks()
{
  int k,b,p,r,n;

  if(!(constant_mask=(int *)malloc(sizeof(int)*NUM_BASES*site_len)))
    out_of_space("constant_mask");
  for(b=0;b<NUM_BASES;b++) {
    for(p=0;p<site_len;p++)
      CONSTANT_MASK(b)[p] = 0;
    for(n=0;n<num_leaves;n++)
      add_base(CONSTANT_MASK(b),n,(0x1 << b));
  }
  if(!(mask=(int *)malloc(sizeof(int)*num_nodes*site_len)))
    out_of_space("mask");
  for(n=0;n<num_leaves;n++) {
    for(p=0;p<site_len;p++)
      MASK(n)[p] = 0;
    add_base(MASK(n),n,GAP);
  }
}

void restrict_sites_nonroot(const Tree *t, int n, int c)
{
  int i,j,b,num_canon=0,num_nonconst,tag,sym,cb;
  int *const_mask,*sl;

  if(!(const_mask=(int *)malloc(sizeof(int)*NUM_BASES*site_len)))
    out_of_space("const_mask");
  if(!(sl=(int *)malloc(sizeof(int)*num_all_sites*(site_len+2))))
    out_of_space("sl");
  for(b=0;b<NUM_BASES;b++)
    siteand(CONST_MASK(b), CONSTANT_MASK(b), MASK(c));

  for(tag=0;tag<num_tags;tag++) {
    num_nonconst = 0;
    for(i=FIRST_CSITE(n)[tag];i<FIRST_CSITE(n)[tag+1];i++) {
      j = CSITES(n)[i];
      if(LEAF(c))
	if((sym=BASE(c)[j]) == GAP)
	  LSITES(c)[i] = onefrag;
	else if(single_base(sym,&cb))
	  LSITES(c)[i] = KFRAG2(c,tag,cb);
	else
	  LSITES(c)[i] = UFRAG3(c,tag,sym);
      else  {
	siteand(NS_SITE(SL(num_nonconst)), SITES(j), MASK(c));
	if((b=sitemember(NS_SITE(SL(num_nonconst)),const_mask,NUM_BASES))
	   < NUM_BASES)
	  LSITES(c)[i] = KFRAG2(c,tag,b);
	else {
	  NS_N(SL(num_nonconst)) = i;
	  NS_INDEX(SL(num_nonconst++)) = j;
	}
      }
    }
    if(num_nonconst>0) {
      qsort((void *)sl,num_nonconst,sizeof(int)*(site_len+2),cmp_nsites);
      CSITES(c)[num_canon] = NS_INDEX(SL(0));
      LSITES(c)[NS_N(SL(0))] = FRAG2(c,num_canon++);
      for(i=1;i<num_nonconst;i++)
	if (siteeq(NS_SITE(SL(i)), NS_SITE(SL(i-1))))
	  LSITES(c)[NS_N(SL(i))] = LSITES(c)[NS_N(SL(i-1))];
	else {
	  CSITES(c)[num_canon] = NS_INDEX(SL(i));
	  LSITES(c)[NS_N(SL(i))] = FRAG2(c,num_canon++);
	}
    }
    FIRST_CSITE(c)[tag+1] = num_canon;
  }
  free(sl);
  free(const_mask);
}

void restrict_sites_root(const Tree *t, int c)
{
  int i,b,num_canon=0,num_nonconst,tag,sym,cb;
  int *const_mask,*sl;


  if(!(const_mask=(int *)malloc(sizeof(int)*NUM_BASES*site_len)))
    out_of_space("const_mask");
  if(!(sl=(int *)malloc(sizeof(int)*num_all_sites*(site_len+2))))
    out_of_space("sl");
  for(b=0;b<NUM_BASES;b++)
    siteand(CONST_MASK(b), CONSTANT_MASK(b), MASK(c));
  
  for(tag=0;tag<num_tags;tag++) {
    num_nonconst = 0;
    for(i=first_site[tag];i<first_site[tag+1];i++) {
      if(LEAF(c))
	if((sym=BASE(c)[i]) == GAP)
	  LSITES(c)[i] = onefrag;
	else if(single_base(sym,&cb))
	  LSITES(c)[i] = KFRAG2(c,tag,cb);
	else
	  LSITES(c)[i] = UFRAG3(c,tag,sym);
      else {
	siteand(NS_SITE(SL(num_nonconst)), SITES(i), MASK(c));
	if((b=sitemember(NS_SITE(SL(num_nonconst)),const_mask,NUM_BASES))
	   < NUM_BASES)
	  LSITES(c)[i] = KFRAG2(c,tag,b);
	else {
	  NS_N(SL(num_nonconst)) = i;
	  NS_INDEX(SL(num_nonconst++)) = i;
	}
      }
    }
    
    if(num_nonconst>0) {
      qsort((void *)sl,num_nonconst,sizeof(int)*(site_len+2),cmp_nsites);
      CSITES(c)[num_canon] = NS_INDEX(SL(0));
      LSITES(c)[NS_N(SL(0))] = FRAG2(c,num_canon++);
      for(i=1;i<num_nonconst;i++)
	if (siteeq(NS_SITE(SL(i)), NS_SITE(SL(i-1))))
	  LSITES(c)[NS_N(SL(i))] = LSITES(c)[NS_N(SL(i-1))];
	else {
	  CSITES(c)[num_canon] = NS_INDEX(SL(i));
	  LSITES(c)[NS_N(SL(i))] = FRAG2(c,num_canon++);
	}
    }
    FIRST_CSITE(c)[tag+1] = num_canon;
  }
  free(sl);
  free(const_mask);
}

void restrict_sites_node(const Tree *t, int n, int c)
{
  if(t->parent[c]==t->root)
    restrict_sites_root(t,c);
  else
    restrict_sites_nonroot(t,n,c);
}

void restrict_sites_all(const Tree *t,int n)
{
  /* n should not be t->root. */

  if(!LEAF(n)) {
    restrict_sites_nonroot(t,n,t->lchild[n]);
    restrict_sites_all(t,t->lchild[n]);
    restrict_sites_nonroot(t,n,t->rchild[n]);
    restrict_sites_all(t,t->rchild[n]);
  }
}

void restrict_sites_pos_root(const Tree *t, int subroot)
{
  if(subroot==t->root) {
    restrict_sites_root(t,t->lchild[subroot]);
    restrict_sites_all(t,t->lchild[subroot]);
    restrict_sites_root(t,t->rchild[subroot]);
    restrict_sites_all(t,t->rchild[subroot]);
  }
  else {
    if(t->parent[subroot]==t->root)
      restrict_sites_root(t,subroot);
    else
      restrict_sites_nonroot(t,t->parent[subroot],subroot);
    restrict_sites_all(t,subroot);
  }
}
    
/*
 * int parent_index(int i,Dlist *list, double *exht)
 * int leafparent(int i,Dlist *list, double *exht)
 * void remove_leaf(int i,int d,Dlist *list,int *tlchild,int *trchild)
 * void remake_tree(int *nodes,double *exht,int *tlchild,int *trchild)
 *
 * In combination, these functions put back together a tree after
 * its excursion has been modified.  nodes is input to remake_tree
 * containing the inorder traversal of the tree, and exht contains
 * the heights of the excursion after modification.  The new children
 * relationships are stored in tlchild and trchild.  nodes is changed
 * so that node[i] = j means that old node i is now labeled j.
 *
 */

#define NHGT(i)		(exht[list[(i)].node])

int parent_index(int i,Dlist *list, double *exht)
     /* Returns -2 if i is the root. Returns -1 if i is not a leaf.
	Otherwise, returns the index of i's parent in the linked list. */
{
  if(list[i].prev<0)
    if(list[i].next<0)
      return -2;
    else
      return (NHGT(i) >= NHGT(list[i].next) ? list[i].next : -1);
  if(list[i].next<0)
    return   (NHGT(i) >= NHGT(list[i].prev) ? list[i].prev : -1);
  if(NHGT(list[i].next) >= NHGT(list[i].prev))
    return   (NHGT(i) >= NHGT(list[i].next) ? list[i].next : -1);
  else
    return   (NHGT(i) >= NHGT(list[i].prev) ? list[i].prev : -1);
}

int leafparent(int i,Dlist *list, double *exht)
/* Simplified version of parent for the leaves of the original tree.
   These leaves are assumed to have parents (not true for a 1-node tree).
   Also, the leaves occur lower in the tree than any ancestor. */
{
  if(list[i].prev<0)
    return list[i].next;
  if(list[i].next<0)
    return list[i].prev;
  return (NHGT(list[i].next) > NHGT(list[i].prev) ? 
	  list[i].next : list[i].prev);
}

void remove_leaf(Tree *t, int i, int d, Dlist *list, double exht[])
/*  Removes i from the linked list. d is the parent of i. 
    Sets i to be the appropriate child of d. */
{
  if(list[i].prev>=0)
    list[list[i].prev].next = list[i].next;
  if(list[i].next>=0)
    list[list[i].next].prev = list[i].prev;
  t->branchlen[list[i].node] = exht[list[i].node]-exht[list[d].node];
  t->parent[list[i].node] = list[d].node;
  if(i>d)
    t->rchild[list[d].node] = list[i].node;
  else
    t->lchild[list[d].node] = list[i].node;
}

int remake_tree(Tree *t, Dlist list[], int count, double *exht)
     /* Change t->lchild[] and t->rchild to reflect the tree with heights
	exht[]. list is an linked list containing an inorder traversal of
	the tree and count is the number of nodes in the tree. */
{
  int i,j,d,new_subroot;

  for(i=0;i<count;i+=2)
  {
    remove_leaf(t,i,leafparent(i,list,exht),list,exht);
    list[i].done = 1;
  }

  for(i=1;i<count;i+=2)
  {
    if(!list[i].done)
    {
      for(j=i,d=parent_index(j,list,exht);
	  d>=0;
	  j=d,d=parent_index(j,list,exht))
      {
	remove_leaf(t,j,d,list,exht);
	list[j].done = 1;
      }
      if(d==-2) {
	new_subroot = list[j].node;
	list[j].done = 1;
      }
    }
  }
  return new_subroot;
}

void propagate_masks(const Tree *t, int n)
{
  if(!LEAF(n)) {
    int lc = t->lchild[n];
    int rc = t->rchild[n];
    
    propagate_masks(t,lc);
    propagate_masks(t,rc);
    siteor(MASK(n), MASK(lc), MASK(rc));
  }
}

/*
 *  find_partial_likelihoods finds the partial likelihood for each non-root
 *			     node and each canonical site.
 *  Receives:	 *t, a tree
 *		 *param, array of evolution parameters
 *  Sets Global: frag[][][] for each non-root node, canonical site of parent,
 *		   and possible base
 */

void find_parent_pl(const Tree *t, const Parameters params, 
			     int n, int *saved)
{ 
  int ls,rs,i,pb,b,k,tag,bs;
  double *ptab;
  double prod[NUM_BASES];

  if(!(ptab=(double *)malloc(sizeof(double)*num_tags*NUM_BASES*NUM_BASES)))
    out_of_space("ptab"); 

  ls = t->lchild[n];
  rs = t->rchild[n];

  find_ptable(ptab,t->branchlen[n],params);

  for(tag=0;tag<num_tags;tag++)
    for(bs=0;bs<NUM_BASES;bs++) {
      for(b=0;b<NUM_BASES;b++) {
	prod[b] = KFRAG2(ls,tag,bs)[b] * KFRAG2(rs,tag,bs)[b];
      }
      for(pb=0;pb<NUM_BASES;pb++) {
	SAVED_KFRAG2(n,tag,bs)[pb] = KFRAG2(n,tag,bs)[pb];
	KFRAG2(n,tag,bs)[pb] = 0;
	for(b=0;b<NUM_BASES;b++) {
	  KFRAG2(n,tag,bs)[pb] += PTAB(tag,b)[pb] * prod[b];
	}
      }
    }
  for(tag=0;tag<num_tags;tag++)
    for(i=FIRST_CSITE(n)[tag];i<FIRST_CSITE(n)[tag+1];i++) {
      for(b=0;b<NUM_BASES;b++)
	prod[b] = LSITES(ls)[i][b] * LSITES(rs)[i][b];
      for(pb=0;pb<NUM_BASES;pb++) {
	SAVED_FRAG2(n,i)[pb] = FRAG2(n,i)[pb];
	FRAG2(n,i)[pb] = 0;
	for(b=0;b<NUM_BASES;b++) 
	  FRAG2(n,i)[pb] += PTAB(tag,b)[pb] * prod[b];
      }
    }
  *saved = FIRST_CSITE(n)[num_tags];
  free(ptab);
}

void copy_kFRAG2(double *kfragx, double *kfragy)
{
  int pb,tag,bs;

  for(tag=0;tag<num_tags;tag++)
    for(bs=0;bs<NUM_BASES;bs++)
      for(pb=0;pb<NUM_BASES;pb++)
	kfragx[(tag*NUM_BASES+bs)*NUM_BASES+pb] =
          kfragy[(tag*NUM_BASES+bs)*NUM_BASES+pb];
}

void copy_uFRAG2(int n, double *ufragx, double *ufragy)
{
  int i,pb,tag,sym;

  for(i=0;i<num_uncertain[n];i++) {
    sym = UNCERTAIN_BASE(n)[i];
    tag = UNCERTAIN_TAG(n)[i];
      for(pb=0;pb<NUM_BASES;pb++)
	ufragx[(tag*NUM_SYMBOLS+sym)*NUM_BASES+pb] = 
	  ufragy[(tag*NUM_SYMBOLS+sym)*NUM_BASES+pb];
  }
}

void copy_FRAG2(int last, double *fragx, double *fragy)
{
  int i,pb;

  for(i=0;i<last;i++)
    for(pb=0;pb<NUM_BASES;pb++)
      fragx[i*NUM_BASES+pb] = fragy[i*NUM_BASES+pb];
}

void restore_parent_pl(int n, int nvs)
{
  copy_kFRAG2(KFRAG(n),SAVED_KFRAG(n));
  copy_FRAG2(nvs,FRAG(n),SAVED_FRAG(n));
}

void find_leaf_pl(const Tree *t, const Parameters params, int n, int *saved)
{
  int i,k,tag,pb,p,m;

  copy_kFRAG2(SAVED_KFRAG(n),KFRAG(n));
  copy_uFRAG2(n,SAVED_UFRAG(n),UFRAG(n));
  find_ptable(KFRAG(n),t->branchlen[n],params);
  for(i=0;i<num_uncertain[n];i++) {
    k = UNCERTAIN_BASE(n)[i];
    tag = UNCERTAIN_TAG(n)[i];
    for(pb=0;pb<NUM_BASES;pb++) {
      UFRAG3(n,tag,k)[pb] = 0;
      for(p=0,m=0x1;p<NUM_BASES;p++,m=(m<<1))
	if(k & m)
	  UFRAG3(n,tag,k)[pb] += KFRAG2(n,tag,p)[pb];
    }
  }
  *saved = 0;
}
    
void find_node_pl(const Tree *t, const Parameters params, int n,
			   int *saved)
{
  if(LEAF(n)) 
    find_leaf_pl(t,params,n,saved);
  else
    find_parent_pl(t,params,n,saved);
}
    
void restore_node_pl(int n, int nvs)
{
  if(nvs!=-1) {
    copy_kFRAG2(KFRAG(n),SAVED_KFRAG(n));
    if(LEAF(n))
      copy_uFRAG2(n,UFRAG(n),SAVED_UFRAG(n));
    else
      copy_FRAG2(nvs,FRAG(n),SAVED_FRAG(n));
  }
}

void find_pl_tree(const Tree *t, const Parameters params, int n,
		  int *nvs)
{
  if(LEAF(n))
    find_leaf_pl(t,params,n,nvs);
  else {
    find_pl_tree(t,params,t->lchild[n],nvs);
    find_pl_tree(t,params,t->rchild[n],nvs);
    find_parent_pl(t,params,n,&nvs[n]);
  }
}

void restore_pl_tree(const Tree *t, int n, int *nvs)
{
  copy_kFRAG2(KFRAG(n),SAVED_KFRAG(n));
  if(LEAF(n)) 
    copy_uFRAG2(n,UFRAG(n),SAVED_UFRAG(n));
  else {
    copy_FRAG2(nvs[n],FRAG(n),SAVED_FRAG(n));
    restore_pl_tree(t,t->lchild[n],nvs);
    restore_pl_tree(t,t->rchild[n],nvs);
  }
}

void restore_pl_treex(const Tree *t, int prop_root, int n, int *nvs)
{
  /* This is called when the node prop_root never had its kfrag and frag
     arrays saved, i.e, a global update was done and prop_root became
     the new root. */

  if(LEAF(n)) {
    copy_kFRAG2(KFRAG(n),SAVED_KFRAG(n));
    copy_uFRAG2(n,UFRAG(n),SAVED_UFRAG(n));
  }
  else {
    if(n==prop_root) {
      restore_pl_tree(t,t->lchild[n],nvs);
      restore_pl_tree(t,t->rchild[n],nvs);
    }
    else {
      copy_kFRAG2(KFRAG(n),SAVED_KFRAG(n));
      copy_FRAG2(nvs[n],FRAG(n),SAVED_FRAG(n));
      restore_pl_treex(t,prop_root,t->lchild[n],nvs);
      restore_pl_treex(t,prop_root,t->rchild[n],nvs);
    }
  }
}

void find_partial_likelihoods(const Tree *t, const Parameters params,
			      int subroot, int *nvs)
{
  int n;

  propagate_masks(t,subroot);
  restrict_sites_pos_root(t,subroot);
  if(!LEAF(subroot)) {
    find_pl_tree(t,params,t->lchild[subroot],nvs); 
    find_pl_tree(t,params,t->rchild[subroot],nvs);
  }
  if(subroot != t->root) {
    find_parent_pl(t,params,subroot,&nvs[subroot]);
    for(n=t->parent[subroot];n!=t->root;n=t->parent[n]) {
      find_parent_pl(t,params,n,&nvs[n]);
    }
  }
}

void restore_partial_likelihoods(const Tree *t, int prop_root,
				 int subroot, int *nvs)
{
  int n;

  propagate_masks(t,subroot);
  restrict_sites_pos_root(t,subroot);

  if(!LEAF(subroot)) 
    if(subroot == t->root) {
      restore_pl_treex(t,prop_root,t->lchild[subroot],nvs); 
      restore_pl_treex(t,prop_root,t->rchild[subroot],nvs); 
    }
    else {
      restore_pl_tree(t,t->lchild[subroot],nvs); 
      restore_pl_tree(t,t->rchild[subroot],nvs); 
    }
  if(subroot != t->root) {
    restore_parent_pl(subroot,nvs[subroot]);
    for(n=t->parent[subroot];n!=t->root;n=t->parent[n])
      restore_parent_pl(n,nvs[n]);
  }
}

/*
 *  find_loglikelihood finds the log of the likelihood probability of a tree
 *  
 *  Receives:  *t, a tree
 *	       *param, parameters
 *  Returns:   The loglikelihood probability of the tree
 *
 *   t->loglikelihood is set
 *
 *  Algorithm:
 */


void find_root_ll(Tree *t,const Parameters params)
{
  int b,k,ls,rs,i,tag,bs;
  double isum,osum;

  ls = t->lchild[t->root];
  rs = t->rchild[t->root];
  osum = 0;
  for(tag=0;tag<num_tags;tag++)
    for(bs=0;bs<NUM_BASES;bs++) {
      isum = 0;
      for(b=0;b<NUM_BASES;b++)
	isum += params[tag].pi[b]*KFRAG2(ls,tag,bs)[b] *
	  KFRAG2(rs,tag,bs)[b];
      osum += ROOT_KSITES(tag)[bs] * log(isum);
    }
  for(tag=0;tag<num_tags;tag++)
    for(i=first_site[tag];i<first_site[tag+1];i++) {
      isum = 0;
      for(b=0;b<NUM_BASES;b++)
	isum += params[tag].pi[b] * LSITES(ls)[i][b] 
	  * LSITES(rs)[i][b];
      osum += sitecount[i] * log(isum);
    }
  t->loglikelihood = osum;
}

void find_loglikelihood(Tree *t, const Parameters params,
			 Saved_state ss)
{
  find_pl_tree(t,params,t->lchild[t->root],ss->nvs); 
  find_pl_tree(t,params,t->rchild[t->root],ss->nvs);
  ss->loglikelihood = t->loglikelihood;
  find_root_ll(t,params);
}

void restore_loglikelihood(Tree *t, Saved_state ss)
{
  restore_pl_tree(t,t->lchild[t->root],ss->nvs); 
  restore_pl_tree(t,t->rchild[t->root],ss->nvs); 
  t->loglikelihood = ss->loglikelihood;
}

void upcase_string(char *p)
{
  for(;*p;p++)
    *p = toupper(*p);
}

void create_initial_tree(Tree *t, const Parameters params)
{
  int i;
  double *exht;
  Dlist *list;
  char tree_str[MAX_LINE];
  int *nvs,*perm;

  if(!(exht=(double *)malloc(sizeof(double)*num_nodes)))
    out_of_space("exht"); 
  if(!(list=(Dlist *)malloc(sizeof(struct dlist)*num_nodes)))
    out_of_space("list"); 
  if(!(nvs=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("nvs"); 
  if(!(perm=(int *)malloc(sizeof(int)*num_leaves)))
    out_of_space("perm"); 

  if(!(t->lchild=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("t->lchild");
  if(!(t->rchild=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("t->rchild");
  if(!(t->parent=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("t->parent");
  if(!(t->branchlen=(double *)malloc(sizeof(double)*num_nodes)))
    out_of_space("t->branchlen");

  for(i=0;i<num_nodes;i++)
    FIRST_CSITE(i)[0] = 0;
  strcpy(tree_str,tree_file);
  upcase_string(tree_str);
  if (strcmp(tree_str,"RANDOM")==0)
  {
    if(newick_format) 
      for(i=0;i<num_leaves-1;i++)
	if(strcmp(ordered_names[i].name,ordered_names[i+1].name)==0)
	  fprintf(stderr,"Warning: Taxa have identical names (%s). Program "
		  "will not be able to read back in final tree.\n",
		  ordered_names[i].name);
    for(i=0;i<num_leaves;i++) {
      exht[i] = max_depth;
      t->lchild[i] = t->rchild[i] = -1;
    }

    /*  Choose random valley depths. */

    for(i=num_leaves;i<num_nodes;i++)
      exht[i] = max_depth * runif();

    find_permutation(perm,num_leaves);

    for(i=0;i<num_nodes;i++) {
      list[i].node = (i%2 ? num_leaves+(i-1)/2 : perm[i/2]);
      list[i].prev = i-1;
      list[i].next = i+1;
      list[i].done = 0;
    }
    list[num_nodes-1].next = -1;

    t->root = remake_tree(t,list,num_nodes,exht);
    t->branchlen[t->root] = 0;
    t->parent[t->root] = -1;
  }
  else
    read_initial_tree(t);

  propagate_masks(t,t->root);
  restrict_sites_pos_root(t,t->root);
  find_pl_tree(t,params,t->lchild[t->root],nvs);
  find_pl_tree(t,params,t->rchild[t->root],nvs);
  find_root_ll(t,params);
  fprint_tree(stdout,t);
  free(perm);
  free(nvs);
  free(list);
  free(exht);
}

double hastings_pi(const double c,const double x[],const double y[],int k)
/* c is the constant where y ~ Dirichlet(cx)
   x is the current state (assumed that sum x = 1)
   y is the proposed state (assumed that sum y = 1)
   k is the length of each array
   return q(y,x) / q(x,y)
*/
{
  double lhr = 0.0,cx,cy;
  int i;
  
  for(i=0;i<k;i++)
  {
    cx = c*x[i];
    cy = c*y[i];
    lhr += (cy-1)*log(x[i]) - (cx-1)*log(y[i]) - lgamma(cy) + lgamma(cx);
  }
  return SAFE_EXP(lhr);
}

double hastings_theta(double c,const double w[],const Parameters x,
                      const Parameters y,int k)
/* 
   c is the constant where w*y ~ Dirichlet(c(w*x))
   w is the array of weights such that sum wx = sum wy = 1
     (it is also true that sum w = 1, but this isn't necessary)
   x is the current state
   y is the proposed state
   k is the length of each array
   return q(y,x) / q(x,y)
*/
{
  double lhr = 0.0,wx,wy,cwx,cwy;
  int i;
  
  for(i=0;i<k;i++)
  {
    wx = w[i]*x[i].theta;
    wy = w[i]*y[i].theta;
    cwx = c*wx;
    cwy = c*wy;
    lhr += (cwy-1)*log(wx) - (cwx-1)*log(wy) - lgamma(cwy) + lgamma(cwx);
  }
  return SAFE_EXP(lhr);
}

void update_params(Parameters newparams, const Parameters params)
{
  int i,b;
  double *alpha1,alpha2[NUM_BASES],*newtheta,newpi[NUM_BASES];

  if(!(alpha1=(double *)malloc(sizeof(double)*num_tags)))
    out_of_space("alpha1"); 
  if(!(newtheta=(double *)malloc(sizeof(double)*num_tags)))
    out_of_space("newtheta"); 

  if(update_kappa)
  {
    newparams[0].kappa = fabs(params[0].kappa + (2*runif()-1)*kappa_win);
    for(i=1;i<num_tags;i++) 
      if(single_kappa)
        newparams[i].kappa = newparams[0].kappa;
      else
        newparams[i].kappa = fabs(params[i].kappa + (2*runif()-1)*kappa_win);
  }
  else
    for(i=0;i<num_tags;i++)
      newparams[i].kappa = params[i].kappa;
  
  if(update_theta)
  {
    for(i=0;i<num_tags;i++)
      alpha1[i] = theta_const * params[i].theta * tag_weights[i];
    rdirich(alpha1,newtheta,num_tags);
    for(i=0;i<num_tags;i++)
      newparams[i].theta = newtheta[i] / tag_weights[i];
  }
  else
    for(i=0;i<num_tags;i++)
      newparams[i].theta = params[i].theta;
  
  if(update_pi)
  {
    for(i=0;i<num_tags;i++)
    {
      for(b=0;b<NUM_BASES;b++)
        alpha2[b] = pi_const * params[i].pi[b];
      rdirich(alpha2,newpi,NUM_BASES);
      for(b=0;b<NUM_BASES;b++)
        newparams[i].pi[b] = newpi[b];
      newparams[i].pi_r = newparams[i].pi[0] + newparams[i].pi[1];
      newparams[i].pi_y = newparams[i].pi[2] + newparams[i].pi[3];
      newparams[i].pry = newparams[i].pi_r / newparams[i].pi_y;
      newparams[i].pyr = newparams[i].pi_y / newparams[i].pi_r;
    }
  }
  else
  {
    for(i=0;i<num_tags;i++)
    {
      for(b=0;b<NUM_BASES;b++)
        newparams[i].pi[b] = params[i].pi[b];
      newparams[i].pi_r = newparams[i].pi[0] + newparams[i].pi[1];
      newparams[i].pi_y = newparams[i].pi[2] + newparams[i].pi[3];
      newparams[i].pry = newparams[i].pi_r / newparams[i].pi_y;
      newparams[i].pyr = newparams[i].pi_y / newparams[i].pi_r;
    }
  }
  
  for(i=0;i<num_tags;i++) {
    if(model==HKY85) {
      newparams[i].ttp = newparams[i].kappa/2;
      newparams[i].gamma = 1;
    }
    else if(model==F84) {
      newparams[i].ttp = (newparams[i].kappa + 2 * newparams[i].pi_r 
		      * newparams[i].pi_y) 
	/ (4 * newparams[i].pi_r * newparams[i].pi_y);
      newparams[i].gamma = ((newparams[i].kappa + newparams[i].pi_y) 
			* newparams[i].pi_r) /
	((newparams[i].kappa + newparams[i].pi_r) * newparams[i].pi_y);
    }
    /*else  model == TN93
    {
      ;
    }*/
  }
  free(newtheta);
  free(alpha1);
}

void copy_params(Parameters params1, const Parameters params2)
{
  int i,b;

  for(i=0;i<num_tags;i++) {
    params1[i].kappa = params2[i].kappa;
    params1[i].theta = params2[i].theta;
    params1[i].ttp = params2[i].ttp;
    params1[i].gamma = params2[i].gamma;
    for(b=0;b<NUM_BASES;b++)
      params1[i].pi[b] = params2[i].pi[b];
    params1[i].pi_r = params2[i].pi_r;
    params1[i].pi_y = params2[i].pi_y;
    params1[i].pry =  params2[i].pry;
    params1[i].pyr =  params2[i].pyr;
  }
}

/*
 *  void swap_children(Tree *t)
 *
 *  swap_children changes the left/right orientation of internal node
 *  with probability 0.5.
 *
 *  Upon exiting: t->lchild and t->rchild have been modified
 */

void swap_children(Tree *t,int n)
     /* Randomly swaps the children of the tree rooted at n. */
{
  int temp;

  if(!LEAF(n)) {
    if(runif()<.5)
      SWAP(temp,t->lchild[n],t->rchild[n]);
    swap_children(t,t->lchild[n]);
    swap_children(t,t->rchild[n]);
  }
}

/* 
 * jiggle_heights adds a uniform perturbation to each valley depth
 *
 *  Upon exiting: exht[i] contains the height of node i
 *		  The height of the root will (a.s.) not be 0
 *		  Differences will be accurate.
 *		  All exht[i] are off by a common offset.
 */

void jiggle_heights_int(const Tree *t, double *exht, int i,
			const double subheight, int bounced[],
			int *count)
{
  int n;

  if(!LEAF(i)) {
    exht[i] += (2*runif()-1)*valley_win;
    n = (int)(exht[i]/subheight);
    if(exht[i]<0)
      n--;
    if((n>=0?n:1-n)/2)
      bounced[(*count)++] = i;
    exht[i] = n%2 ? (n+1)*subheight-exht[i] : exht[i]-n*subheight;
    jiggle_heights_int(t,exht,t->lchild[i],subheight,bounced,count);
    jiggle_heights_int(t,exht,t->rchild[i],subheight,bounced,count);
  }
}
    
void jiggle_heights_nonroot(const Tree *t, double *exht, int subroot,
			    const double subheight)
{
  int *bounced, *perm, count = 0, i;
  double *exht2;
  int x;  
  
  if(!(bounced=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("bounced"); 
  if(!(perm=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("perm"); 
  if(!(exht2=(double *)malloc(sizeof(double)*num_nodes)))
    out_of_space("exht2"); 
  if(!LEAF(subroot)) {
    jiggle_heights_int(t,exht,t->lchild[subroot],subheight,bounced,&count);
    jiggle_heights_int(t,exht,t->rchild[subroot],subheight,bounced,&count);
  }
  if(count>0) {
    for(i=0;i<count;i++)
      exht2[bounced[i]] = exht[bounced[i]];
    bounced[count++] = subroot;
    exht2[subroot] = 0;
    find_permutation(perm,count);
    for(i=0;i<count;i++)
      exht[bounced[i]] = exht2[bounced[perm[i]]];
    }
  free(exht2);
  free(perm);
  free(bounced);
}

void jiggle_heights_root(const Tree *t, double *exht, int i,
			 const double height)
{
  if(!LEAF(i)) {
    exht[i] += (2*runif()-1)*valley_win;
    if(exht[i] > height)
      exht[i] = 2*height-exht[i];
    jiggle_heights_root(t,exht,t->lchild[i],height);
    jiggle_heights_root(t,exht,t->rchild[i],height);
  }
}

void jiggle_heights_nc(const Tree *t, double *exht, int subroot,
		       Dlist list[], int count)
{
  /* Jiggles the heights of the nodes of tree rooted at subroot for the
     non-clock case. Changes exht. Assumes the list is actually
     stored in inorder numbering. */

  int i;     /* Interior node in the inorder traversal to be processed */
  double d1; /* New distance (hgt(i-1)-hgt(i)) */
  double d2; /* New distance (hgt(i+1)-hgt(i)) */
  double ht; /* New height of the leaf before i */
  int j;

  ht = exht[list[0].node];
  for(i=1;i<count;i+=2) {
    d1 = fabs(exht[list[i-1].node]-exht[list[i].node]
	      +(2*runif()-1)*valley_win);
    d2 = fabs(exht[list[i+1].node]-exht[list[i].node]
	      +(2*runif()-1)*valley_win);
    exht[list[i-1].node] = ht;
    exht[list[i].node] = ht - d1;
    ht += d2 - d1;
  } 
  exht[list[count-1].node] = ht;
}
  
void jiggle_heights(const Tree *t, double *exht, int subroot,
		    const double subheight, Dlist list[], int count)
{
  /* subroot is the root of the affected tree. Subheight is its height.
     Changes exht to reflect random changes to the heights of the
     interior nodes. If affected tree is not the whole tree, its overall 
     height is not changed. */

  if(!mclock)
    jiggle_heights_nc(t,exht,subroot,list,count);
  else if(subroot == t->root)
    jiggle_heights_root(t,exht,t->root,subheight);
  else
    jiggle_heights_nonroot(t,exht,subroot,subheight);
}

void list_nodes_tree(const Tree *t, int i, Dlist list[], int *count)
{
  if(!LEAF(i))
    list_nodes_tree(t,t->lchild[i],list,count);
  list[*count].node = i;
  list[*count].prev = (*count)-1;
  list[*count].next = (*count)+1;
  list[*count].done = 0;
  (*count)++;
  if(!LEAF(i))
    list_nodes_tree(t,t->rchild[i],list,count);
}

void list_nodes(const Tree *t, int subroot, Dlist list[], int *count,
		double exht[])
{
  /* Returns list, a linked list containing an inorder traversal of
     the tree rooted at subroot, count, the number of nodes in that
     tree, and exht, randomly changed heights of the nodes. */

  double subheight;
  
  find_heights(t,exht,subroot,0,&subheight);
  *count = 0;
  list_nodes_tree(t,subroot,list,count);
  list[(*count)-1].next = -1;
  jiggle_heights(t,exht,subroot,subheight,list,*count); 
}

void regional_update(Tree *t, int subroot,
		     Saved_state ss, const Parameters params) 
{
  Dlist *list;
  int count,new_subroot;
  double *exht,subroot_branchlen;
  int subroot_parent;

  if(!(list=(Dlist *)malloc(sizeof(struct dlist)*num_nodes)))
    out_of_space("list"); 
  if(!(exht=(double *)malloc(sizeof(double)*num_nodes)))
    out_of_space("exht"); 

  copy_subtree_topology(ss->t,subroot,t);
  ss->n = subroot;
  swap_children(t,subroot);
  list_nodes(t,subroot,list,&count,exht);
  subroot_branchlen = t->branchlen[subroot];
  subroot_parent = t->parent[subroot];
  new_subroot = remake_tree(t,list,count,exht);
  t->parent[new_subroot] = subroot_parent;
  t->branchlen[new_subroot] = subroot_branchlen;
  /* The ancestors of the subroot are not changed by local_update.
     However, since the subroot may be different, we must change
     its parent pointer. */
  if(subroot == t->root) {
    t->root = new_subroot;
    t->parent[t->root] = -1;
  }
  else
    if(subroot == t->lchild[subroot_parent])
      t->lchild[subroot_parent] = new_subroot;
    else 
      t->rchild[subroot_parent] = new_subroot;
  find_partial_likelihoods(t,params,new_subroot,ss->nvs);
  find_root_ll(t,params);
  free(exht);
  free(list);
}

void regional_restore(Tree *t, Saved_state ss) 
{
  int prop_root = t->root;

  copy_subtree_topology(t,ss->n,ss->t);
  restore_partial_likelihoods(t,prop_root,ss->n,ss->nvs);
}

void global_update(Tree *t, Saved_state ss, 
		   const Parameters params) 
{
  int new_rootchild,s,*ancestors,num_ancest=0,i,j;
  Tree *t2;
  int minnode, maxnode, midnode;
  int *nvs;

  if(!mclock && runif() < REROOT_PROB) {
    if(!(ancestors=(int *)malloc(sizeof(int)*num_nodes)))
      out_of_space("ancestors"); 
    t2 = make_tree();
    if(!(nvs=(int *)malloc(sizeof(int)*num_nodes)))
      out_of_space("nvs"); 

    /* Pick a node other than the root or its children. */
    new_rootchild = (int)(runif()*(num_nodes-3));
    minnode = MINIMUM(t->root,MINIMUM(t->lchild[t->root],t->rchild[t->root]));
    maxnode = MAXIMUM(t->root,MAXIMUM(t->lchild[t->root],t->rchild[t->root]));
    midnode = t->root + t->lchild[t->root] + t->rchild[t->root] 
      - minnode - maxnode;
    if(new_rootchild >= minnode)
      new_rootchild++;
    if(new_rootchild >= midnode)
      new_rootchild++;
    if(new_rootchild >= maxnode)
      new_rootchild++;
    copy_subtree_topology(t2,t->root,t);
    s = split_branch(t,new_rootchild);
    for(i=t->parent[s];i!=-1;i=t->parent[i]) {
      ancestors[num_ancest++] = i;
      siteor(MASK(i), MASK(t->lchild[i]), MASK(t->rchild[i]));
    }
    restrict_sites_root(t,t->lchild[t->root]);
    restrict_sites_root(t,t->rchild[t->root]);
    for(j=num_ancest-2;j>=0;j--) {
      restrict_sites_nonroot(t,ancestors[j],t->lchild[ancestors[j]]);
      restrict_sites_nonroot(t,ancestors[j],t->rchild[ancestors[j]]);
    }
    if(!LEAF(s)) {
      restrict_sites_nonroot(t,s,t->lchild[s]);
      restrict_sites_nonroot(t,s,t->rchild[s]);
    }
    for(j=0;j<num_ancest;j++) {
      find_node_pl(t, params, t->lchild[ancestors[j]], 
		   &(ss->nvs[t->lchild[ancestors[j]]]));
      find_node_pl(t, params, t->rchild[ancestors[j]], 
		   &(ss->nvs[t->rchild[ancestors[j]]]));
    }
    find_root_ll(t,params);
    free(ancestors);
    free_tree(t2);
    free(nvs);
  }
  else 
    regional_update(t,t->root,ss,params);
}

void global_restore(Tree *t, Saved_state ss) 
{
  regional_restore(t,ss);
}

Relation opposite(Relation r)
{
  /* Returns the relation opposite to r. */

  switch (r) {
  case LEFTCHILD:  return PARENTL;
  case RIGHTCHILD: return PARENTR;
  case PARENTL:    return LEFTCHILD;
  case PARENTR:    return RIGHTCHILD;
  case ROOTLEFT:   return ROOTRIGHT;
  case ROOTRIGHT:  return ROOTLEFT;
  default:	   return PARENTL;
  }
}

void nonchild_neighbor(const Tree *t, int x, Neighbor *n)
{
  /* Sets n to be the non-child neighbor of x in t. */

  if(mclock || t->parent[x] != t->root) {
    n->nd = t->parent[x]; 
    n->rel = (t->lchild[n->nd] == x ? PARENTL : PARENTR);
  }
  else if(t->lchild[t->root] == x) {
    n->nd = t->rchild[t->root]; 
    n->rel = ROOTRIGHT;
  }
  else {
    n->nd = t->lchild[t->root]; 
    n->rel = ROOTLEFT;
  }
}

void neighbors(const Tree *t, int x, Neighbor *n1, Neighbor *n2, 
	       Neighbor *n3)
{
  /* n cannot be the root of the tree. n3 is always the non-child 
     neighbor. */

  n1->nd = t->lchild[x]; 
  n1->rel = LEFTCHILD;
  n2->nd = t->rchild[x]; 
  n2->rel = RIGHTCHILD;
  nonchild_neighbor(t,x,n3);
}

void other_neighbors(const Tree *t, int x, Relation r,
		     Neighbor *n2, Neighbor *n3)
{
  /* Returns the two other neighbors of x, given that one neighbor
     is in relation r to it.
     If non-clock, n cannot be the root of the tree. */

  switch (r) {
  case LEFTCHILD: 
    n2->nd = t->rchild[x]; 
    n2->rel = RIGHTCHILD;
    nonchild_neighbor(t,x,n3);
    break;
  case RIGHTCHILD:
    n2->nd = t->lchild[x]; 
    n2->rel = LEFTCHILD;
    nonchild_neighbor(t,x,n3);
    break;
  case PARENTL: case PARENTR: case ROOTRIGHT: case ROOTLEFT:
    n2->nd = t->lchild[x]; 
    n2->rel = LEFTCHILD;
    n3->nd = t->rchild[x]; 
    n3->rel = RIGHTCHILD;
  }
}

double get_edge_length(const Tree *t, int x, Neighbor n)
{
  /* returns the length of the edge between node x and its neighbor n. */

  switch (n.rel) {
  case LEFTCHILD: case RIGHTCHILD: return t->branchlen[n.nd];
  case PARENTL: case PARENTR: return t->branchlen[x];
  case ROOTLEFT: case ROOTRIGHT: return t->branchlen[x]+t->branchlen[n.nd];
  default:       return 1.0;
  }
}
    
void set_edge_length(Tree *t, int x, Neighbor n, double m)
{
  /* sets the length of the edge between node x and its neighbor n to m. */

  switch (n.rel) {
  case LEFTCHILD: case RIGHTCHILD: t->branchlen[n.nd] = m; break;
  case PARENTL: case PARENTR:  t->branchlen[x] = m; break;
  case ROOTLEFT: case ROOTRIGHT: 
    t->branchlen[x] = t->branchlen[n.nd] = m/2;
    break;
  }
}
    
void set_neighbor(Tree *t, int a, Relation r, int b)
{
  /* sets b to be a's neighbor as indicated by r. */
  
  switch (r) {
  case LEFTCHILD:
    t->lchild[a] = b;
    t->parent[b] = a;
    break;
  case RIGHTCHILD:
    t->rchild[a] = b;
    t->parent[b] = a;
    break;
   case PARENTL: 
    t->lchild[b] = a;
    t->parent[a] = b;
    break;
  case PARENTR: 
    t->rchild[b] = a;
    t->parent[a] = b;
    break;
  case ROOTLEFT:
    t->lchild[t->root] = b;
    t->parent[b] = t->root;
    break;
  case ROOTRIGHT:
    t->rchild[t->root] = b;
    t->parent[b] = t->root;
    break;
  }
}

typedef struct {
  Neighbor nb;
  double ht;
} Child;
    
int cmp_heights(const void *n1,const void *n2)
{  
  double c =((Child *)n1)->ht - ((Child *)n2)->ht;

  return (c<0 ? -1 : (c>0 ? 1 : 0));
}

void local_update(Tree *t, int n, Saved_state ss,
		   const Parameters params,
		   double *hr)
{
 /* 
  * ---------------------------------------------------------------
  * Update does not depend on rooting of the tree.
  * n is the child node of a randomly chosen internal edge.
  * The tree is locally changed, keeping the sum of all branch 
  * lengths constant.
  * if n is a child of the root, e is the other child of the root.
  * In the unrooted tree, randomly move n and e between a and c.
  *
  *        b  d
  *        |  |
  *     a--n--e--c
  *
  * The two children of n are randomly assigned to be a and b.
  * If e is the parent of n, n's sibling and grandparent are randomly
  * chosen to be c and d.
  * If e is n's sibling (because n's parent is the root), the sibling's
  * two children are randomly chosen to be c and d.
  * The two nodes are conceptually equivalent in the unrooted tree.
  * If n is randomly set below e in the stored rooted tree, relabel them
  * to make the code for updating masks easier.
  */

  Neighbor temp;
  Child child[3];
  double m, old_m, x, y, h, tempd, *hgt, min_h;
  double old_x, old_y, mratio;
  int g, p, min_child;
  int i,b,pb;

  if(!(hgt=(double *)malloc(sizeof(double)*num_nodes)))
    out_of_space("hgt"); 

  /* need to fix saved state information */
  ss->n = n;
  ss->loglikelihood = t->loglikelihood;
  neighbors(t,n,&(ss->a),&(ss->b),&(ss->e));
  ss->t->branchlen[n] = t->branchlen[n];
  ss->t->branchlen[ss->a.nd] = t->branchlen[ss->a.nd];
  ss->t->branchlen[ss->b.nd] = t->branchlen[ss->b.nd];
  ss->t->branchlen[ss->e.nd] = t->branchlen[ss->e.nd];

  if(mclock) {
    if(ss->e.rel == PARENTL) {
      ss->c.nd = t->rchild[ss->e.nd];
      ss->c.rel = RIGHTCHILD;
    }
    else { /* ss->e.rel == PARENTR */
      ss->c.nd = t->lchild[ss->e.nd];
      ss->c.rel = LEFTCHILD;
    }
    ss->t->branchlen[ss->c.nd] = t->branchlen[ss->c.nd];
    h = (ss->e.nd == t->root ? 0 : t->branchlen[ss->e.nd]);
    child[0].nb = ss->a;
    child[0].ht = hgt[ss->a.nd] = t->branchlen[ss->a.nd] + t->branchlen[n] + h;
    child[1].nb = ss->b;
    child[1].ht = hgt[ss->b.nd] = t->branchlen[ss->b.nd] + t->branchlen[n] + h;
    child[2].nb = ss->c;
    child[2].ht = hgt[ss->c.nd] = t->branchlen[ss->c.nd] + h;
    qsort(child,3,sizeof(Child),cmp_heights);
    min_child = child[0].nb.nd;
    min_h = hgt[min_child];
    if(ss->e.nd==t->root) {
      hgt[min_child] *=  SAFE_EXP(LAMBDA2 * (runif() - 0.5));
      hgt[child[1].nb.nd] += hgt[min_child] - min_h;
      hgt[child[2].nb.nd] += hgt[min_child] - min_h;
      y = 0;
    }
    else
      y = runif() * min_h;

    x = runif() * hgt[child[1].nb.nd];
    if(x<y)
      SWAP(tempd,x,y);
    if(x>hgt[min_child]) {
      ss->d = child[0].nb; /* ss->d is the child to be closest to the gp. */
      *hr = (t->branchlen[n]+h>min_h ? 1.0 : 1.0/3.0);
    }
    else {
      ss->d = child[(int)(runif()*3)].nb;
      *hr = (t->branchlen[n]+h>min_h ? 3.0 : 1.0);
    }
    t->branchlen[n] = x - y;
    t->branchlen[ss->e.nd] = y;
    if(ss->e.nd==t->root)
      *hr *= (hgt[min_child] * hgt[min_child]) / (min_h * min_h);
    if(ss->d.nd == ss->c.nd) { /* Topology does not change. */
      ss->rotated = 0;
      t->branchlen[ss->a.nd] = hgt[ss->a.nd] - x;
      t->branchlen[ss->b.nd] = hgt[ss->b.nd] - x;
      t->branchlen[ss->c.nd] = hgt[ss->c.nd] - y;
    }
    else { /* Topology changes */
      ss->rotated = 1;
      g = (ss->d.nd == ss->a.nd ? ss->b.nd : ss-> a.nd);
      set_neighbor(t,n,ss->d.rel,ss->c.nd);
      set_neighbor(t,ss->e.nd,ss->c.rel,ss->d.nd);
      t->branchlen[g] = hgt[g] - x;
      t->branchlen[ss->c.nd] = hgt[ss->c.nd] - x;
      t->branchlen[ss->d.nd] = hgt[ss->d.nd] - y;
      siteor(MASK(n), MASK(g), MASK(ss->c.nd));
      restrict_sites_node(t,ss->e.nd,n);
      restrict_sites_node(t,ss->e.nd,ss->d.nd);
      restrict_sites_nonroot(t,n,ss->c.nd);	
      restrict_sites_nonroot(t,n,g);
    }
    find_node_pl(t,params,ss->a.nd,&ss->nvs[ss->a.nd]);
    find_node_pl(t,params,ss->b.nd,&ss->nvs[ss->b.nd]);
    find_node_pl(t,params,ss->c.nd,&ss->nvs[ss->c.nd]);
    find_parent_pl(t,params,n,&ss->nvs[n]);
    for(p=ss->e.nd;p!=t->root;p=t->parent[p])
      find_parent_pl(t,params,p,&ss->nvs[p]);
  }
  else { /* non-clock */
    other_neighbors(t,ss->e.nd,opposite(ss->e.rel),&(ss->c),&(ss->d));
    ss->t->branchlen[ss->c.nd] = t->branchlen[ss->c.nd];
    ss->t->branchlen[ss->d.nd] = t->branchlen[ss->d.nd];
    if(runif() < 0.5) 
      SWAP(temp,ss->a,ss->b);
    if(runif() < 0.5) 
      SWAP(temp,ss->c,ss->d);
    old_x = get_edge_length(t,n,ss->a);
    old_y = old_x + get_edge_length(t,n,ss->e);
    old_m = old_y + get_edge_length(t,ss->e.nd,ss->c);
    m = old_m * SAFE_EXP(LAMBDA2 * (runif() - 0.5));
    mratio = m / old_m;
    *hr = mratio * mratio;
    if(runif() < 0.5)
    {
      x = old_x * mratio;
      y = runif() * m;
    }
    else
    {
      x = runif() * m;
      y = old_y * mratio;
    }
    if(x < y) { /* tree topology does not change. */
      ss->rotated = 0;
      set_edge_length(t,n,ss->a,x);
      set_edge_length(t,n,ss->e,y-x);
      set_edge_length(t,ss->e.nd,ss->c,m-y);
    }
    else { /* tree topology changes. */
      ss->rotated = 1;
      set_edge_length(t,n,ss->a,y);
      set_edge_length(t,n,ss->e,x-y);
      set_edge_length(t,ss->e.nd,ss->c,m-x);
      if(ss->d.rel==LEFTCHILD || ss->d.rel==RIGHTCHILD) {
	set_neighbor(t,n,ss->b.rel,ss->d.nd);
	set_neighbor(t,ss->e.nd,ss->d.rel,ss->b.nd);
	siteor(MASK(n), MASK(ss->a.nd), MASK(ss->d.nd));
	if(ss->e.rel==ROOTLEFT || ss->e.rel==ROOTRIGHT) {
	  siteor(MASK(ss->e.nd), MASK(ss->b.nd), MASK(ss->c.nd));
	  restrict_sites_root(t,ss->e.nd);
	  restrict_sites_nonroot(t,ss->e.nd,ss->c.nd);
	  restrict_sites_root(t,n);
	}
	else 
	  restrict_sites_nonroot(t,ss->e.nd,n);
	restrict_sites_nonroot(t,ss->e.nd,ss->b.nd);
	restrict_sites_nonroot(t,n,ss->d.nd);
	restrict_sites_nonroot(t,n,ss->a.nd);
      }
      else { /* d must be the sibling or parent of e, so n is not a child of 
		the root. */
	set_neighbor(t,n,ss->a.rel,ss->c.nd);
	set_neighbor(t,ss->e.nd,ss->c.rel,ss->a.nd);
	siteor(MASK(n), MASK(ss->b.nd), MASK(ss->c.nd));
	restrict_sites_nonroot(t,ss->e.nd,n);
	restrict_sites_nonroot(t,ss->e.nd,ss->a.nd);
	restrict_sites_nonroot(t,n,ss->c.nd);      
	restrict_sites_nonroot(t,n,ss->b.nd);      
      }
    }

    find_node_pl(t,params,ss->a.nd,&ss->nvs[ss->a.nd]);
    if(t->parent[ss->e.nd]!=ss->c.nd)
      find_node_pl(t,params,ss->c.nd,&ss->nvs[ss->c.nd]);
    find_parent_pl(t,params,n,&ss->nvs[n]);
    for(p=ss->e.nd;p!=t->root;p=t->parent[p])
      find_parent_pl(t,params,p,&ss->nvs[p]);
  }
  find_root_ll(t,params);
  free(hgt);
}

void local_restore(Tree *t,  Saved_state ss,const Parameters params) 
{
  int p;
  int i,b,pb;

  t->loglikelihood = ss->loglikelihood;
  t->branchlen[ss->n] = ss->t->branchlen[ss->n];
  t->branchlen[ss->a.nd] = ss->t->branchlen[ss->a.nd];
  t->branchlen[ss->b.nd] = ss->t->branchlen[ss->b.nd];
  t->branchlen[ss->c.nd] = ss->t->branchlen[ss->c.nd];
  t->branchlen[ss->e.nd] = ss->t->branchlen[ss->e.nd];
  if(!mclock)
    t->branchlen[ss->d.nd] = ss->t->branchlen[ss->d.nd];
  if(ss->rotated)
    if(mclock) {
      set_neighbor(t,ss->n,ss->d.rel,ss->d.nd);
      set_neighbor(t,ss->e.nd,ss->c.rel,ss->c.nd);
      siteor(MASK(ss->n), MASK(ss->a.nd), MASK(ss->b.nd));
      restrict_sites_node(t,ss->e.nd,ss->n);
      restrict_sites_node(t,ss->e.nd,ss->c.nd);
      restrict_sites_nonroot(t,ss->n,ss->a.nd);
      restrict_sites_nonroot(t,ss->n,ss->b.nd);
    }
    else if(ss->d.rel==LEFTCHILD || ss->d.rel==RIGHTCHILD) { /* non-clock */ 
      set_neighbor(t,ss->n,ss->b.rel,ss->b.nd);
      set_neighbor(t,ss->e.nd,ss->d.rel,ss->d.nd);
      siteor(MASK(ss->n), MASK(t->lchild[ss->n]), MASK(t->rchild[ss->n]));
      if(ss->e.rel==ROOTLEFT || ss->e.rel==ROOTRIGHT) {
	siteor(MASK(ss->e.nd), MASK(t->lchild[ss->e.nd]), 
	       MASK(t->rchild[ss->e.nd]));
	restrict_sites_root(t,ss->e.nd);
	restrict_sites_nonroot(t,ss->e.nd,ss->c.nd);
	restrict_sites_root(t,ss->n);
      }
      else
	restrict_sites_nonroot(t,ss->e.nd,ss->n);
      restrict_sites_nonroot(t,ss->e.nd,ss->d.nd);
      restrict_sites_nonroot(t,ss->n,ss->b.nd);
      restrict_sites_nonroot(t,ss->n,ss->a.nd);
    }
    else {
      set_neighbor(t,ss->n,ss->a.rel,ss->a.nd);
      set_neighbor(t,ss->e.nd,ss->c.rel,ss->c.nd);
      siteor(MASK(ss->n), MASK(t->lchild[ss->n]), MASK(t->rchild[ss->n]));
      restrict_sites_nonroot(t,ss->e.nd,ss->n);
      restrict_sites_nonroot(t,ss->e.nd,ss->c.nd);
      restrict_sites_nonroot(t,ss->n,ss->a.nd);
      restrict_sites_nonroot(t,ss->n,ss->b.nd);
    }
  restore_node_pl(ss->a.nd,ss->nvs[ss->a.nd]);
  if(mclock)
    restore_node_pl(ss->b.nd,ss->nvs[ss->b.nd]);
  if(t->parent[ss->e.nd]!=ss->c.nd)
    restore_node_pl(ss->c.nd,ss->nvs[ss->c.nd]);
  restore_parent_pl(ss->n,ss->nvs[ss->n]);
  for(p=ss->e.nd;p!=t->root;p=t->parent[p])
    restore_parent_pl(p,ss->nvs[p]);
}

void print_top_int(FILE *f, const Tree *t, int i, int last,
		   int check[])
{
  check[i] = 1;
  if(LEAF(i))
    fprintf(f,"%d",i);
  else {
    fprintf(f,"(%d ",i);
    print_top_int(f,t,t->lchild[i],0,check);
    print_top_int(f,t,t->rchild[i],1,check);
    fprintf(f,")");
  }
  if(!last)
    fprintf(f," ");
}

void print_top(FILE *f, const Tree *t)
{
  int *check,i,sum=0;

  if(!(check=(int *)malloc(sizeof(int)*num_nodes)))
    out_of_space("check"); 
  for(i=0;i<num_nodes;i++)
    check[i] = 0;
  print_top_int(f,t,t->root,0,check);
  fprintf(f,"\n");
  for(i=0;i<num_nodes;i++)
    sum += check[i];
  if(sum!=num_nodes)
    fprintf(f,"Error: Too few nodes in tree.\n");
  free(check);
}

void fprint_param(FILE *f, int cycle, int *accept, int *kaccept, 
		  const Parameters param, const Tree *t, int interval)
{
  int i,b;

  fprintf(f,"%6d %6.1lf %1.3lf %1.6lf ",
	  cycle,t->loglikelihood,(double)(*accept)/(double)interval,
	  valley_win);
  for(i=0;i<num_tags;i++)
  {
    fprintf(f,"%lf ",param[i].kappa);
    fprintf(f,"%lf ",param[i].theta);
    for(b=0;b<NUM_BASES;b++)
      fprintf(f,"%lf ",param[i].pi[b]);
  }
  fprintf(f,"%1.3lf %lf\n",
	  (double)(*kaccept)*(double)param_update_interval/(double)interval,
	  tree_radius(t));
  *accept = 0;
  *kaccept = 0;
}

void update_burn_intervals(Tree *t, int cycle, Parameters param,
				  Accept_counts *ac, Files *files)
{
  if (cycle % sample_interval == 0) {
    fprint_tree_topology(files->ftop,t);
    fprint_param(files->fkappa,cycle,&ac->accepts,&ac->kaccepts,param,t,
		 sample_interval);
    fprintf(files->flogp,"%lf\n",t->loglikelihood);
  }
  if (cycle % tune_interval == 0)
    if((double)(ac->accept)/(double)tune_interval < .15)
      valley_win *= .5;
  if (cycle % window_interval == 0) {
    fprint_tree_topology(stdout,t);
    fprint_param(stdout,cycle,&ac->accept,&ac->kaccept,param,t,
		 window_interval);
  }
}

void update_main_intervals(Tree *t, int cycle, int num_sc, 
				   Parameters param, Accept_counts *ac, 
				   Files *files)
{
  if (cycle % sample_interval == 0) {
    fprint_tree_topology(files->ftop,t);
    fprint_param(files->fkappa,cycle,&ac->accepts,&ac->kaccepts,param,t,
		 sample_interval*num_sc);
    fprintf(files->flogp,"%lf\n",t->loglikelihood);
  }
  if (cycle % window_interval == 0) {
    fprint_tree_topology(stdout,t);
    fprint_param(stdout,cycle,&ac->accept,&ac->kaccept,param,t,
		 window_interval*num_sc);
  }
}

void burn_global(Tree *t, Parameters params,
		 Files *files, Accept_counts *ac)
{
  int cycle;
  double old_loglikelihood;
  Saved_state ss;
  
  ss = make_saved_state();
  for(cycle=1; cycle<= burn; cycle++) {
    old_loglikelihood = t->loglikelihood;
    global_update(t,ss,params);
    if(SAFE_EXP(t->loglikelihood - old_loglikelihood) > runif()) {
      ac->accept++;
      ac->accepts++;
    }
    else
      global_restore(t,ss); 
    update_burn_intervals(t,cycle,params,ac,files);
    
  }
  free_saved_state(ss);
}

void burn_local(Tree *t, Parameters params,
		Files *files, Accept_counts *ac)
{
  int cycle, k, n, temp, bad1, bad2;
  double old_loglikelihood;
  Saved_state ss;
  double hr;

  ss = make_saved_state();
  for(cycle=1; cycle<= burn; cycle++) {
    if(mclock) {
      n = (int)(runif()*(num_leaves-2)+num_leaves);
      if(n >= t->root)
	n++;
    }
    else {
      bad1 = t->root;
      bad2 = LEAF(t->rchild[bad1]) ? t->lchild[bad1] : t->rchild[bad1];
      if(bad1 > bad2)
	SWAP(temp,bad1,bad2);
      n = (int)(runif()*(num_leaves-3)+num_leaves);
      if(n >= bad1)
	n++;
      if(n >= bad2)
	n++;
    }
    old_loglikelihood = t->loglikelihood;
    local_update(t,n,ss,params,&hr);
    if(SAFE_EXP(t->loglikelihood - old_loglikelihood) * hr > runif()) {
      ac->accept++;
      ac->accepts++;
    }
    else
      local_restore(t,ss, params); 
    update_burn_intervals(t,cycle,params,ac,files);
  }
  free_saved_state(ss);
}

void check_param_update(Tree *t, int cycle, Parameters cur_params,
			       Accept_counts *ac)
{
  double old_loglikelihood;
  struct parameters *prop_params;
  Saved_state ss;
  double hr=1;
  int i;

  if(!(prop_params=(struct parameters *)malloc(sizeof(struct parameters)
					       * num_tags)))
    out_of_space("prop_params");
  ss = make_saved_state();
  if (param_update_interval && ((cycle-1) % param_update_interval == 0)) {
    update_params(prop_params,cur_params);
    old_loglikelihood = t->loglikelihood;
    find_loglikelihood(t, prop_params ,ss);
    if(update_theta)
      hr *= hastings_theta(theta_const,tag_weights,cur_params,prop_params,
			   num_tags);
    if(update_pi)
      for(i=0;i<num_tags;i++)
	hr *= hastings_pi(pi_const,cur_params[i].pi,prop_params[i].pi,
			  NUM_BASES);
    if(SAFE_EXP(t->loglikelihood - old_loglikelihood) * hr > runif()) {
      copy_params(cur_params,prop_params);
      ac->kaccept++;
      ac->kaccepts++;
    }
    else
      restore_loglikelihood(t,ss);
  }
  free_saved_state(ss);
  free(prop_params);
}

void main_global(Tree *t, Parameters params,
		 Files *files, Accept_counts *ac)
{
  int cycle, k;
  double old_loglikelihood;
  Saved_state ss;

  ss = make_saved_state();
  for(cycle=1; cycle<= num_cycles; cycle++) {
    check_param_update(t,cycle,params,ac);
    old_loglikelihood = t->loglikelihood;
    global_update(t,ss,params);
    if(SAFE_EXP(t->loglikelihood - old_loglikelihood) > runif()) {
      ac->accept++;
      ac->accepts++;
    }
    else
      global_restore(t,ss); 
    update_main_intervals(t,cycle,1,params,ac,files);
  }
  free_saved_state(ss);
}

void main_local(Tree *t, Parameters params,
		Files *files, Accept_counts *ac)
{
  int cycle, k, n, temp, bad1, bad2;
  double old_loglikelihood;
  Saved_state ss;
  double hr;

  ss = make_saved_state();
  for(cycle=1; cycle<= num_cycles; cycle++) {
    check_param_update(t,cycle,params,ac);
    if(mclock) {
      n = (int)(runif()*(num_leaves-2)+num_leaves);
      if(n >= t->root)
	n++;
    }
    else {
      bad1 = t->root;
      bad2 = LEAF(t->rchild[bad1]) ? t->lchild[bad1] : t->rchild[bad1];
      if(bad1 > bad2)
	SWAP(temp,bad1,bad2);
      n = (int)(runif()*(num_leaves-3)+num_leaves);
      if(n >= bad1)
	n++;
      if(n >= bad2)
	n++;
    }
    old_loglikelihood = t->loglikelihood;
    local_update(t,n,ss,params,&hr);
    if(SAFE_EXP(t->loglikelihood - old_loglikelihood) * hr > runif()) {
      ac->accept++;
      ac->accepts++;
    }
    else
      local_restore(t,ss, params); 
    update_main_intervals(t,cycle,1,params,ac,files);
  }
  free_saved_state(ss);
}

void create_arrays()
{
  if(!(frag=(double *)malloc(sizeof(double)*num_nodes*
			     num_unique_sites*NUM_BASES)))
    out_of_space("frag");
  if(!(saved_frag=(double *)malloc(sizeof(double)*num_nodes*
				   num_unique_sites*NUM_BASES)))
    out_of_space("saved_frag");
  if(!(kfrag=(double *)malloc(sizeof(double)*num_nodes*num_tags*NUM_BASES*
			      NUM_BASES)))
    out_of_space("kfrag");
  if(!(saved_kfrag=(double *)malloc(sizeof(double)*num_nodes*num_tags*
				    NUM_BASES* NUM_BASES)))
    out_of_space("saved_kfrag");
  if(!(ufrag=(double *)malloc(sizeof(double)*num_leaves*num_tags*
			      NUM_SYMBOLS*NUM_BASES)))
    out_of_space("ufrag");
  if(!(saved_ufrag=(double *)malloc(sizeof(double)*num_leaves*num_tags*
				    NUM_SYMBOLS*NUM_BASES)))
    out_of_space("saved_ufrag");
  if(!(lsites=(double **)malloc(sizeof(double *)*num_nodes*
				num_unique_sites)))
    out_of_space("lsites");
  if(!(csites=(int *)malloc(sizeof(int)*num_nodes*num_unique_sites)))
    out_of_space("csites");
  if(!(first_csite=(int *)malloc(sizeof(int)*num_nodes*(num_tags+1))))
    out_of_space("first_csite");
}

void initialize(Files *files, Parameters *params, Tree *cur_tree)
{

  fprintf(stdout,"Reading Runtime Parameters....\n");
  read_input_params(stdin);

  fprintf(stdout,"Opening Files....\n");
  open_files(files);

  fprintf(stdout,"Reading Data file....\n");
  read_datafile(data_file);

  fprintf(stdout,"Creating initial parameters....\n");
  create_initial_parameters(params);

  generate_masks();

  fprintf(stdout,"Finding Unique Sites....\n");
  find_unique_sites();

  fprintf(stdout,"Creating arrays....\n");
  create_arrays();

  fprintf(stdout,"Writing initial summary information to file....\n");
  print_initial_information(files->fout,*params);
  fclose(files->fout);

  fprintf(stdout,"Creating initial tree....\n");
  create_initial_tree(cur_tree,*params);
}

int main()
{
  Tree t;
  Parameters params;
  Files files;
  Accept_counts ac = {0,0,0,0};
  int i;

  initialize(&files,&params,&t);

  fprint_tree_topology(stdout,&t);

  fprintf(stdout,"Starting burn in....\n");

  switch (burn_algorithm) {
  case GLOBAL : burn_global(&t,params,&files,&ac); break;
  case LOCAL: burn_local(&t,params,&files,&ac); break;
  }

  fprintf(stdout,"Entering main loop....\n");
  switch (main_algorithm) {
  case GLOBAL: main_global(&t,params,&files,&ac); break;
  case LOCAL: main_local(&t,params,&files,&ac); break;
  }
  /*check_calc(&t,params);*/

  fprint_tree_topology(stdout,&t);
  fprint_tree(files.ftree,&t);
  fprint_tree(stdout,&t);
  close_files(&files);
  return 0;
}
